/** \file PclOpFluidBodyInteraction.h

    \brief Particle operation to interact vorton-based fluid with rigid bodies.

    \see Accompanying articles for more information:
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-1/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-2/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-3/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-4/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-5/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-6/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-7/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-8/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-9/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-10/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-11/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-12/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-13/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-14/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-15/

        - http://www.gamasutra.com/view/feature/4164/sponsored_feature_fluid_.php
        - http://www.gamasutra.com/view/feature/4176/sponsored_feature_fluid_.php

        - http://www.mijagourlay.com/

    \author Copyright 2009-2013 Dr. Michael Jason Gourlay; All rights reserved.  Contact me at mijagourlay.com for licensing.
*/
#ifndef PARTICLE_OPERATION_FLUID_BODY_INTERACTION_H
#define PARTICLE_OPERATION_FLUID_BODY_INTERACTION_H

#include "Particles/Operation/particleOperation.h"

/** Particle operation to interact vorton-based fluid with rigid bodies.
*/
class PclOpFluidBodyInteraction : public IParticleOperation
{
    public:
        PclOpFluidBodyInteraction()
            : mDensityGrid( 0 )
        #if REDUCE_CONVERGENCE
            , mVortonIndicesGrid( 0 )
        #endif
            , mAmbientFluidDensity( 0.0f )
            , mFluidSpecificHeatCapacity( 1.0f )
            , mGravityAcceleration( Vec3( 0.0f , 0.0f , 0.0f ) )
            , mPhysicalObjects( 0 )
            , mRespectAngVel( true )
        {}

        VIRTUAL_CONSTRUCTORS( PclOpFluidBodyInteraction ) ;

    #if REDUCE_CONVERGENCE
        ////////////////////////////////////////////////////////////////////////////////
        // This REDUCE_CONVERGENCE code was a temporary work-around for issues in the
        // vorton simulation introduced in Part 14.  It uses an ad-hoc divergence
        // reduction scheme.  Do not spend time trying to unravel its mysteries.  Expect
        // it to disappear soon.
        ////////////////////////////////////////////////////////////////////////////////
        /** Reduce divergence in number density of particles, while satisfying boundary conditions.
        */
        void SolveBoundaryConditionsAndReduceDivergence(  VECTOR< Particle > & particles , float timeStep , unsigned uFrame )
        {
            extern float gVortonSim_DisplacementMax    ;
            extern int   gVortonSim_NumRelaxationIters ;
            float displacementMax     = - FLT_MAX ;
            float displacementMaxPrev =   FLT_MAX ;
            int iter ;
            for( iter = 0 ; iter < 3 /* 32 */ ; ++ iter )
            {
                FluidBodySim::SolveBoundaryConditions( particles , mAmbientFluidDensity , mFluidSpecificHeatCapacity , * mPhysicalObjects , mRespectAngVel ) ;
                if( mVortonIndicesGrid )
                {
                    displacementMax = VortonSim::ReduceDivergence( reinterpret_cast< VECTOR< Vorton > * >( & particles ) , * mVortonIndicesGrid ) ;
                }
                if( ( displacementMax < FLT_EPSILON ) && ( displacementMaxPrev < FLT_EPSILON ) )
                {   // Worst displacement was small enough to ignore.
                    break ; // Stop relaxing particles.
                }
                displacementMaxPrev = displacementMax ; // Remember worst displacement for reporting.
                gVortonSim_DisplacementMax = displacementMaxPrev ;    // Report previous worst displacement (since last one was zero and not interesting).
            }
            gVortonSim_NumRelaxationIters = iter ;
        }
    #endif


        void Operate(  VECTOR< Particle > & particles , float timeStep , unsigned uFrame )
        {
            QUERY_PERFORMANCE_ENTER ;

        #if REDUCE_CONVERGENCE
            SolveBoundaryConditionsAndReduceDivergence( particles , timeStep , uFrame ) ;
        #else
            FluidBodySim::SolveBoundaryConditions( particles , mAmbientFluidDensity , mFluidSpecificHeatCapacity , * mPhysicalObjects , mRespectAngVel ) ;
        #endif

            if( mDensityGrid != 0 )
            {   // Caller wants to use this fluid to buoy bodies submerged in it.
                FluidBodySim::BuoyBodies( * mDensityGrid , mAmbientFluidDensity , mGravityAcceleration , * mPhysicalObjects ) ;
            }
            QUERY_PERFORMANCE_EXIT( PclOp_FluidBodyInteraction ) ;
        }


        const UniformGrid< float > *            mDensityGrid                ;   ///< Grid of fluid density values.

    #if REDUCE_CONVERGENCE
        const UniformGrid< VECTOR< unsigned > > * mVortonIndicesGrid        ;
    #endif

        float                                   mAmbientFluidDensity        ;   ///< Fluid density in the absence of fluid particles
        float                                   mFluidSpecificHeatCapacity  ;   ///< Specific heat capacity of the fluid.
        Vec3                                    mGravityAcceleration        ;   ///< Acceleration due to gravity
        VECTOR< Impulsion::PhysicalObject * > * mPhysicalObjects            ;   ///< Dynamic array of addresses of physical objects.
        bool                                    mRespectAngVel              ;   ///< Whether to use particle angular velocity in interaction
} ;

#endif
