/** \file pclOpEmit.cpp

    \brief Operation to emit particles.

    \see http://www.mijagourlay.com/

    \author Copyright 2009-2012 Dr. Michael Jason Gourlay; All rights reserved.  Contact me at mijagourlay.com for licensing.
*/

#include <stdlib.h>

#include "Core/Performance/perf.h"
#include "Core/SpatialPartition/uniformGridMath.h"
#include "../particle.h"
#include "pclOpEmit.h"


static const float  sOneMinusEpsilon        = 1.0f - FLT_EPSILON ;




/** Create particles within the given region.

    \param particles    (out) Dynamic array of particles to populate.

    \param uniformGrid  Region within which to emit particles.

    \param multiplier   Number of particles to emit along each dimension
                        within each cell of the given uniformGrid.
                        This is the cube root of the number of particles to emit
                        per cell.

    \param vortons      Optional address of Dynamic array of vortons.  If
                        provided then particles are only emitted where vortons exist.
                        This allows for placing tracers at the location of vortons.

*/
void PclOpEmit::Emit( VECTOR< Particle > & particles , const UniformGridGeometry & uniformGrid , unsigned multiplier , const VECTOR< Particle > * vortons )
{
    static const float ambientFluidDensity = 1.0f ; // TODO: FIXME: Pass in as a parameter.

    const Vec3      vSpacing        = uniformGrid.GetCellSpacing() ;
    // Must keep particles away from maximal boundary by at least cell.  Note the +vHalfSpacing in loop.
    //const unsigned  begin[3]        = { 1*uniformGrid.GetNumCells(0)/8 , 1*uniformGrid.GetNumCells(1)/8 , 1*uniformGrid.GetNumCells(2)/8 } ;
    //const unsigned  end[3]          = { 7*uniformGrid.GetNumCells(0)/8 , 7*uniformGrid.GetNumCells(1)/8 , 7*uniformGrid.GetNumCells(2)/8 } ;
    //const unsigned  begin[3]        = { 1,1,1 } ;
    //const unsigned  end[3]          = { uniformGrid.GetNumCells(0) - 1 , uniformGrid.GetNumCells(1) - 1 , uniformGrid.GetNumCells(2) - 1 } ;
    const unsigned  begin[3]        = { 0,0,0 } ;
    const unsigned  end[3]          = { uniformGrid.GetNumCells(0) , uniformGrid.GetNumCells(1) , uniformGrid.GetNumCells(2) } ;
    const float     pclSize         = 0.5f * powf( vSpacing.x * vSpacing.y * vSpacing.z , 1.0f / 3.0f ) / float( multiplier ) ;
    unsigned        idx[3]          ;

    const unsigned  nt[3]           = { multiplier , multiplier , multiplier } ;
    bool            emitParticle    = true ;

    UniformGrid< float > densityDeviationGrid  ; // used to mask out tracer particles
    float densityDeviationMin =  FLT_MAX ;
    float densityDeviationMax = -FLT_MAX ;

    if( vortons )
    {   // Caller provided vortons, to seed tracers.

        densityDeviationGrid.Clear() ;
        densityDeviationGrid.CopyShape( uniformGrid ) ;
        densityDeviationGrid.Init( 0.0f );

        // Populate density deviation grid based on vortons.
        // Note that this is only a rough approximation of a proper density grid.
        // Here we only want to know whether a given cell has deviant density,
        // to get the overall shape of the density distribution.
        // This density grid will NOT have the same total mass as the particles.
        const size_t numVortons = vortons->Size() ;
        for( size_t uVorton = 0 ; uVorton < numVortons ; ++ uVorton )
        {   // For each vorton in the array...
            const Particle  &   rVorton     = (*vortons)[ uVorton ] ;
            const Vec3      &   rPosition   = rVorton.mPosition   ;
            
            // Convert from unsigned multiplicative gain to signed additive deviation.
            const float densityDeviation = rVorton.mDensity - ambientFluidDensity ;
            densityDeviationGrid.Accumulate( rPosition , densityDeviation ) ;
        }
        // Aggregate density deviation min and max.
        for( idx[2] = begin[2] ; idx[2] <= end[2] ; ++ idx[2] )
        for( idx[1] = begin[1] ; idx[1] <= end[1] ; ++ idx[1] )
        for( idx[0] = begin[0] ; idx[0] <= end[0] ; ++ idx[0] )
        {   // For each grid cell...
            const size_t offset = densityDeviationGrid.OffsetFromIndices( idx ) ;
            const float densityDeviation = densityDeviationGrid[ offset ] ;
            densityDeviationMin = Min2( densityDeviation , densityDeviationMin ) ;
            densityDeviationMax = Max2( densityDeviation , densityDeviationMax ) ;
        }
    }
    const float densityDeviationScale   = Max2( fabsf( densityDeviationMax ) , fabsf( densityDeviationMin ) ) ;
    float       probMax                 = 0.0f ;

    // Shift each particle to center the distribution within each cell.
    const Vec3  vShift                  = 0.5f * vSpacing / float( multiplier ) ;

    // Jitter should distribute particles within their little subdomains,
    // and must not place particles outside grid, even by a tiny amount.
    const Vec3  jitterMagnitude         = 1.9999f * sOneMinusEpsilon * vShift ;

    for( idx[2] = begin[2] ; idx[2] < end[2] ; ++ idx[2] )
    for( idx[1] = begin[1] ; idx[1] < end[1] ; ++ idx[1] )
    for( idx[0] = begin[0] ; idx[0] < end[0] ; ++ idx[0] )
    {   // For each grid cell...
        Vec3 vPosMinCorner ;
        uniformGrid.PositionFromIndices( vPosMinCorner , idx ) ;
        Particle pcl ;
        pcl.mVelocity	        = Vec3( 0.0f , 0.0f , 0.0f ) ;
        pcl.mOrientation	    = Vec3( 0.0f , 0.0f , 0.0f ) ;
        pcl.mAngularVelocity	= Vec3( 0.0f , 0.0f , 0.0f ) ;
        pcl.mDensity            = ambientFluidDensity ;
    #if ENABLE_FIRE
        pcl.mFuelFraction       = 0.0f ;
        pcl.mFlameFraction      = 0.0f ;
        pcl.mSmokeFraction      = 1.0f ;
    #endif
        pcl.mSize		        = pclSize ;
        pcl.mBirthTime          = 0 ;
        //const unsigned  offset    = densityDeviationGrid.OffsetFromIndices( idx ) ;

        if( ( ! vortons )                           // Either vorton array was not provided...
            ||  ( densityDeviationScale != 0.0f )   // ...or it was provided and there is mass variation in this domain.
            )
        {   // Want to place tracer particles here.
            unsigned it[3] ;
            for( it[2] = 0 ; it[2] < nt[2] ; ++ it[2] )
            for( it[1] = 0 ; it[1] < nt[1] ; ++ it[1] )
            for( it[0] = 0 ; it[0] < nt[0] ; ++ it[0] )
            {   // For each subdomain within each grid cell...
                const Vec3 vDisplacement(   float( it[0] ) / float( nt[0] ) * vSpacing.x ,
                                            float( it[1] ) / float( nt[1] ) * vSpacing.y ,
                                            float( it[2] ) / float( nt[2] ) * vSpacing.z ) ;

                const Vec3 jitter = RandomSpread( jitterMagnitude ) ;
                pcl.mPosition   = vPosMinCorner + vDisplacement + vShift + jitter ;
                if( vortons )
                {   // Vortons were passed in so assign density based on
                    float densityDeviationHere ;
                    densityDeviationGrid.Interpolate( densityDeviationHere , pcl.mPosition ) ;
                    // Convert from additive deviation to multiplicative gain. (Inverse of above.)
                    pcl.mDensity = ( densityDeviationHere + 1.0f ) * ambientFluidDensity ;
                    const float probability = fabsf( densityDeviationHere / densityDeviationScale ) ;
                    probMax = Max2( probMax , probability ) ;
                    //const float dieRoll     = float( rand() ) / float( RAND_MAX ) ;
                    //emitParticle = dieRoll <= probability ;
                    emitParticle = probability >= 0.5f ;
                }
                if( emitParticle )
                {
                    particles.PushBack( pcl ) ;
                }
            }
        }
    }
}




void PclOpEmit::Operate(  VECTOR< Particle > & particles , float timeStep , unsigned uFrame )
{
    QUERY_PERFORMANCE_ENTER ;

    const float fNumToEmit = timeStep * mEmitRate + mRemainder ;
    const int   iNumToEmit = int( fNumToEmit ) ;

    // Remember fractional particles for future emission.
    mRemainder = fNumToEmit - float( iNumToEmit ) ;

    for( int iPcl = 0 ; iPcl < iNumToEmit ; ++ iPcl )
    {   // For each new particle to emit...
        particles.PushBack( mTemplate ) ;
        Particle & rParticleNew = particles.Back() ;
        rParticleNew.mPosition          += RandomSpread( mSpread.mPosition          ) ;
        rParticleNew.mVelocity          += RandomSpread( mSpread.mVelocity          ) ;
        rParticleNew.mOrientation       += RandomSpread( mSpread.mOrientation       ) ;
        rParticleNew.mAngularVelocity   += RandomSpread( mSpread.mAngularVelocity   ) ;
        rParticleNew.mDensity           += RandomSpread( mSpread.mDensity           ) ;
    #if ENABLE_FIRE
        rParticleNew.mFuelFraction      += RandomSpread( mSpread.mFuelFraction      ) ;
        rParticleNew.mFlameFraction     += RandomSpread( mSpread.mFlameFraction     ) ;
        rParticleNew.mSmokeFraction     += RandomSpread( mSpread.mSmokeFraction     ) ;
    #endif
        rParticleNew.mSize              += RandomSpread( mSpread.mSize              ) ;
        rParticleNew.mBirthTime          = uFrame ;
    }

    QUERY_PERFORMANCE_EXIT( Particles_Emit ) ;
}
