/** \file inteSiVis.cpp

    \brief Application for interactive simulation and visualization

    \see Accompanying articles for more information:
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-1/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-2/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-3/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-4/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-5/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-6/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-7/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-8/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-9/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-10/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-11/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-12/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-13/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-14/
        - http://software.intel.com/en-us/articles/fluid-simulation-for-video-games-part-15/

        - http://www.gamasutra.com/view/feature/4164/sponsored_feature_fluid_.php
        - http://www.gamasutra.com/view/feature/4176/sponsored_feature_fluid_.php

        - http://www.mijagourlay.com/

    \author Copyright 2009-2012 Dr. Michael Jason Gourlay; All rights reserved.  Contact me at mijagourlay.com for licensing.
*/

#include <stdarg.h>

#include <assert.h>
#if defined( WIN32 )
    #include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include "glut.h"


#include "Core/Math/vec2.h"
#include "Core/Math/vec4.h"
#include "Core/Math/mat4.h"
#include "Core/Performance/perf.h"
#include "Core/SpatialPartition/uniformGridMath.h"

#include "Image/Operation/imgOpApplyGamma.h"
#include "Image/Operation/imgOpBlur.h"
#include "Image/Operation/imgOpDrawBox.h"
#include "Image/Operation/imgOpDrawCircle.h"
#include "Image/Operation/imgOpMakeNoise.h"
#include "Image/Operation/imgOpOverlay.h"
#include "Image/Operation/imgOpLinearGradient.h"
#include "Image/Operation/imgOpRadialGradient.h"
#include "Image/Operation/imgOpTint.h"

#include "Sim/Vorton/vorticityDistribution.h"

#include "inteSiVis.h"

// Private variables --------------------------------------------------------------

static const float  sOneMinusEpsilon        = 1.0f - FLT_EPSILON ;
static const float  sOnePlusEpsilon         = 1.0f + FLT_EPSILON ;
static const float  sOneOver30              = 1.0f / 30.0f ;
static const float  sOneOver60              = 1.0f / 60.0f ;
static const float  sOneOver120             = 1.0f / 120.0f ;
static const float  sTimeStep               = sOneOver30 ;

static InteSiVis * sInstance = 0 ;
static const Vec3 sGravityDirection( 0.0f , 0.0f , -1.0f ) ; ///< Direction of acceleration due to gravity
static const Vec3 sGravityAcceleration( 10.0f * sGravityDirection ) ; ///< Acceleration due to gravity

static const size_t vortonStride                = sizeof( Vorton ) ;
static const size_t vortonOffsetToAngVel        = offsetof( Vorton , mAngularVelocity   ) ;
static const size_t vortonOffsetToSize          = offsetof( Vorton , mSize              ) ;
static const size_t vortonOffsetToDensity       = offsetof( Vorton , mDensity           ) ;

static const size_t tracerStride                = sizeof( Particle ) ;
static const size_t tracerOffsetToAngVel        = offsetof( Particle , mAngularVelocity ) ;
static const size_t tracerOffsetToSize          = offsetof( Particle , mSize            ) ;
static const size_t tracerOffsetToDensity       = offsetof( Particle , mDensity         ) ;
#if ENABLE_FIRE
static const size_t tracerOffsetToFuelFraction  = offsetof( Particle , mFuelFraction    ) ;
static const size_t tracerOffsetToFlameFraction = offsetof( Particle , mFlameFraction   ) ;
static const size_t tracerOffsetToSmokeFraction = offsetof( Particle , mSmokeFraction   ) ;
#endif

static int sMousePrevX = -999 ;
static int sMousePrevY = -999 ;

static const float gScale = 1.0f ;

bool gPrintProfileData = false ;

#if defined( PROFILE )
static const int sFrameCountMax = 600   ;   // In profile builds, run for a fixed number of frames
static size_t    sNumTracersSum = 0     ;   // Sum of current number of tracers
static size_t    sNumVortonsSum = 0     ;   // Sum of current number of vortons
#endif

static Vec4 sColorRampCyan[] =
{   // Cyan
    Vec4( 0.0f , 0.5f , 0.5f , 0.5f ) ,
    Vec4( 0.0f , 0.8f , 0.8f , 0.8f ) ,
    Vec4( 0.0f , 0.9f , 0.9f , 0.9f ) ,
    Vec4( 0.0f , 1.0f , 1.0f , 1.0f ) ,
} ;
static const int NUM_COLORS_IN_RAMP_CYAN = sizeof( sColorRampCyan ) / sizeof( sColorRampCyan[0] ) ;


static Vec4 sColorRampRainbow[] =
{   // Rainbow
    Vec4( 0.0f , 0.0f , 1.0f , 0.3f ) ,
    Vec4( 0.0f , 0.5f , 1.0f , 0.4f ) ,
    Vec4( 0.0f , 0.8f , 0.0f , 0.5f ) ,
    Vec4( 1.0f , 0.5f , 0.0f , 0.6f ) ,
    Vec4( 1.0f , 0.0f , 0.0f , 0.7f ) ,
    Vec4( 1.0f , 0.0f , 0.5f , 0.8f ) ,
} ;
static const int NUM_COLORS_IN_RAMP_RAINBOW = sizeof( sColorRampRainbow ) / sizeof( sColorRampRainbow[0] ) ;

// Functions --------------------------------------------------------------

/** Given a color ramp and a value in [0,1], compute and return a blended color.

    \param colorRamp - array of colors.

    \param numColorsInRamp - number of elements in colorRamp

    \param zeroToOne - value in [0,1] used to choose a color.

    This routine assumes the colors in the ramp should be uniformly distributed
    across the range [0,1] of input values.  So, colorRamp[0] corresponds to 0,
    colorRamp[numColorsInRamp-1] corresponds to 1, and all colorRamp values in
    between are evenly distributed in [0,1].

*/
static inline Vec4 GetColorFromRamp( const Vec4 * colorRamp , int numColorsInRamp , float zeroToOne )
{
    if( IsNan( zeroToOne ) )
    {
        return Vec4( 1.0f , 0.0f , 1.0f , 1.0f ) ;
    }
    if( zeroToOne < 0.0f )
    {
        return Vec4( 1.0f , 0.5f , 1.0f , 1.0f ) ;
    }
    if( zeroToOne > 1.0f )
    {
        return Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) ;
    }
    const float         idxMinFlt               = zeroToOne * float( numColorsInRamp - 1 ) * sOneMinusEpsilon ;
    const int           idxMin                  = int( idxMinFlt ) ;
    // Interpolate between adjacent colors within the ramp.
    const int           idxMax                  = idxMin + 1 ;
    const float         spanBetweenRampElements = sOnePlusEpsilon / float( numColorsInRamp - 1 ) ;
    const float         tween                   = ( zeroToOne - spanBetweenRampElements * float( idxMin ) ) / spanBetweenRampElements ;
    return colorRamp[ idxMin ] * ( 1.0f - tween ) + colorRamp[ idxMax ] * tween ;
}




/** Create fluid vorton particle systems.

    This routine stand proxy for defining particle systems using some
    definitions from a file.
*/
static ParticleGroup * CreateVortonParticleSystem( FluidVortonPclGrpInfo & vortonPclGrpInfo , const float viscosity , const float ambientFluidDensity , const float fluidSpecificHeatCapacity , VECTOR< Impulsion::PhysicalObject * > & physicalObjects )
{
    vortonPclGrpInfo.mParticleGroup = new ParticleGroup() ;

    static const int    killAgeMax  = 90    ;

    vortonPclGrpInfo.mPclOpKillAge          = new PclOpKillAge() ;
    vortonPclGrpInfo.mPclOpKillAge->mAgeMax = killAgeMax ;
    vortonPclGrpInfo.mParticleGroup->PushBack( vortonPclGrpInfo.mPclOpKillAge ) ;

    Particle emitterTemplate , emitterSpread ;
    emitterTemplate.mPosition           = Vec3( -0.5f , 0.0f , 0.0f ) ;
    emitterTemplate.mVelocity           = Vec3( 2.0f , 0.0f , 0.0f ) ; // Mostly useless when Wind is active
    emitterTemplate.mSize               = 0.05f ;   // Also see where Particles::Emit reassigns mSize.  See calculation of pclSize.
    emitterTemplate.mDensity            = ambientFluidDensity ;
#if ENABLE_FIRE
    emitterTemplate.mFuelFraction       = 0.0f ;
    emitterTemplate.mFlameFraction      = 0.0f ;
    emitterTemplate.mSmokeFraction      = 1.0f ;
#endif
    emitterSpread.mPosition             = Vec3( emitterTemplate.mVelocity.x / 30.0f , 0.5f , 0.5f ) ;

    vortonPclGrpInfo.mPclOpEmit = new PclOpEmit() ;
    vortonPclGrpInfo.mPclOpEmit->mTemplate                  = emitterTemplate ;
    vortonPclGrpInfo.mPclOpEmit->mTemplate.mAngularVelocity = FLT_EPSILON * Vec3( 1.0f , 1.0f , 1.0f ) ;
    vortonPclGrpInfo.mPclOpEmit->mSpread                    = emitterSpread ;
    vortonPclGrpInfo.mPclOpEmit->mSpread.mDensity           = 0.0f ; // No variance for density.
    vortonPclGrpInfo.mPclOpEmit->mEmitRate                  = 0.0f ; // Set by InitialConditions
    vortonPclGrpInfo.mParticleGroup->PushBack( vortonPclGrpInfo.mPclOpEmit ) ;

    // PclOpVortonSim must run after finding bounding box for tracers
    // because VortonSim populates the velocity grid, which relies on
    // the grid dimensions being set from the bounding box.
    // VortonSim must rum before assign-velocity-from-field because that
    // accesses the velocity grid populated by PclOpVortonSim.
    vortonPclGrpInfo.mPclOpVortonSim = new PclOpVortonSim() ;
    //vortonPclGrpInfo.mPclOpVortonSim->mMinCorner = & mPclOpFindBoundingBox.GetMinCorner() ;
    //vortonPclGrpInfo.mPclOpVortonSim->mMaxCorner = & mPclOpFindBoundingBox.GetMaxCorner() ;
    vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetVortons( reinterpret_cast< VECTOR< Vorton > * >( & vortonPclGrpInfo.mParticleGroup->GetParticles() ) ) ;
    vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetViscosity( viscosity ) ;
    vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetAmbientDensity( ambientFluidDensity ) ;
    vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetGravitationalAcceleration( sGravityAcceleration ) ;
    vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSpecificHeatCapacity( fluidSpecificHeatCapacity ) ;

    vortonPclGrpInfo.mParticleGroup->PushBack( vortonPclGrpInfo.mPclOpVortonSim ) ;

#if 1 || COMPUTE_VELOCITY_AT_VORTONS
    // TODO: Verify the new scheme works in all scenarios, then eliminate mPclOpAssignVelocityFromField entirely.
    vortonPclGrpInfo.mPclOpAssignVelocityFromField = 0 ;
#else
#error OBSOLETE.
    // Assign vorton velocities from grid.  This is no longer needed, for at least two reasons.
    //
    // VortonSim can solve velocity at gridpoints, rather than at vortons, and
    // the vortons need to have their velocities assigned in order to be
    // advected and for PclOpFluidBodyInteraction to run. BUT there is probably
    // no advantage to solving those initial velocities on a grid when using
    // Direct or Tree velocity solvers.  (The Poisson solver uses a grid
    // directly.)  Furthermore, PclOpPopulateVelocityGrid (below) populates the
    // velocity grid from vortons after solving boundary conditions.
    //
    // Even when solving velocity on a grid, it still makes sense to assign
    // vorton velocity within UpdateVpm, rather than in a separate particle
    // operation (e.g. here) for other bookkeeping reasons.
    vortonPclGrpInfo.mPclOpAssignVelocityFromField = new PclOpAssignVelocityFromField() ;
    vortonPclGrpInfo.mPclOpAssignVelocityFromField->mVelocityGrid   = & vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid() ;
    vortonPclGrpInfo.mParticleGroup->PushBack( vortonPclGrpInfo.mPclOpAssignVelocityFromField ) ;
#endif

    // Wind runs after assign-velocity-from-field, because that overwrites
    // velocity, and Wind adds to that velocity.
    vortonPclGrpInfo.mPclOpWind = new PclOpWind() ;
    // mVortonPclGrpInfo.mPclOpWind->mWind        = Vec3( 0.0f , 0.0f , 0.0f ) ; Set in InitialConditions
    vortonPclGrpInfo.mPclOpWind->mWindWeight = 1.0f ;
    vortonPclGrpInfo.mPclOpWind->mSrcWeight  = 1.0f ;
    vortonPclGrpInfo.mParticleGroup->PushBack( vortonPclGrpInfo.mPclOpWind ) ;

    // Evolve runs after all operations that set velocity.
    vortonPclGrpInfo.mPclOpEvolve = new PclOpEvolve() ;
    vortonPclGrpInfo.mParticleGroup->PushBack( vortonPclGrpInfo.mPclOpEvolve ) ;

    // Fluid-Body interaction must occur after evolve, because evolve will move
    // particles, potentially outside grids, and fluid-body interaction does not
    // depend on those grids. Likewise, no particle motion can happen between
    // when grid bounds are found and when grids are populated.
    vortonPclGrpInfo.mPclOpFluidBodInte = new PclOpFluidBodyInteraction() ;
    vortonPclGrpInfo.mPclOpFluidBodInte->mDensityGrid               = 0 ;
    vortonPclGrpInfo.mPclOpFluidBodInte->mPhysicalObjects           = & physicalObjects ;
    vortonPclGrpInfo.mPclOpFluidBodInte->mAmbientFluidDensity       = ambientFluidDensity ;
    vortonPclGrpInfo.mPclOpFluidBodInte->mFluidSpecificHeatCapacity = fluidSpecificHeatCapacity ;
    vortonPclGrpInfo.mPclOpFluidBodInte->mGravityAcceleration       = sGravityAcceleration ;
    vortonPclGrpInfo.mPclOpFluidBodInte->mRespectAngVel             = true ;
    vortonPclGrpInfo.mParticleGroup->PushBack( vortonPclGrpInfo.mPclOpFluidBodInte ) ;

#if USE_SMOOTHED_PARTICLE_HYDRODYNAMICS /* This should probably work but is not tested: || COMPUTE_VELOCITY_AT_VORTONS  */ 
    // Populate velocity grid from vortons, after PclOpFluidBodyInteraction,
    // since satisfying boundary conditions will change velocity.  This is done
    // for the benefit of advecting tracer particles.
    vortonPclGrpInfo.mPclOpPopulateVelocityGrid = new PclOpPopulateVelocityGrid() ;
    vortonPclGrpInfo.mPclOpPopulateVelocityGrid->mVelocityGrid = & vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid() ;
    vortonPclGrpInfo.mPclOpPopulateVelocityGrid->mBoundingBox  = const_cast< Vec3 * >( & vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetMinCorner() ) ;
    vortonPclGrpInfo.mParticleGroup->PushBack( vortonPclGrpInfo.mPclOpPopulateVelocityGrid ) ;
#else // When using strictly Vortex Particle Method, computing velocity on grid, then the velocity grid is already populated so this operation is not necessary.
    vortonPclGrpInfo.mPclOpPopulateVelocityGrid = 0 ;
#endif

    return vortonPclGrpInfo.mParticleGroup ;
}




static ParticleGroup * CreateTracerParticleSystem( FluidTracerPclGrpInfo & tracerPclGrpInfo , FluidVortonPclGrpInfo & vortonPclGrpInfo , const float ambientFluidDensity )
{
    tracerPclGrpInfo.mParticleGroup = new ParticleGroup() ;

    tracerPclGrpInfo.mPclOpKillAge = new PclOpKillAge() ;
    static const int    killAgeMax  = 90    ;

    tracerPclGrpInfo.mPclOpKillAge->mAgeMax = killAgeMax ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpKillAge ) ;

    tracerPclGrpInfo.mPclOpAssignDensityFromGrid = new PclOpAssignScalarFromGrid() ;
    tracerPclGrpInfo.mPclOpAssignDensityFromGrid->mMemberOffsetInBytes      = tracerOffsetToDensity ;
    tracerPclGrpInfo.mPclOpAssignDensityFromGrid->mScalarGrid               = & vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetDensityGrid() ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpAssignDensityFromGrid ) ;

#if ENABLE_FIRE
    tracerPclGrpInfo.mPclOpAssignFuelFromGrid = new PclOpAssignScalarFromGrid() ;
    tracerPclGrpInfo.mPclOpAssignFuelFromGrid->mMemberOffsetInBytes = tracerOffsetToFuelFraction ;
    tracerPclGrpInfo.mPclOpAssignFuelFromGrid->mScalarGrid          = & vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetFuelGrid() ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpAssignFuelFromGrid ) ;

    tracerPclGrpInfo.mPclOpAssignFlameFromGrid = new PclOpAssignScalarFromGrid() ;
    tracerPclGrpInfo.mPclOpAssignFlameFromGrid->mMemberOffsetInBytes    = tracerOffsetToFlameFraction ;
    tracerPclGrpInfo.mPclOpAssignFlameFromGrid->mScalarGrid             = & vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetFlameGrid() ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpAssignFlameFromGrid ) ;

    tracerPclGrpInfo.mPclOpAssignSmokeFromGrid = new PclOpAssignScalarFromGrid() ;
    tracerPclGrpInfo.mPclOpAssignSmokeFromGrid->mMemberOffsetInBytes    = tracerOffsetToSmokeFraction ;
    tracerPclGrpInfo.mPclOpAssignSmokeFromGrid->mScalarGrid             = & vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetSmokeGrid() ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpAssignSmokeFromGrid ) ;
#endif

    tracerPclGrpInfo.mPclOpAssignVelocityFromField = new PclOpAssignVelocityFromField() ;
    tracerPclGrpInfo.mPclOpAssignVelocityFromField->mVelocityGrid    =  & vortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid() ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpAssignVelocityFromField ) ;

    tracerPclGrpInfo.mPclOpWind = new PclOpWind() ;
    * tracerPclGrpInfo.mPclOpWind = * vortonPclGrpInfo.mPclOpWind ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpWind ) ;

    tracerPclGrpInfo.mPclOpEvolve = new PclOpEvolve() ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpEvolve ) ;

    tracerPclGrpInfo.mPclOpFluidBodInte = new PclOpFluidBodyInteraction() ;
    tracerPclGrpInfo.mPclOpFluidBodInte->mPhysicalObjects      = vortonPclGrpInfo.mPclOpFluidBodInte->mPhysicalObjects ;
    tracerPclGrpInfo.mPclOpFluidBodInte->mAmbientFluidDensity  = ambientFluidDensity ;
    tracerPclGrpInfo.mPclOpFluidBodInte->mRespectAngVel        = false ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpFluidBodInte ) ;

    // Note that if tracers get emitted outside the bounding box computed last
    // frame, then mVelocityGrid will not encompass them. To be safe, these
    // emitted particles should not be advected on their first frame.  That
    // could lead to weird artifacts where just-emitted particles disobey wind
    // and other advectors, on their first frame.  Could mitigate that either by
    // providing suitable initial velocity, or could move Emit to just after
    // Kill, then each FindBoundingBox call would also include the entire Emit
    // region, in anticipation of particles to be emitted next frame.
    tracerPclGrpInfo.mPclOpEmit = new PclOpEmit() ;
    tracerPclGrpInfo.mPclOpEmit->mTemplate              = vortonPclGrpInfo.mPclOpEmit->mTemplate ;
    tracerPclGrpInfo.mPclOpEmit->mSpread                = vortonPclGrpInfo.mPclOpEmit->mSpread ;
    tracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.x    = vortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.x        ;
    tracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.y    = vortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.y * 0.3f ;
    tracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.z    = vortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.z * 0.3f ;
    tracerPclGrpInfo.mPclOpEmit->mSpread.mDensity       = 0.0f ; // No variance for density. Doesn't really matter since tracers adopt density from vortons eventually.
    tracerPclGrpInfo.mPclOpEmit->mEmitRate              = 0.0f ; // Set by InitialConditions
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpEmit ) ;

    // Finding bounding box must occur last in this group, and before
    // VortonSim in the other group (because VortonSim uses the bounding
    // box to create a grid of appropriate size to capture all relevant
    // terms of the vorticity equation). Safest to run FindBoundingBox last.
    tracerPclGrpInfo.mPclOpFindBoundingBox = new PclOpFindBoundingBox() ;
    tracerPclGrpInfo.mParticleGroup->PushBack( tracerPclGrpInfo.mPclOpFindBoundingBox ) ;

    return tracerPclGrpInfo.mParticleGroup ;
}




ParticleSystem * CreateFluidParticleSystems( FluidVortonPclGrpInfo & vortonPclGrpInfo , FluidTracerPclGrpInfo & tracerPclGrpInfo , const float viscosity , const float ambientFluidDensity , const float fluidSpecificHeatCapacity , VECTOR< Impulsion::PhysicalObject * > & physicalObjects )
{
    ParticleSystem * fluidParticleSystem = new ParticleSystem() ;

    // Set up particle group for vortons.
    CreateVortonParticleSystem( vortonPclGrpInfo , viscosity , ambientFluidDensity , fluidSpecificHeatCapacity , physicalObjects ) ;

    // Set up particle group for passive tracers.
    CreateTracerParticleSystem( tracerPclGrpInfo , vortonPclGrpInfo , ambientFluidDensity ) ;

    // Process tracers before vortons, so vortons know tracer bounding box.
    fluidParticleSystem->PushBack( tracerPclGrpInfo.mParticleGroup ) ;
    fluidParticleSystem->PushBack( vortonPclGrpInfo.mParticleGroup ) ;

    // Patch vortex particle system to reference results from PclOpFindBoundingBox.
    fluidParticleSystem->AddIndirectAddressAssignment(
            IndirectAddress( vortonPclGrpInfo.mPclOpVortonSim       , & vortonPclGrpInfo.mPclOpVortonSim->mMinCorner           )
        ,   IndirectAddress( tracerPclGrpInfo.mPclOpFindBoundingBox , & tracerPclGrpInfo.mPclOpFindBoundingBox->GetMinCorner() ) ) ;

    fluidParticleSystem->AddIndirectAddressAssignment(
            IndirectAddress( vortonPclGrpInfo.mPclOpVortonSim       , & vortonPclGrpInfo.mPclOpVortonSim->mMaxCorner           )
        ,   IndirectAddress( tracerPclGrpInfo.mPclOpFindBoundingBox , & tracerPclGrpInfo.mPclOpFindBoundingBox->GetMaxCorner() ) ) ;

    return fluidParticleSystem ;
}




ParticleSystem * CreateNonFluidParticleSystem( size_t i )
{
    ParticleSystem  * particleSystem = new ParticleSystem ;

    {
        ParticleGroup * pclGrpFlames = new ParticleGroup ;

        {
            PclOpKillAge * pclOpKillAge = new PclOpKillAge ;
            pclOpKillAge->mAgeMax    = 90 ;
            pclGrpFlames->PushBack( pclOpKillAge ) ;
        }
        {
            Particle emitterTemplate ;
            const float gap = 1.0f * float( i ) ;
            emitterTemplate.mPosition               = Vec3( gap - 10.0f , 0.0f , 0.0f ) ;
            emitterTemplate.mVelocity               = Vec3( 0.0f , 0.0f , 0.0f ) ; // Mostly useless when Wind is active
            emitterTemplate.mSize                   = 0.0625f ;   // Also see where Particles::Emit reassigns mSize.  See calculation of pclSize.
            emitterTemplate.mDensity                = 1.0f ; // Actual value does not matter for this case.
        #if ENABLE_FIRE
            emitterTemplate.mFuelFraction           = 0.0f ;
            emitterTemplate.mFlameFraction          = 0.0f ;
            emitterTemplate.mSmokeFraction          = 1.0f ;
        #endif

            Particle emitterSpread ;
            emitterSpread.mPosition                 = Vec3( 0.5f , 0.5f , 0.5f ) ;

            PclOpEmit     * pclOpEmit               = new PclOpEmit ;

            pclOpEmit->mTemplate                    = emitterTemplate ;
            pclOpEmit->mTemplate.mAngularVelocity   = FLT_EPSILON * Vec3( 1.0f , 1.0f , 1.0f ) ;
            pclOpEmit->mSpread                      = emitterSpread ;
            pclOpEmit->mSpread.mDensity             = 0.0f ;
            pclOpEmit->mEmitRate                    = 100.0f ;

            pclGrpFlames->PushBack( pclOpEmit ) ;
        }
        {
            PclOpWind   * pclOpWind = new PclOpWind ;

            pclOpWind->mWindWeight  = 1.0f ;
            pclOpWind->mSrcWeight   = 0.0f ;
            pclOpWind->mWind        = Vec3( 0.0f , 0.0f , 1.0f ) ;
            pclGrpFlames->PushBack( pclOpWind ) ;

        }
        {
            PclOpEvolve * pclOpEvolve = new PclOpEvolve ;
            // Evolve runs after all operations that set velocity.
            pclGrpFlames->PushBack( pclOpEvolve ) ;
        }

        particleSystem->PushBack( pclGrpFlames ) ;
    }

    return particleSystem ;
}


        

/** Construct application for interactive simulation and visualization.

    \param viscosity - Fluid viscosity.

    \param ambientFluidDensity - Ambient fluid density.  This is the density of
            the fluid in the absence of particles.  Particles have a density,
            and when particles are present then they can modify this ambient density.

    \param fluidSpecificHeatCapacity - Heat capacity per unit mass of the fluid.

*/
InteSiVis::InteSiVis()
#if 0
    : mCamera( 320 , 240 )  // Low resolution
#elif 1
    : mCamera( 640 , 480 )  // Standard definition
#elif 1
    : mCamera( 1280 , 960 ) // 4:3 aspect ratio, high definition
#else
    : mCamera( 1280 , 720 ) // 16:9 720p high definition
#endif
    , mDiagnosticVortonRenderer( 0 , vortonStride , vortonOffsetToAngVel , vortonOffsetToSize , vortonOffsetToDensity , mDiagnosticVortonMaterial )
    , mSimpleVortonRenderer    ( 0 , vortonStride , vortonOffsetToAngVel , vortonOffsetToSize , vortonOffsetToDensity , mSimpleVortonMaterial     )
#if ENABLE_FIRE
    , mDyeRenderer   ( 0 , tracerStride , tracerOffsetToAngVel , tracerOffsetToSize , tracerOffsetToDensity         , mDyeMaterial    )
    , mSmokeRenderer ( 0 , tracerStride , tracerOffsetToAngVel , tracerOffsetToSize , tracerOffsetToSmokeFraction   , mSmokeMaterial  )
    , mFlameRenderer ( 0 , tracerStride , tracerOffsetToAngVel , tracerOffsetToSize , tracerOffsetToFlameFraction   , mFlameMaterial  )
    , mFuelRenderer  ( 0 , tracerStride , tracerOffsetToAngVel , tracerOffsetToSize , tracerOffsetToFuelFraction    , mFuelMaterial   )
#else
    , mDyeRenderer   ( 0 , tracerStride , tracerOffsetToAngVel , tracerOffsetToSize , tracerOffsetToDensity         , mDyeMaterial    )
    , mSmokeRenderer ( 0 , tracerStride , tracerOffsetToAngVel , tracerOffsetToSize , tracerOffsetToDensity         , mSmokeMaterial  )
    , mFlameRenderer ( 0 , tracerStride , tracerOffsetToAngVel , tracerOffsetToSize , tracerOffsetToDensity         , mFlameMaterial  )
    , mFuelRenderer  ( 0 , tracerStride , tracerOffsetToAngVel , tracerOffsetToSize , tracerOffsetToDensity         , mFuelMaterial   )
#endif
    , mRenderWindow( 0 )
    , mStatusWindow( 0 )
    , mFrame( 0 )
    , mTimeNow( 0.0 )
    , mTimeStep( sTimeStep )
    , mTimeStepping( PLAY )
    , mInitialized( false )
    , mGridDecorations( GRID_DECO_NONE )
    , mGridField( GRID_FIELD_DENSITY_GRADIENT )
    , mVortonRendering( VORTON_RENDER_NONE )
    , mVortonProperty( VORTON_PROPERTY_VORTICITY )
    , mTracerRendering( TRACER_RENDER_FIRE )
    , mDiagnosticText( DIAG_TEXT_NONE )
    , mEmphasizeCameraTarget( true )
{
    sInstance = this ;


    mMouseButtons[0] = mMouseButtons[1] = mMouseButtons[2] = 0 ;

    InitialConditions( 26 /* 12 */ /* 32 */ /* 28 */) ;

    // Set the camera in motion.
    //mCamera.SetOrbitalTrajectory( Vec3( -0.002f , -0.000f , 0.0f ) ) ;

}




/** Return a value based on the distance of the given position to the camera target.

    Value lies in [0,1] where the value 1 is for any distance smaller than a grid cell.
*/
float InteSiVis::CameraFocusEmphasis( const Vec3 position ) const
{
    if( mEmphasizeCameraTarget )
    {
        const Vec3 &    target               = mCamera.GetTarget() ;
        const Vec3      displacementToLookat = position - target ;
        const float     distToEye            = displacementToLookat.Magnitude() ;
        const float     characteristicSize   = powf( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid().GetCellVolume() , 0.33333333333333333333333f ) ;
        const float     distRatio            = Min2( characteristicSize / distToEye , 1.0f ) ;
        return Pow3( distRatio ) ;
    }
    else
    {
        return 1.0f ;
    }
}




#if USE_ORIENTED_SORTED_PARTICLES
/** Cheap lighting hack to make an internal glow effect inside smoke.

    \param vGlowPos - position of point light

*/
static void SetGlow( const Vec3 vGlowPos )
{
    glEnable( GL_LIGHTING ) ;

    // Code below here belongs in "QdLight"
    // Simplify (and speed up) specular computation
    glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE ) ;

    // Global ambient light
    {
        static const int iLight = GL_LIGHT0 ;
        float globalAmbientLight[] = { 1.0f , 1.0f , 1.0f , 1.0f } ;
        glLightModelfv( GL_LIGHT_MODEL_AMBIENT , globalAmbientLight ) ;
        glEnable( iLight ) ;
    }

    // Glow point light
    {
        glPushMatrix() ;
        static const int iLight = GL_LIGHT1 ;
        const Vec4 diffuse( 18.0 , 9.0 , 0.0 , 1.0 ) ;
        glLightfv( iLight , GL_DIFFUSE , (float*) & diffuse ) ;
        Vec4 pos( vGlowPos , 1.0 ) ;
        glLightfv( iLight , GL_POSITION , (float*) & pos ) ;
        glLightf( iLight , GL_CONSTANT_ATTENUATION , 0.0f ) ;
        glLightf( iLight , GL_LINEAR_ATTENUATION , 0.0f ) ;
        glLightf( iLight , GL_QUADRATIC_ATTENUATION , 10.0 ) ;
        glEnable( iLight ) ;
        glPopMatrix() ;
    }

    // Directional light from above
    {
        glPushMatrix() ;
        static const int iLight = GL_LIGHT2 ;
        const Vec4 ambient( 0.05 , 0.05 , 0.05 , 1.0 ) ;
        glLightfv( iLight , GL_AMBIENT , (float*) & ambient ) ;
        const Vec4 diffuse( 1.0 , 1.0 , 1.0 , 1.0 ) ;
        glLightfv( iLight , GL_DIFFUSE , (float*) & diffuse ) ;
        const Vec4 position( 0.0 , 0.0 , -1.0 , 0.0 ) ;
        glLightfv( iLight , GL_POSITION , (float*) & position ) ;
        glEnable( iLight ) ;
        glPopMatrix() ;
    }
}
#endif




/** Construct application for interactive simulation and visualization.

    \param ic - which initial conditions to impose

    This is just a demonstration of a few initial conditions.
    There is nothing fundamental or especially important about
    the initial conditions that this routine implements.
    These are just examples.

*/
void InteSiVis::InitialConditions( unsigned ic )
{
#if ENABLE_PARTICLE_HISTORY
    extern size_t gPclHistoryFrame ;
    extern bool gPclHistoryFirstRun ;
    if( gPclHistoryFrame != 0 ) gPclHistoryFirstRun = false ; // This is a subsequent run.  History is already populated with the basis for comparison.
    gPclHistoryFrame = 0 ;  // Reset particle history recorder
    ++ gPclHistoryFrame ; // Advance to next frame in particle history buffer.
#endif


    mScenario           = ic  ;
    sInstance->mFrame   = 0   ;
    sInstance->mTimeNow = 0.0 ;

    mPclSysMgr.Clear() ;

    static const float ambientFluidDensity  = 1.0f ;
    {
        static const float viscosity = 0.01f ;
        static const float fluidSpecificHeatCapacity = 10.0f ;

        ParticleSystem * fluidParticleSystem = CreateFluidParticleSystems( mVortonPclGrpInfo , mTracerPclGrpInfo , viscosity , ambientFluidDensity , fluidSpecificHeatCapacity , GetPhysicalObjects() ) ;
        mPclSysMgr.PushBack( fluidParticleSystem ) ;
    }

    SetTallyDiagnosticIntegrals() ;

    GetSpheres().Clear() ;                      // Remove all spherical physical objects from the simulation
    GetBoxes().Clear() ;                        // Remove all box physical objects from the simulation
    mEntities.Clear() ;                         // Remove all simulation entities

    // Set up lights.
    mLights.resize( 2 ) ;
    mLights[0].mPosition = Vec3( 0.7f , -0.2f , 0.7f ) ; // actually a direction
    mLights[0].mColor    = Vec3( 0.9f , 0.9f , 0.9f ) ;
    mLights[0].mType     = QdLight::LT_DIRECTIONAL ;

    mLights[1].mPosition = Vec3( -0.7f , -0.2f , 0.7f ) ; // actually a direction
    mLights[1].mColor    = Vec3( 0.9f , 0.9f , 0.9f ) ;
    mLights[1].mType     = QdLight::LT_DIRECTIONAL ;

    mLights[0].mAmplitudes[ 1 ]  = 0.01f ;
    mLights[0].mFrequencies[ 1 ] = 0.0f ;
    mLights[0].mAmplitudes[ 2 ]  = 0.0f ;
    mLights[0].mFrequencies[ 2 ] = 0.0f ;
    mLights[0].mAmplitudes[ 3 ]  = 0.0f ;
    mLights[0].mFrequencies[ 3 ] = 0.0f ;

    static const float          fRadius                     = 1.0f ;
    static const float          fThickness                  = 1.0f ;
    static const float          fMagnitude                  = 20.0f ;
    static const unsigned       numCellsPerDim              = 16 ;
    static const unsigned       numVortonsMax               = numCellsPerDim * numCellsPerDim * numCellsPerDim ;
    unsigned                    numTracersPerCellCubeRoot   = 5 ;
    VortonSim &                 vortonSim                   = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;
    VECTOR< Vorton > &          vortons                     = * vortonSim.GetVortons() ;
    static const float          veryLargeMass               = 1.0e8f ; // large but not large enough that 1/mass is a denormal.  Denormal values occur when veryLargeMass is around 1.0e34f. Also needs to be small enough that squaring it fits in FLT_MAX.
    const VECTOR< Particle > *  vortonsImplyingTracers      = 0 ;
    Vec3                        tracerGridScale             ( 1.0f , 1.0f , 1.0f ) ;

    mVortonPclGrpInfo.mPclOpKillAge->mAgeMax    = INT_MAX ;                                     // By default, disable vorton killing.
    mTracerPclGrpInfo.mPclOpKillAge->mAgeMax    = mVortonPclGrpInfo.mPclOpKillAge->mAgeMax ;    // By default, disable tracer killing.
    mVortonPclGrpInfo.mPclOpEmit->mEmitRate     = 0.0f ;                                        // By default, disable vorton emission.
    mTracerPclGrpInfo.mPclOpEmit->mEmitRate     = 0.0f ;                                        // By default, disable tracer emission.
    mVortonPclGrpInfo.mPclOpWind->mWind         = Vec3( 0.0f , 0.0f , 0.0f ) ;                  // By default, disable wind.

    mVortonPclGrpInfo.mPclOpFluidBodInte->mDensityGrid = 0 ;

    srand( 1 ) ;    // Seed pseudo-random number generator to make results repeatable.

    switch( ic )
    {   // Switch on initial conditions
        case 0: // vortex ring -- vorticity in [0,1]
            AssignVortons( vortons , fMagnitude , numVortonsMax , VortexRing( fRadius , fThickness , Vec3( 1.0f , 0.0f , 0.0f ) ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 10.f , 0.f , 0.f ) ) ;
            mCamera.SetEye( Vec3( 10.f , -10.0f , 0.0f ) ) ;
        break ;

        case 1: // "jet" vortex ring -- velocity in [0,1]
            AssignVortons( vortons , fMagnitude , numVortonsMax , JetRing( fRadius , fThickness , Vec3( 1.0f , 0.0f , 0.0f ) ) ) ;

            //vortonsImplyingTracers = ( const VECTOR< Particle > * )( & vortons ) ;
            numTracersPerCellCubeRoot -= 2 ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 10.f ,    0.f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 10.f , -10.0f , 0.0f ) ) ;
        break ;

        case 2: // Projectile with no initial spin
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax / 2 , VortexNoise( Vec3( 4.0f * fThickness , 0.125f * fThickness , 0.5f * fThickness ) ) ) ;

            numTracersPerCellCubeRoot ++ ;

            // Add a sphere
            GetSpheres().PushBack( RbSphere( Vec3( -3.0f , 0.0f , 0.0f ) , Vec3( 2.0f , 0.0f , 0.0f ) , 0.02f , 0.2f ) ) ;
            GetSpheres().Back().GetBody()->ApplyImpulsiveTorque( Vec3( 0.0f , 0.0f , 0.0f ) ) ; // Make sphere spin

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 1.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 1.0f , -3.0f , 0.0f ) ) ;
        break ;

        case 3: // Projectile with spin about longitudinal axis: rifled.
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax / 2 , VortexNoise( Vec3( 4.0f * fThickness , 0.3f * fThickness , 0.3f * fThickness ) ) ) ;

            numTracersPerCellCubeRoot -- ;

            // Add a sphere
            GetSpheres().PushBack( RbSphere( Vec3( -3.0f , 0.0f , 0.0f ) , Vec3( 5.0f , 0.0f , 0.0f ) , 0.02f , 0.2f ) ) ;
            GetSpheres()[0].GetBody()->ApplyImpulsiveTorque( Vec3( 0.002f , 0.0f , 0.0f ) ) ; // Make sphere spin

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        break ;

        case 4: // Projectile with spin about transverse axis: curve ball.
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax , VortexNoise( Vec3( 4.0f * fThickness , 0.125f * fThickness , 0.5f * fThickness ) ) ) ;

            numTracersPerCellCubeRoot -- ;

            // Add a sphere.
            GetSpheres().PushBack( RbSphere( Vec3( -3.0f , 0.0f , 0.0f ) , Vec3( 4.0f , 0.0f , 0.0f ) , 0.02f , 0.2f ) ) ;
            GetSpheres().Back().GetBody()->ApplyImpulsiveTorque( Vec3( 0.0f , 0.05f , 0.0f ) ) ; // Make sphere spin

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 1.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 1.0f , -3.0f , 0.0f ) ) ;
        break ;

        case 5: // Spinning sphere in initially stationary fluid
            // Make a box of vortons surrounding the origin.
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax / 2 , VortexNoise( Vec3( 2.0f * fThickness , 2.0f * fThickness , 2.0f * fThickness ) ) ) ;

            // Make a thinner layer of tracers.
            numTracersPerCellCubeRoot += 6 ;
            tracerGridScale = Vec3( 1.0f , 0.125f , 1.0f ) ;

#if 1
            // Add a sphere
            GetSpheres().PushBack( RbSphere( Vec3( 0.f , 0.0f , 0.0f ) , Vec3( 0.f , 0.0f , 0.0f ) , 1000.0f , fThickness * 0.3f ) ) ;
            GetSpheres().Back().GetBody()->ApplyImpulsiveTorque( Vec3( 0.0f , 0.0f , 100.0f ) ) ; // Make sphere spin
#else
            // Add a box
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f ,  0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 1000.0f , Vec3( 1.0f , 1.0f , 1.0f ) ) ) ;
            GetBoxes().Back().GetBody()->ApplyImpulsiveTorque( Vec3( 0.0f , 0.0f , 100.0f ) ) ; // Make box spin
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , 0.0f , 3.0f ) ) ;
        break ;

        case 6: // Sphere in spinning fluid
            AssignVortons( vortons , 2.0f * fMagnitude , numVortonsMax , VortexTube( fThickness , /* variation */ 0.0f , /* width */ 4.0f * fThickness , 2 , 0 ) ) ;

            numTracersPerCellCubeRoot += 3 ;

            // Add a sphere
            GetSpheres().PushBack( RbSphere( Vec3( 0.f , 0.0f , 0.0f ) , Vec3( 0.f , 0.0f , 0.0f ) , 0.5f , 0.1f ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( -2.0f , 0.0f , 0.0f ) ) ;
        break ;

        case 7: // Fluid moves past ball
        {
            // Add a sphere
            GetSpheres().PushBack( RbSphere( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , veryLargeMass , 0.2f ) ) ;
            //GetSpheres()[0].ApplyImpulsiveTorque( Vec3( 0.0f , 0.0f , 0.0f ) ) ; // Make sphere spin

            mVortonPclGrpInfo.mPclOpKillAge->mAgeMax            = 100 ;
            mTracerPclGrpInfo.mPclOpKillAge->mAgeMax            = mVortonPclGrpInfo.mPclOpKillAge->mAgeMax ;

            Particle emitterTemplate , emitterSpread ;
            emitterTemplate.mPosition       = Vec3( -0.2f , 0.0f , 0.0f ) ;
            emitterTemplate.mVelocity       = Vec3( 1.0f , 0.0f , 0.0f ) ; // Mostly useless when Wind is active
            emitterTemplate.mSize           = 0.0625f ;
            emitterTemplate.mDensity        = ambientFluidDensity ;
        #if ENABLE_FIRE
            emitterTemplate.mFuelFraction   = 0.0f ;
            emitterTemplate.mFlameFraction  = 0.0f ;
            emitterTemplate.mSmokeFraction  = 1.0f ;
        #endif
            emitterSpread.mPosition         = Vec3( emitterTemplate.mVelocity.x / 30.0f , 0.5f , 0.5f ) ;

            mVortonPclGrpInfo.mPclOpEmit->mTemplate                     = emitterTemplate ;
            mVortonPclGrpInfo.mPclOpEmit->mTemplate.mAngularVelocity    = FLT_EPSILON * Vec3( 1.0f , 1.0f , 1.0f ) ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread                       = emitterSpread ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread.mDensity              = 0.0f ;
            mVortonPclGrpInfo.mPclOpEmit->mEmitRate                     = 500.0f ;

            mTracerPclGrpInfo.mPclOpEmit->mEmitRate                     = 10000.0f ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate                     = mVortonPclGrpInfo.mPclOpEmit->mTemplate ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate.mSize               = 0.0078f ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread                       = mVortonPclGrpInfo.mPclOpEmit->mSpread ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.x           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.x        ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.y           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.y * 0.3f ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.z           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.z * 0.3f ;

            mVortonPclGrpInfo.mPclOpWind->mWind                         = Vec3( 1.0f , 0.0f , 0.0f ) ;

            numTracersPerCellCubeRoot -- ; // Emit rate controls tracer density, but this effectively controls their radius.

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 1.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 1.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        case 8: // Vortex sheet with spanwise variation
            AssignVortons( vortons , fMagnitude , numVortonsMax , VortexSheet( fThickness , /* variation */ 0.2f , /* width */ 7.0f * fThickness ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -20.0f , 0.0f ) ) ;
        break ;

        case 9: // Vortex tube
            AssignVortons( vortons , fMagnitude , numVortonsMax , VortexTube( fThickness , /* variation */ 0.0f , /* width */ 2.0f * fThickness , 2 , 0 ) ) ;
        break ;

        case 10: // 2 orthogonal vortex tubes
            AssignVortons( vortons , fMagnitude , numVortonsMax , VortexTube( fThickness , /* variation */ 0.0f , /* width */ 4.0f * fThickness , 2 , -1 ) ) ;
            AssignVortons( vortons , fMagnitude , numVortonsMax , VortexTube( fThickness , /* variation */ 0.0f , /* width */ 4.0f * fThickness , 2 ,  1 ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        break ;

        ////////////////////////////////////////////////////////////////////////

        case 11: // 2D sheet
            AssignVortons( vortons , fMagnitude , numVortonsMax , VortexSheet( fThickness , /* variation */ 0.0f , /* width */ 2.0f * fThickness ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -10.0f , 0.0f ) ) ;
        break ;

        case 12: // ball of initially stationary heavy fluid
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBall( fRadius , 2.0f ) , Vec3( 0.0f , 0.0f , 5.0f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot += 2 ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -10.0f , 0.0f ) ) ;
        break ;

        case 13: // Two balls of initially stationary fluid, one heavy, one light, heavy one above light one.
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBall( fRadius , -0.1f ) , Vec3( 0.0f , 0.0f , -2.0f ) ) ;
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBall( fRadius ,  0.1f ) , Vec3( 0.0f , 0.0f ,  2.0f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -5.0f , 0.0f ) ) ;
        break ;

        case 14: // Two boxes of initially stationary fluid, one heavy, one light, light one above heavy one.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) ,  0.6f ) , Vec3( 0.0f , 0.0f , -0.5f ) ) ;
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) , -0.6f ) , Vec3( 0.0f , 0.0f ,  0.5f ) ) ;

            // Set up density grid to facilitate applying buoyancy to bodies.
            mVortonPclGrpInfo.mPclOpFluidBodInte->mDensityGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetDensityGrid() ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            const float sphereRadius = 0.2f ;
            const float neutrallyBuoyantMass = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetAmbientDensity() * 4.0f * PI * Pow3( sphereRadius ) / 3.0f ;
            GetSpheres().PushBack( RbSphere( Vec3( 0.f , 0.0f , 0.5f ) , Vec3( 0.f , 0.0f , 0.0f ) , neutrallyBuoyantMass , sphereRadius ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -3.0f , 0.0f ) ) ;
        }
        break ;

        case 15: // Ball heats fluid
        {
            mVortonPclGrpInfo.mPclOpKillAge->mAgeMax    = 300 ;
            mTracerPclGrpInfo.mPclOpKillAge->mAgeMax    = mVortonPclGrpInfo.mPclOpKillAge->mAgeMax ;

            Particle emitterTemplate , emitterSpread ;
            emitterTemplate.mPosition       = Vec3( 0.0f , 0.0f , 0.0f ) ;
            emitterTemplate.mVelocity       = Vec3( 0.0f , 0.0f , 0.0f ) ; // Mostly useless when Wind is active
            emitterTemplate.mSize           = 0.0625f ;   // Also see where Particles::Emit reassigns mSize.  See calculation of pclSize.
            emitterTemplate.mDensity        = ambientFluidDensity ;
        #if ENABLE_FIRE
            emitterTemplate.mFuelFraction   = 0.0f ;
            emitterTemplate.mFlameFraction  = 0.0f ;
            emitterTemplate.mSmokeFraction  = 0.0f ;
        #endif
            emitterSpread.mPosition         = Vec3( 0.5f , 0.5f , 0.5f ) ;

            mVortonPclGrpInfo.mPclOpEmit->mTemplate                     = emitterTemplate ;
            mVortonPclGrpInfo.mPclOpEmit->mTemplate.mAngularVelocity    = FLT_EPSILON * Vec3( 1.0f , 1.0f , 1.0f ) ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread                       = emitterSpread ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread.mDensity              = 0.0f ;
            mVortonPclGrpInfo.mPclOpEmit->mEmitRate                     = 100.0f ;

            mTracerPclGrpInfo.mPclOpEmit->mEmitRate                     = 5000.0f ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate                     = mVortonPclGrpInfo.mPclOpEmit->mTemplate ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate.mSize               = 0.0105f ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread                       = mVortonPclGrpInfo.mPclOpEmit->mSpread ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.x           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.x ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.y           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.y ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.z           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.z ;

            mVortonPclGrpInfo.mPclOpWind->mWind                         = Vec3( 0.0f , 0.0f , 0.0f ) ;

            // Add a sphere
            GetSpheres().PushBack( RbSphere( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , veryLargeMass , 0.2f ) ) ;
            GetSpheres().Back().GetThermalProperties().SetTemperature( Impulsion::PhysicalObject::sAmbientTemperature + 300.0f ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 2.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -3.0f , 2.0f ) ) ;
        }
        break ;

        case 16: // Fluid combusts and pushes a ball up.
        {
            // Set up lights.
            mLights[0].mPosition    = Vec3( 0.0f , 0.0f , 0.7f ) ;
            mLights[0].mColor       = 1.2f * Vec3( 1.0f , 0.5f , 0.3f ) ;
            mLights[0].mAttenuation = Vec3( 0.0f , 0.2f , 0.8f ) ;
            mLights[0].mType        = QdLight::LT_POINT ;
            // Make light fluctuate to imply it comes from flames.
            mLights[0].mAmplitudes [ 1 ] = 0.1f ;
            mLights[0].mFrequencies[ 1 ] = 2.0f ;
            mLights[0].mAmplitudes [ 2 ] = 0.05f ;
            mLights[0].mFrequencies[ 2 ] = 5.0f ;
            mLights[0].mAmplitudes [ 3 ] = 0.02f ;
            mLights[0].mFrequencies[ 3 ] = 10.0f ;

            mLights[1].mPosition    = Vec3( 0.0f , 0.0f , 1.0f ) ; // actually a direction
            mLights[1].mColor       = 0.2f * Vec3( 1.0f , 1.0f , 1.0f ) ;
            mLights[1].mType        = QdLight::LT_DIRECTIONAL ;

            mVortonPclGrpInfo.mPclOpKillAge->mAgeMax    = 300 ;
            mTracerPclGrpInfo.mPclOpKillAge->mAgeMax    = mVortonPclGrpInfo.mPclOpKillAge->mAgeMax ;

            Particle emitterTemplate , emitterSpread ;
            emitterTemplate.mPosition       = Vec3( 0.0f , 0.0f , 0.0f ) ;
            emitterTemplate.mVelocity       = Vec3( 0.0f , 0.0f , 0.0f ) ; // Mostly useless when Wind is active
            emitterTemplate.mSize           = 0.055f ;   // Also see where Particles::Emit reassigns mSize.  See calculation of pclSize.
            emitterTemplate.mDensity        = ambientFluidDensity ;
        #if ENABLE_FIRE
            emitterTemplate.mFuelFraction   = 0.01f ;   // Add a little bit of fluid. Note that the fuel-air ratio in a combustion engine is around 0.068.
            emitterTemplate.mFlameFraction  = 0.0f  ;   // No flames initially -- those will occur from ignition.
            emitterTemplate.mSmokeFraction  = 0.99f ;   // Mostly smoke.
        #endif
            emitterSpread.mPosition         = Vec3( 0.5f , 0.5f , 0.5f ) ;

            mVortonPclGrpInfo.mPclOpEmit->mTemplate                     = emitterTemplate ;
            mVortonPclGrpInfo.mPclOpEmit->mTemplate.mAngularVelocity    = FLT_EPSILON * Vec3( 1.0f , 1.0f , 1.0f ) ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread                       = emitterSpread ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread.mDensity              = 0.0f ;
            mVortonPclGrpInfo.mPclOpEmit->mEmitRate                     = 100.0f ;

            mTracerPclGrpInfo.mPclOpEmit->mEmitRate                     = 5000.0f ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate                     = mVortonPclGrpInfo.mPclOpEmit->mTemplate ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate.mSize               = 0.0095f ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread                       = mVortonPclGrpInfo.mPclOpEmit->mSpread ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.x           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.x ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.y           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.y ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.z           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.z ;

            mVortonPclGrpInfo.mPclOpWind->mWind                         = Vec3( 0.0f , 0.0f , 0.0f ) ;

        #if ENABLE_FIRE
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetCombustionTemperature( 1.0f * Particle_sAmbientTemperature + 1.5f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetCombustionRateFactor( 2.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSmokeTemperature( 2000.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSmokeRateFactor( 5.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSpecificFreeEnergy( 10000.0f ) ;
        #endif

            // Add rigid bodies.
            GetSpheres().PushBack( RbSphere( Vec3(  0.0f , 0.0f , -0.5f ) , Vec3( 0.0f , 0.0f , 0.0f ) , veryLargeMass , 0.2f ) ) ;
            GetSpheres().PushBack( RbSphere( Vec3(  0.2f , 0.0f ,  2.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 1.0f /* veryLargeMass */ , 0.5f ) ) ;
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f ,  1.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , veryLargeMass , Vec3( 1.0f , 0.5f , 0.1f ) ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.
            GetBoxes().Back().GetBody()->SetAngularVelocity( Vec3( 0.01f , 0.02f , 0.0f ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 1.5f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -3.0f , 1.5f ) ) ;
        }
        break ;

        case 17: // Candle.
        {
            // Set up lights.
            mLights[0].mPosition    = Vec3( 0.0f , 0.0f , 0.7f ) ;
            mLights[0].mColor       = 1.2f * Vec3( 1.0f , 0.5f , 0.3f ) ;
            mLights[0].mAttenuation = Vec3( 0.0f , 0.2f , 0.8f ) ;
            mLights[0].mType        = QdLight::LT_POINT ;
            // Make light fluctuate to imply it comes from candle flame.
            mLights[0].mAmplitudes [ 1 ] = 0.05f ;
            mLights[0].mFrequencies[ 1 ] = 1.0f ;
            mLights[0].mAmplitudes [ 2 ] = 0.02f ;
            mLights[0].mFrequencies[ 2 ] = 2.0f ;
            mLights[0].mAmplitudes [ 3 ] = 0.01f ;
            mLights[0].mFrequencies[ 3 ] = 5.0f ;

            mLights[1].mPosition    = Vec3( 0.0f , 0.0f , 1.0f ) ; // actually a direction
            mLights[1].mColor       = 0.2f * Vec3( 1.0f , 1.0f , 1.0f ) ;
            mLights[1].mType        = QdLight::LT_DIRECTIONAL ;

            mVortonPclGrpInfo.mPclOpKillAge->mAgeMax    = 400 ;
            mTracerPclGrpInfo.mPclOpKillAge->mAgeMax    = mVortonPclGrpInfo.mPclOpKillAge->mAgeMax ;

            Particle emitterTemplate , emitterSpread ;
            emitterTemplate.mPosition       = Vec3( 0.0f , 0.0f , 0.0f ) ;
            emitterTemplate.mVelocity       = Vec3( 0.0f , 0.0f , 0.0f ) ; // Mostly useless when Wind is active
            emitterTemplate.mSize           = 0.05f ;
            emitterTemplate.mDensity        = ambientFluidDensity ;
        #if ENABLE_FIRE
            emitterTemplate.mFuelFraction   = 0.01f ;   // Add a little bit of fluid. Note that the fuel-air ratio in a combustion engine is around 0.068.
            emitterTemplate.mFlameFraction  = 0.0f  ;   // No flames initially -- those will occur from ignition.
            emitterTemplate.mSmokeFraction  = 0.99f ;   // Mostly smoke.
        #endif
            emitterSpread.mPosition         = Vec3( 0.2f , 0.2f , 0.2f ) ;

            mVortonPclGrpInfo.mPclOpEmit->mTemplate                     = emitterTemplate ;
            mVortonPclGrpInfo.mPclOpEmit->mTemplate.mAngularVelocity    = FLT_EPSILON * Vec3( 1.0f , 1.0f , 1.0f ) ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread                       = emitterSpread ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread.mDensity              = 0.0f ;
            mVortonPclGrpInfo.mPclOpEmit->mEmitRate                     = 100.0f ;

            mTracerPclGrpInfo.mPclOpEmit->mEmitRate                     = 1000.0f ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate                     = mVortonPclGrpInfo.mPclOpEmit->mTemplate ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate.mSize               = 0.0085f ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread                       = mVortonPclGrpInfo.mPclOpEmit->mSpread ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.x           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.x ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.y           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.y ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.z           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.z ;

            mVortonPclGrpInfo.mPclOpWind->mWind                         = Vec3( 0.0f , 0.0f , 0.0f ) ;

        #if ENABLE_FIRE
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetCombustionTemperature( Particle_sAmbientTemperature + 1.3f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetCombustionRateFactor( 5.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSmokeTemperature( 2000.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSmokeRateFactor( 10.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSpecificFreeEnergy( 10000.0f ) ;
        #endif

            // Add a sphere
            GetSpheres().PushBack( RbSphere( Vec3(  0.0f , 0.0f , -0.35f ) , Vec3( 0.0f , 0.0f , 0.0f ) , veryLargeMass , 0.2f ) ) ;
            GetSpheres().PushBack( RbSphere( Vec3(  0.2f , 0.0f ,  5.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 1.0f /* veryLargeMass */ , 0.5f ) ) ;
            GetSpheres().PushBack( RbSphere( Vec3( -1.0f , 1.0f ,  1.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , veryLargeMass , 0.3f ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 1.5f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -3.0f , 1.5f ) ) ;
        }
        break ;

        case 18: // Large flame
        {
            // Set up lights.
            mLights[0].mPosition    = Vec3( 0.0f , 0.0f , 0.7f ) ;
            mLights[0].mColor       = 1.2f * Vec3( 1.0f , 0.5f , 0.3f ) ;
            mLights[0].mAttenuation = Vec3( 0.0f , 0.2f , 0.8f ) ;
            mLights[0].mType        = QdLight::LT_POINT ;
            // Make light fluctuate to imply it comes from flames.
            mLights[0].mAmplitudes [ 1 ] = 0.1f ;
            mLights[0].mFrequencies[ 1 ] = 2.0f ;
            mLights[0].mAmplitudes [ 2 ] = 0.05f ;
            mLights[0].mFrequencies[ 2 ] = 5.0f ;
            mLights[0].mAmplitudes [ 3 ] = 0.02f ;
            mLights[0].mFrequencies[ 3 ] = 10.0f ;

            mLights[1].mPosition    = Vec3( 0.0f , 0.0f , 1.0f ) ; // actually a direction
            mLights[1].mColor       = 0.2f * Vec3( 1.0f , 1.0f , 1.0f ) ;
            mLights[1].mType        = QdLight::LT_DIRECTIONAL ;

            mVortonPclGrpInfo.mPclOpKillAge->mAgeMax    = 300 ;
            mTracerPclGrpInfo.mPclOpKillAge->mAgeMax    = mVortonPclGrpInfo.mPclOpKillAge->mAgeMax ;

            Particle emitterTemplate , emitterSpread ;
            emitterTemplate.mPosition       = Vec3( 0.0f , 0.0f , 0.0f ) ;
            emitterTemplate.mVelocity       = Vec3( 0.0f , 0.0f , 0.0f ) ; // Mostly useless when Wind is active
            emitterTemplate.mSize           = 0.0625f ;   // Also see where Particles::Emit reassigns mSize.  See calculation of pclSize.
            emitterTemplate.mDensity        = ambientFluidDensity ;
        #if ENABLE_FIRE
            emitterTemplate.mFuelFraction   = 0.01f ;   // Add a little bit of fluid. Note that the fuel-air ratio in a combustion engine is around 0.068.
            emitterTemplate.mFlameFraction  = 0.0f  ;   // No flames initially -- those will occur from ignition.
            emitterTemplate.mSmokeFraction  = 0.99f ;   // Mostly smoke.
        #endif
            emitterSpread.mPosition         = Vec3( 1.0f , 1.0f , 0.5f ) ;

            mVortonPclGrpInfo.mPclOpEmit->mTemplate                     = emitterTemplate ;
            mVortonPclGrpInfo.mPclOpEmit->mTemplate.mAngularVelocity    = FLT_EPSILON * Vec3( 1.0f , 1.0f , 1.0f ) ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread                       = emitterSpread ;
            mVortonPclGrpInfo.mPclOpEmit->mSpread.mDensity              = 0.0f ;
            mVortonPclGrpInfo.mPclOpEmit->mEmitRate                     = 200.0f ;

            mTracerPclGrpInfo.mPclOpEmit->mEmitRate                     = 10000.0f ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate                     = mVortonPclGrpInfo.mPclOpEmit->mTemplate ;
            mTracerPclGrpInfo.mPclOpEmit->mTemplate.mSize               = 0.0105f ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread                       = mVortonPclGrpInfo.mPclOpEmit->mSpread ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.x           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.x ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.y           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.y ;
            mTracerPclGrpInfo.mPclOpEmit->mSpread.mPosition.z           = mVortonPclGrpInfo.mPclOpEmit->mSpread.mPosition.z ;

            mVortonPclGrpInfo.mPclOpWind->mWind                         = Vec3( 0.0f , 0.0f , 0.0f ) ;

        #if ENABLE_FIRE
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetCombustionTemperature( 1.0f * Particle_sAmbientTemperature + 1.5f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetCombustionRateFactor( 2.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSmokeTemperature( 5000.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSmokeRateFactor( 2.0f ) ;
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSpecificFreeEnergy( 20000.0f ) ;
        #endif

            // Add a sphere
            GetSpheres().PushBack( RbSphere( Vec3(  0.0f , 0.0f , -0.5f ) , Vec3( 0.0f , 0.0f , 0.0f ) , veryLargeMass , 0.2f ) ) ;
            GetSpheres().PushBack( RbSphere( Vec3(  0.2f , 0.0f ,  5.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 1.0f /* veryLargeMass */ , 0.5f ) ) ;
            GetSpheres().PushBack( RbSphere( Vec3( -1.0f , 1.0f ,  1.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , veryLargeMass , 0.3f ) ) ;

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 1.5f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -3.0f , 1.5f ) ) ;
        }
        break ;

        case 20: // Stationary cube to test RemoveEmbedded
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax / 2 , VortexNoise( Vec3( 0.5f * fThickness , 0.5f * fThickness , 0.5f * fThickness ) ) ) ;

            numTracersPerCellCubeRoot ++ ;

            // Add a cube
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.01f ,  0.0f ) , Vec3( 0.00f , 0.0f , 0.0f ) , veryLargeMass , Vec3( 0.5f , 0.5f , 0.5f ) ) ) ;
            {
            Mat33 orientation ;
            orientation.SetRotationX( 0.5f ) ;
            GetBoxes().Back().GetBody()->SetOrientation( orientation ) ; // Orient box
            }

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -3.0f , 0.0f ) ) ;
        break ;

        ////////////////////////////////////////////////////////////////////////

        case 21: // Gradient sphere, for testing SPH mass density and density gradient calculations.
            numTracersPerCellCubeRoot = 1 ;
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax * 64 , DensityGradientBall( fRadius , 1.0f ) ) ;
            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -3.0f , 0.0f ) ) ;
        break ;

        case 22: // Box projectile
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax / 2 , VortexNoise( Vec3( 4.0f * fThickness , 0.125f * fThickness , 0.5f * fThickness ) ) ) ;

            numTracersPerCellCubeRoot ++ ;

            // Add a box
        #if 0   // Cube.
            GetBoxes().PushBack( RbBox( Vec3( -3.0f , 0.01f ,  0.0f ) , Vec3( 4.0f , 0.0f , 0.0f ) , 0.01 , Vec3( 0.2f , 0.2f , 0.2f ) ) ) ;
            {
            Mat33 orientation ;
            orientation.SetRotationY( 0.5f ) ;
            GetBoxes().Back().GetBody()->SetOrientation( orientation ) ; // Orient box
            }
        #else   // Long thin board.
            GetBoxes().PushBack( RbBox( Vec3( -3.0f , 0.0f ,  -0.05f ) , Vec3( 10.0f , 0.0f , 0.0f ) , 0.02 , Vec3( 1.0f , 1.0f , 0.1f ) ) ) ;
            {
            Mat33 orientation ;
            orientation.SetRotationY( -0.2f ) ; // Give plate some angle of attack.
            GetBoxes().Back().GetBody()->SetOrientation( orientation ) ; // Orient box
            }
        #endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 1.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 1.0f , -3.0f , 0.0f ) ) ;
        break ;

        case 23: // Box projectile with spin about longitudinal axis
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax / 2 , VortexNoise( Vec3( 4.0f * fThickness , 0.3f * fThickness , 0.3f * fThickness ) ) ) ;

            numTracersPerCellCubeRoot -- ;

            // Add a box.
            GetBoxes().PushBack( RbBox( Vec3( -3.0f , 0.0f , 0.0f ) , Vec3( 4.0f , 0.0f , 0.0f ) , 0.02f , Vec3( 0.2f , 0.2f , 0.2f ) ) ) ;
            GetBoxes().Back().GetBody()->ApplyImpulsiveTorque( Vec3( 0.001f , 0.0f , 0.0f ) ) ; // Make box spin

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        break ;

        case 24: // Box projectile with spin about transverse axis
            AssignVortons( vortons , 0.125f * FLT_EPSILON , numVortonsMax , VortexNoise( Vec3( 4.0f * fThickness , 0.125f * fThickness , 0.5f * fThickness ) ) ) ;

            numTracersPerCellCubeRoot -- ;

            // Add a box.
            GetBoxes().PushBack( RbBox( Vec3( -3.0f , 0.0f , 0.0f ) , Vec3( 2.0f , 0.0f , 0.0f ) , 0.015f , Vec3( 0.2f , 0.5f , 0.1f ) ) ) ;
            GetBoxes().Back().GetBody()->ApplyImpulsiveTorque( Vec3( 0.0f , 0.002f , 0.0f ) ) ; // Make box spin

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 1.0f , 0.0f , 0.0f ) ) ;
            mCamera.SetEye( Vec3( 1.0f , -3.0f , 0.0f ) ) ;
        break ;

        case 25: // particle fountain
        {
            ParticleSystem * pclSys = CreateNonFluidParticleSystem( 0 ) ;

            mPclSysMgr.PushBack( pclSys ) ;

            // TODO: FIXME: Clone this effect multiple times using new virtual constructors.

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 2.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -10.0f , 2.0f ) ) ;
        }
        break ;

        case 26: // ball of initially stationary heavy fluid inside a box.
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBall( fRadius , 10.0f ) , Vec3( 0.0f , 0.0f , 1.0f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot += 2 ;

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 5.0f , 5.0f , 5.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -4.0f , 4.0f ) ) ;
        break ;

        case 27: // Two boxes of initially stationary fluid, one heavy, one light, light one above heavy one, inside container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) ,  0.6f ) , Vec3( 0.0f , 0.0f , -0.5f ) ) ;
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) , -0.6f ) , Vec3( 0.0f , 0.0f ,  0.5f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSpecificHeatCapacity( 1.0f ) ; // Diagnosing: Does this influence vorton motion?


            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        case 28: // One box of initially stationary fluid, heavy, at bottom of container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) ,  0.6f ) , Vec3( 0.0f , 0.0f , -0.5f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;

            // Decorations and other rendering options
            //mGridDecorations        = GRID_DECO_POINTS ;
            //mGridField              = GRID_FIELD_DENSITY_GRADIENT ;
            //mVortonRendering        = VORTON_RENDER_DIAGNOSTIC_PARTICLES ;
            //mTracerRendering        = TRACER_RENDER_DYE ;
            //mEmphasizeCameraTarget  = false ;
        }
        break ;

        case 29: // One box of initially stationary fluid, light, at top of container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) , -0.6f ) , Vec3( 0.0f , 0.0f ,  0.5f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        ////////////////////////////////////////////////////////////////////////

        case 32: // One rotating box of initially stationary fluid, heavy, at side of container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBox( Vec3( 1.5f , 0.5f , 2.0f ) ,  1000.0f ) , Vec3( 0.75f , 0.0f , 0.0f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 1.0f / FLT_EPSILON , Vec3( 3.0f , 0.5f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ;
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ;
            GetBoxes().Back().GetBody()->SetAngularVelocity( Vec3( 0.0f , 2.0f , 0.0f ) ) ;

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -3.0f , 0.0f ) ) ;
        }
        break ;

        case 33: // One box of initially stationary fluid, heavy, at side of container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBox( Vec3( 1.5f , 0.5f , 2.0f ) ,  100.0f ) , Vec3( 0.75f , 0.0f , 0.0f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 3.0f , 0.5f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -1.8f , 0.0f ) ) ;
        }
        break ;

        case 34: // One box of initially stationary fluid, heavy, at side of container.
        {
            AssignVortons( vortons , 1.0f , 2 * numVortonsMax , DensityBox( Vec3( 1.5f , 0.5f , 2.0f ) ,  100.0f ) , Vec3( 0.75f , 0.0f , 0.0f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 3.0f , 0.5f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        case 35: // Two boxes of initially stationary fluid, one heavy, one light, heavy above light, inside container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) , -0.6f ) , Vec3( 0.0f , 0.0f , -0.5f ) ) ;
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) ,  0.6f ) , Vec3( 0.0f , 0.0f ,  0.5f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetSpecificHeatCapacity( 1.0f ) ; // Diagnosing: Does this influence vorton motion?

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        case 36: // One box of initially stationary fluid, light, at bottom of container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( fRadius , fRadius , fRadius ) , -0.6f ) , Vec3( 0.0f , 0.0f , -0.5f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        case 37: // One box of initially stationary fluid, heavy, at top of container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBox( Vec3( fRadius , fRadius , fRadius ) ,  1000.0f ) , Vec3( 0.0f , 0.0f ,  0.5f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        case 38: // One box of initially stationary fluid, heavy, at side of container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax / 2 , DensityBox( Vec3( 0.5 , 1.0f , 2.0f ) ,  1000.0f ) , Vec3( 0.0f , 0.0f ,  0.0 ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        case 39: // One box of initially stationary fluid, heavy, at middle of container, nearly filling container.
        {
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBox( Vec3( 1.0f , 1.0f , 1.9f ) ,  1000.0f ) , Vec3( 0.0f , 0.0f ,  0.0f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

        case 40: // One box of initially stationary fluid, neutral, at middle of container, nearly filling container.
        {
            //AssignVortons( vortons , 1.0f , numVortonsMax , VortexNoise( Vec3( 1.0f , 1.0f , 1.9f ) ) , Vec3( 0.0f , 0.0f ,  0.0f ) ) ;
            AssignVortons( vortons , 1.0f , numVortonsMax , DensityBox( Vec3( 1.0f , 1.0f , 1.9f ) ,  1.0f ) , Vec3( 0.0f , 0.0f ,  0.0f ) ) ;

            vortonsImplyingTracers = reinterpret_cast< const VECTOR< Particle > * >( & vortons ) ;
            numTracersPerCellCubeRoot -- ;

            // Add a box-shaped hole.
            GetBoxes().PushBack( RbBox( Vec3( 0.0f , 0.0f , 0.0f ) , Vec3( 0.0f , 0.0f , 0.0f ) , 100000.0f , Vec3( 1.0f , 1.0f , 2.0f ) , true ) ) ;
            GetBoxes().Back().GetThermalProperties().SetOneOverHeatCapacity( 1.0f ) ; // Very small heat capacity.  Want this box to respond to heat quickly, so it does not alter temperature.
            GetBoxes().Back().GetThermalProperties().SetThermalConductivity( 0.0f ) ; // Prevent heat transfer between fluid and container.

#if REDUCE_CONVERGENCE
            mVortonPclGrpInfo.mPclOpFluidBodInte->mVortonIndicesGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortonIndicesGrid() ;
#endif

            // Position and orient camera.
            mCamera.SetTarget( Vec3( 0.0f ,  0.0f , 0.0f ) ) ;
            mCamera.SetEye   ( Vec3( 0.0f , -2.0f , 0.0f ) ) ;
        }
        break ;

    }

    // Apply same wind to tracers that vortons experience.
    * mTracerPclGrpInfo.mPclOpWind = * mVortonPclGrpInfo.mPclOpWind ;

    // Tell Fluid-Body simulation about rigid bodies.
    GetPhysicalObjects().Clear() ;
    for( size_t iSphere = 0 ; iSphere < GetSpheres().Size() ; ++ iSphere )
    {   // For each sphere in the simulation...
        RbSphere * pSphere = & GetSpheres()[ iSphere ] ;
        GetPhysicalObjects().PushBack( pSphere ) ;
    }
    for( size_t iBox = 0 ; iBox < GetBoxes().Size() ; ++ iBox )
    {   // For each box in the simulation...
        RbBox * pBox = & GetBoxes()[ iBox ] ;
        GetPhysicalObjects().PushBack( pBox ) ;
    }
    if( mInitialized )
    {   // Render API is initialized.  Initial conditions probably changed (i.e. this is not the first call to this function).
        // Can and should create entities now, to bind physical objects to their render objects.
        CreateEntities() ;
    }

#if COMPUTE_DENSITY_AND_GRADIENT_AT_AND_WITH_VORTONS_USING_SPH || POISON_DENSITY_GRADIENT_BASED_ON_VORTONS_HITTING_WALLS || POISON_DENSITY_BASED_ON_GRIDPOINTS_INSIDE_WALLS || POISON_DENSITY_GRADIENT_BASED_ON_GRIDPOINTS_INSIDE_WALLS || COMPUTE_PRESSURE_GRADIENT
    mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetPhysicalObjects( & GetPhysicalObjects() ) ;
#endif

    if( ! vortonSim.GetVortons()->empty() )
    {   // Initial conditions include pre-existing vortons (in contrast to an effect where all particles are emitted later).
    #if( ENABLE_FLUID_BODY_SIMULATION )
        FluidBodySim::RemoveEmbeddedParticles( reinterpret_cast< VECTOR< Particle > & > ( * vortonSim.GetVortons() ) , GetPhysicalObjects() ) ;
        FluidBodySim::RemoveEmbeddedParticles( mTracerPclGrpInfo.mParticleGroup->GetParticles() , GetPhysicalObjects() ) ;
    #endif

        vortonSim.Initialize() ;
        UniformGridGeometry tracerGrid( vortonSim.GetGrid() ) ;
        tracerGrid.Scale( vortonSim.GetGrid() , tracerGridScale ) ;

        PclOpEmit::Emit( mTracerPclGrpInfo.mParticleGroup->GetParticles() , tracerGrid , numTracersPerCellCubeRoot , vortonsImplyingTracers ) ;

    #if( ENABLE_FLUID_BODY_SIMULATION )
        FluidBodySim::RemoveEmbeddedParticles( reinterpret_cast< VECTOR< Particle > & >( * vortonSim.GetVortons() ) , GetPhysicalObjects() ) ;
        FluidBodySim::RemoveEmbeddedParticles( mTracerPclGrpInfo.mParticleGroup->GetParticles() , GetPhysicalObjects() ) ;
    #endif
    }

#if PROFILE
    fprintf( stderr , "Initial condition %i, simulation objects: %i tracers    %i vortons    %i spheres %i boxes\n"
        , ic , mTracerPclGrpInfo.mParticleGroup->GetNumParticles() , vortons.Size() , GetSpheres().Size() , GetBoxes().size() ) ;
#endif

}




/** Destruct application for interactive simulation and visualization.
*/
InteSiVis::~InteSiVis()
{
    sInstance = 0 ;
}




/** Function that GLUT calls when window resizes.
*/
/* static */ void InteSiVis::GlutReshapeGlutCallback(int width, int height)
{
    const int   width_4 = (width / 4) * 4;
    sInstance->mCamera.SetViewport( width_4 ,height ) ;
    glViewport( 0 , 0 , sInstance->mCamera.GetWidth() , sInstance->mCamera.GetHeight() ) ;
    glutReshapeWindow( width_4 , height ) ;
}




/// Set the projection matrix to orthographic, for rendering text.
void setOrthographicProjection()
{
    // switch to projection mode
    glMatrixMode(GL_PROJECTION);

    // save previous matrix which contains the
    //settings for the perspective projection
    glPushMatrix();

    // reset matrix
    glLoadIdentity();

    int viewportXYWH[4] ;
    glGetIntegerv( GL_VIEWPORT , viewportXYWH ) ;

    // set a 2D orthographic projection
    gluOrtho2D(0, viewportXYWH[ 2 ] , viewportXYWH[ 3 ] , 0);

    // switch back to modelview mode
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

    glLoadIdentity();
}




/// Set the projection matrix to projection, for rendering 3D objects.
void restorePerspectiveProjection()
{
    glMatrixMode(GL_PROJECTION);
    // restore previous projection matrix
    glPopMatrix();

    // get back to modelview mode
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix() ;
}




/// Render the given string using the given font.
static void oglRenderBitmapString( void * font , const char * string )
{
    for( const char * c = string ; *c != '\0' ; c ++ )
    {
        glutBitmapCharacter( font , *c ) ;
    }
}




static void oglRenderBitmapString( const Vec3 & pos , void * font , const Vec4 & color , const char * string , bool useScreenSpace )
{
    if( useScreenSpace )
    {
        setOrthographicProjection() ;
    }

    glPushAttrib( GL_DEPTH_BUFFER_BIT | GL_LIGHTING_BIT | GL_CURRENT_BIT ) ; // Should include GL_ENABLE_BIT or GL_ALL_ATTRIB_BITS but for some reason that *sometimes* makes things disappear, like vortex lines.

    glDisable( GL_LIGHTING ) ;
    glDisable( GL_TEXTURE_2D ) ;
    glEnable( GL_BLEND ) ;
    glBlendFunc( GL_SRC_ALPHA , GL_ONE_MINUS_SRC_ALPHA ) ;  // Typical alpha blending

    if( useScreenSpace )
    {   // Only when drawing in screen space...
        glDisable( GL_DEPTH_TEST ) ;    // Disable depth test so text appears as if in front of everything already drawn.
    }

    glColor4fv( reinterpret_cast< const GLfloat * >( & color ) ) ;

    //int viewportXYWH[4] ;
    //glGetIntegerv( GL_VIEWPORT , viewportXYWH ) ;

    glRasterPos3f( pos.x , pos.y , pos.z ) ;
    oglRenderBitmapString( font , string ) ;

    glPopAttrib() ;

    if( useScreenSpace )
    {
        restorePerspectiveProjection() ;
    }
}




/// Render the given string at the given screen-space position using the given font.
void oglRenderString( const Vec3 & pos , void * font , const char * format , ... )
{
    char stringBuffer[ 512 ] ;
    va_list args ;
    va_start( args , format ) ;
    _vsnprintf( stringBuffer , sizeof( stringBuffer ) , format , args ) ;
    va_end( args ) ;
    oglRenderBitmapString( pos , font , Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) , stringBuffer , true ) ;
}




/// Render the given string at the given world-space position using the given font.
void oglRenderStringWorld( const Vec3 & pos , void * font , const Vec4 & color , const char * format , ... )
{
    char stringBuffer[ 512 ] ;
    va_list args ;
    va_start( args , format ) ;
    vsprintf( stringBuffer , format , args ) ;
    va_end( args ) ;
    oglRenderBitmapString( pos , font , color , stringBuffer , false ) ;
}




Vec4 SetColorForGridPoint( const UniformGrid<Vec3> & grid , const unsigned ix , const unsigned iy , const unsigned iz , float magMin , float oneOverMagRange )
{
    const Vec3 & value  = grid.Get( ix , iy , iz ) ;
    const float valMag  = value.Magnitude() ;
    const float val0to1 = ( valMag - magMin ) * oneOverMagRange ;

    Vec4 color = GetColorFromRamp( sColorRampCyan , NUM_COLORS_IN_RAMP_CYAN , val0to1 ) ; // Color based on magnitude.
    color.w *= sInstance->CameraFocusEmphasis( grid.PositionFromIndices( ix , iy , iz ) ) ;  // Modulate opacity based on distance to camera target.
    glColor4f( color.x , color.y , color.z , color.w ) ;
    return color ;
}




void DrawGridCellEdge( const UniformGrid<Vec3> & grid , const unsigned ix0 , const unsigned iy0 , const unsigned iz0
                                                      , const unsigned ix1 , const unsigned iy1 , const unsigned iz1
                                                      , float magMin , float oneOverMagRange , InteSiVis::GridDecorationsE gridDecorations )
{
    glBegin( GL_LINES ) ;

    SetColorForGridPoint( grid , ix0 , iy0 , iz0 , magMin , oneOverMagRange ) ;
    const Vec3 gp0 = grid.PositionFromIndices( ix0 , iy0 , iz0 ) ;
    glVertex3f( gp0.x , gp0.y , gp0.z ) ;

    SetColorForGridPoint( grid , ix1 , iy1 , iz1 , magMin , oneOverMagRange ) ;
    const Vec3 gp1 = grid.PositionFromIndices( ix1 , iy1 , iz1 ) ;
    glVertex3f( gp1.x , gp1.y , gp1.z ) ;

    glEnd() ;
}




/** Draw grid cells.

    \param grid         Grid to draw.

    \param material     Material to use to draw grid.
*/
static void DrawGridCells( const UniformGrid<Vec3> & grid , const QdMaterial & material , InteSiVis::GridDecorationsE gridDecorations , bool bRenderDiagnosticText )
{

    // Draw bounding box
    material.UseMaterial() ;

    // Gather min and max statistics for grid contents (e.g. speed or density gradient)
    float magMin ;
    float magMax ;
    FindMagnitudeRange( grid , magMin , magMax ) ;
    magMin *= ( 1.0f - FLT_EPSILON ) ;  // Account for diff between fsqrtf and sqrtf
    magMax *= ( 1.0f + FLT_EPSILON ) ;  // Account for diff between fsqrtf and sqrtf
    const float oneOverMagRange = 1.0f / ( magMax - magMin + FLT_EPSILON ) ;    // Add FLT_EPSILON to avoid divide-by-zero when domain is empty.
    const bool speedRangeIsNonZero = magMax != magMin ;

    const Vec3  cellSize   = grid.GetCellSpacing() /* shrink cell to visualize each, otherwise edges overlap */ * 0.99f ;

    {
        // Draw each grid cell.
        Vec4            color   = material.GetColor() ;
        const unsigned  nx      = grid.GetNumCells( 0 ) ;
        const unsigned  ny      = grid.GetNumCells( 1 ) ;
        const unsigned  nz      = grid.GetNumCells( 2 ) ;
        unsigned        indices[ 4 ] ;
        unsigned &      ix      = indices[ 0 ] ;
        unsigned &      iy      = indices[ 1 ] ;
        unsigned &      iz      = indices[ 2 ] ;
        for( iz = 0 ; iz < nz ; ++ iz )
        {
            for( iy = 0 ; iy < ny ; ++ iy )
            {
                for( ix = 0 ; ix < nx ; ++ ix )
                {
                    if( speedRangeIsNonZero )
                    {
                        glLineWidth( 0.5f ) ;

                        DrawGridCellEdge( grid , ix   , iy   , iz   , ix+1 , iy   , iz   , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix   , iy+1 , iz   , ix+1 , iy+1 , iz   , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix   , iy   , iz   , ix   , iy+1 , iz   , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix+1 , iy   , iz   , ix+1 , iy+1 , iz   , magMin , oneOverMagRange , gridDecorations ) ;

                        DrawGridCellEdge( grid , ix   , iy   , iz+1 , ix+1 , iy   , iz+1 , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix   , iy+1 , iz+1 , ix+1 , iy+1 , iz+1 , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix   , iy   , iz+1 , ix   , iy+1 , iz+1 , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix+1 , iy   , iz+1 , ix+1 , iy+1 , iz+1 , magMin , oneOverMagRange , gridDecorations ) ;

                        DrawGridCellEdge( grid , ix   , iy   , iz   , ix   , iy   , iz+1 , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix+1 , iy   , iz   , ix+1 , iy   , iz+1 , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix   , iy+1 , iz   , ix   , iy+1 , iz+1 , magMin , oneOverMagRange , gridDecorations ) ;
                        DrawGridCellEdge( grid , ix+1 , iy+1 , iz   , ix+1 , iy+1 , iz+1 , magMin , oneOverMagRange , gridDecorations ) ;
                    }
                }
            }
        }
    }
}




/** Draw grid points.

    \param grid         Grid to draw.

    \param material     Material to use to draw grid.
*/
static void DrawGridPoints( const UniformGrid<Vec3> & grid , const QdMaterial & material , InteSiVis::GridDecorationsE gridDecorations , bool bRenderDiagnosticText )
{

    // Draw bounding box
    material.UseMaterial() ;

    // Gather min and max statistics for grid contents (speeds)
    float magMin ;
    float magMax ;
    FindMagnitudeRange( grid , magMin , magMax ) ;
    magMin *= ( 1.0f - FLT_EPSILON ) ;  // Account for diff between fsqrtf and sqrtf
    magMax *= ( 1.0f + FLT_EPSILON ) ;  // Account for diff between fsqrtf and sqrtf
    const float oneOverMagRange = 1.0f / ( magMax - magMin + FLT_EPSILON ) ;    // Add FLT_EPSILON to avoid divide-by-zero when domain is empty.

    const Vec3  cellSize   = grid.GetCellSpacing() /* shrink cell to visualize each, otherwise edges overlap */ * 0.99f ;
    const float cellLength = powf( grid.GetCellVolume() , 0.33333333333333333f ) ;

    const float lineWidth = InteSiVis::GRID_DECO_POINTS == gridDecorations ? 1.0f : 2.0f ; // thinner for just points, thicker for points and cells.
    const float pointSize = InteSiVis::GRID_DECO_POINTS == gridDecorations ? 2.0f : 4.0f ; // thinner for just points, thicker for points and cells.

    {
        // Draw each grid cell.
        const unsigned  nx      = grid.GetNumPoints( 0 ) ;
        const unsigned  ny      = grid.GetNumPoints( 1 ) ;
        const unsigned  nz      = grid.GetNumPoints( 2 ) ;
        const unsigned  npx     = grid.GetNumPoints( 0 ) ;
        const unsigned  npy     = grid.GetNumPoints( 1 ) ;
        const unsigned  npxy    = npx * npy ;
        unsigned        indices[ 4 ] ;
        unsigned &      ix      = indices[ 0 ] ;
        unsigned &      iy      = indices[ 1 ] ;
        unsigned &      iz      = indices[ 2 ] ;
        for( iz = 0 ; iz < nz ; ++ iz )
        {
            const unsigned zOffset = iz * npxy ;
            for( iy = 0 ; iy < ny ; ++ iy )
            {
                const unsigned yzOffset = zOffset + iy * npx ;
                for( ix = 0 ; ix < nx ; ++ ix )
                {
                    const unsigned xyzOffset = ix + yzOffset ;

                    {
                        
                        
                    }

                    const Vec3 & value  = grid[ xyzOffset ] ;
                    const float valMag  = value.Magnitude() ;
                    const float val0to1 = ( valMag - magMin ) * oneOverMagRange ;
                    const Vec3 gridpointPosition = grid.PositionFromIndices( indices ) ;

                    const Vec4 color = SetColorForGridPoint( grid , ix , iy , iz , magMin , oneOverMagRange ) ;

                    if( bRenderDiagnosticText && color.w > 0.01f )
                    {   // Diagnostic text is enabled and opacity is sufficient to render it.
                    #if 1
                        // Draw value magnitude
                        oglRenderStringWorld( gridpointPosition , GLUT_BITMAP_HELVETICA_10 , 2.0f * color , "%g" , value.Magnitude() ) ;
                    #else
                        // Draw position and value.
                        oglRenderStringWorld( gridpointPosition , GLUT_BITMAP_HELVETICA_10 , 2.0f * color , "%g,%g,%g %g,%g,%g"
                            , gridpointPosition.x , gridpointPosition.y , gridpointPosition.z
                            , value.x , value.y , value.z
                            ) ;
                    #endif
                    }
                    {
                        // Draw an arrow indicating value direction and magnitude.
                        // Scale arrow according to cell size and speed.
                        Vec3 arrowEnd( gridpointPosition + value.GetDir() * val0to1 * cellLength ) ;
                        glLineWidth( lineWidth ) ;
                        glBegin( GL_LINES ) ;
                        glVertex3fv( reinterpret_cast< const GLfloat * >( & gridpointPosition ) ) ;
                        glVertex3fv( reinterpret_cast< const GLfloat * >( & arrowEnd  ) ) ;
                        glEnd() ;
                    }
                    {
                        glPointSize( pointSize ) ;
                        glBegin( GL_POINTS ) ;
                        glVertex3fv( reinterpret_cast< const GLfloat * >( & gridpointPosition ) ) ;
                        glEnd() ;
                    }
                }
            }
        }
    }
}




static void DrawBoundingBox( const UniformGridGeometry & gridGeometry , const QdMaterial & material )
{
    if( gridGeometry.HasZeroExtent() )
    {   // Grid has no size.
        return ;    // There is nothing to draw.
    }

    // Draw bounding box
    material.UseMaterial() ;
    {
        glPushMatrix() ;

        const Vec3 vCenter = gridGeometry.GetCenter() ;
        glTranslatef( vCenter.x , vCenter.y , vCenter.z ) ;
        const Vec3 gridExtent = gridGeometry.GetExtent() * 1.01f ;
        glScalef( gridExtent.x , gridExtent.y , gridExtent.z ) ;

        glutWireCube( 1.0 ) ;

        glPopMatrix() ;
    }
}




static void DrawGrid( const UniformGrid<Vec3> & grid , const QdMaterial & material , InteSiVis::GridDecorationsE gridDecorations , bool bRenderDiagnosticText )
{
    if( grid.HasZeroExtent() )
    {   // Grid has no size.
        return ;    // There is nothing to draw.
    }

    // Draw bounding box
    DrawBoundingBox( grid , material ) ;

    if( grid.Empty() )
    {   // Grid has no data.
        return ;    // There is no data to draw.
    }

    material.UseMaterial() ;
    if(     ( InteSiVis::GRID_DECO_CELLS            == gridDecorations )
        ||  ( InteSiVis::GRID_DECO_CELLS_AND_POINTS == gridDecorations ) )
    {
        DrawGridCells( grid , material , gridDecorations , bRenderDiagnosticText ) ;
    }
    if(     ( InteSiVis::GRID_DECO_POINTS           == gridDecorations )
        ||  ( InteSiVis::GRID_DECO_CELLS_AND_POINTS == gridDecorations ) )
    {
        DrawGridPoints( grid , material , gridDecorations , bRenderDiagnosticText ) ;
    }
}




void ImgOpSeq_Noise( PeGaSys::Image & noise , float gamma , const Vec4 & tint )
{
    using namespace PeGaSys ;
    const unsigned  numSmoothingPasses  = noise.GetHeight() / 16 ;
    ImgOpMakeNoise::MakeNoise( noise ) ;
    ImgOpBlur::Blur( noise , numSmoothingPasses ) ;
    ImgOpApplyGamma::ApplyGamma( noise , gamma ) ;
    ImgOpTint::Tint( noise , tint ) ;
}




void ImgOpSeq_GradientNoise( PeGaSys::Image & noise , float gamma , const Vec4 & tint )
{
    using namespace PeGaSys ;
    const unsigned  numSmoothingPasses  = noise.GetHeight() / 16 ;
    ImgOpMakeNoise::MakeNoise( noise ) ;
    ImgOpBlur::Blur( noise , numSmoothingPasses ) ;
    ImgOpApplyGamma::ApplyGamma( noise , gamma ) ;
    ImgOpTint::Tint( noise , tint ) ;
    ImgOpLinearGradient::Gradient( noise , Vec2( 0.0f , 1.0f ) , Vec4( 0.1f , 0.1f , 0.1f , 1.0f ) , Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) ) ;
}




void ImgOpSeq_NoiseBall( PeGaSys::Image & noiseBall , float gamma , const Vec4 & tint )
{
    using namespace PeGaSys ;
    ImgOpMakeNoise::MakeNoise( noiseBall ) ;

    {   // Inscribe a ring inside the texture.
        // This simulates limn darkening.
        Image ring ;
        ring.CopyShape( noiseBall ) ;
        const float radius = float( ring.GetWidth() / 2 - 4 ) ;
        static const float thickness = 2.1f ;
        ImgOpDrawCircle::DrawCircle( ring , radius , thickness ) ;
        const Vec4 limnTint( tint.x * 0.0f , tint.y * 0.0f , tint.z * 0.0f , tint.w ) ;
        ImgOpTint::Tint( ring , limnTint ) ;
        ImgOpOverlay::Overlay( noiseBall , ring , noiseBall ) ;
    }

    const unsigned  numSmoothingPasses  = noiseBall.GetHeight() / 2 ;

    ImgOpBlur::Blur( noiseBall , numSmoothingPasses ) ;
    ImgOpRadialGradient::RadialGradient( noiseBall , /* power */ 1.0f , /* max */ 1.0f , /* alpha component index */ 3 ) ;
    ImgOpApplyGamma::ApplyGamma( noiseBall , gamma ) ;
    ImgOpTint::Tint( noiseBall , tint ) ;
}




void ImgOpSeq_DecoratedNoiseBall( PeGaSys::Image & noiseBall , float gamma , const Vec4 & color )
{
    using namespace PeGaSys ;
    ImgOpSeq_NoiseBall( noiseBall , gamma , color ) ;

    {   // Draw a faint frame around the texture border.
        Image frame ;
        frame.CopyShape( noiseBall ) ;
        ImgOpDrawBox::DrawBox( frame , 16 ) ;
        ImgOpTint::Tint( frame , Vec4( 1.0f , 1.0f , 1.0f , 0.25f ) ) ;
        ImgOpOverlay::Overlay( noiseBall , frame , noiseBall ) ;
    }

    {   // Inscribe a ring inside the texture.
        // This demarks the vorton boundary.
        Image ring ;
        ring.CopyShape( noiseBall ) ;
        ImgOpDrawCircle::DrawCircle( ring , /* radius */ float( ring.GetWidth() / 2 - 4 ) , /* thickness */ 2.1f ) ;
        ImgOpBlur::Blur( ring , /* num smoothing passes */ 2 ) ;
        ImgOpTint::Tint( ring , Vec4( 1.0f , 1.0f , 1.0f , 0.5f ) ) ;
        ImgOpOverlay::Overlay( noiseBall , ring , noiseBall ) ;
    }

    {   // Draw a faint dot at the center of the texture.
        Image dot ;
        dot.CopyShape( noiseBall ) ;
        ImgOpDrawCircle::DrawCircle( dot , /* radius */ 2 , /* thickness */ 2.1f ) ;
        ImgOpBlur::Blur( dot , /* num smoothing passes */ 2 ) ;
        ImgOpTint::Tint( dot , Vec4( 1.0f , 1.0f , 1.0f , 0.25f ) ) ;
        ImgOpOverlay::Overlay( noiseBall , dot , noiseBall ) ;
    }
}




/** Make a material with a simple noise texture.

    \param texWidth - see AllocateTexture

    \param texHeight - see AllocateTexture

    \param numTexPages - see AllocateTexture

    \param gamma - see AssignTexturePage

    \note This can only run after the device is initialized.
            Due to order-of-operations this cannot easily run
            in constructor.
*/
void MakeSimpleNoiseTexture( QdMaterial & material , int texWidth , int texHeight , int numTexPages , float gamma )
{
    const Vec4 blankColor( 1.0f , 1.0f , 1.0f , 1.0f ) ;

    PeGaSys::Image noise( texWidth , texHeight , 4 , numTexPages ) ;
    ImgOpSeq_Noise( noise , gamma , blankColor ) ;
    material.AssignTexture( noise ) ;
}




/** Make a material with a gradient noise texture.

    \param texWidth - see AllocateTexture

    \param texHeight - see AllocateTexture

    \param numTexPages - see AllocateTexture

    \param gamma - see AssignTexturePage

    \note This can only run after the device is initialized.
            Due to order-of-operations this cannot easily run
            in constructor.
*/
void MakeGradientNoiseTexture( QdMaterial & material , int texWidth , int texHeight , int numTexPages , float gamma )
{
    const Vec4 blankColor( 1.0f , 1.0f , 1.0f , 1.0f ) ;

    PeGaSys::Image noise( texWidth , texHeight , 4 , numTexPages ) ;
    ImgOpSeq_GradientNoise( noise , gamma , blankColor ) ;
    material.AssignTexture( noise ) ;
}




/** Create materials for rendering particles.
*/
void InteSiVis::CreateParticleMaterials()
{

    // Light & heavy diagnostic vorton materials.
    // This material is used to render vortons diagnostically, that is, for
    // diagnosing the algorithm.  This material is meant to reveal the vortons
    // and their state more clearly.
    {
        static const unsigned width         = 256 ;
        static const unsigned height        = 256 ;
        static const unsigned numChannels   = 4 ;
        static const unsigned numPages      = 2 ;

        const Vec4 translucentYellowish( 0.9f , 0.9f , 0.1f , 0.125f ) ;
        const Vec4 translucentCyanish  ( 0.1f , 0.9f , 0.9f , 0.25f ) ;

        PeGaSys::Image noiseBall( width , height , numChannels , numPages ) ;
        PeGaSys::Image noiseBallLight( noiseBall , 0 ) ;   // Page 0 of noiseBall
        PeGaSys::Image noiseBallHeavy( noiseBall , 1 ) ;   // Page 1 of noiseBall
        ImgOpSeq_DecoratedNoiseBall( noiseBallLight , 1.5f , translucentYellowish ) ;
        ImgOpSeq_DecoratedNoiseBall( noiseBallHeavy , 0.5f , translucentCyanish   ) ;
        mDiagnosticVortonMaterial.AssignTexture( noiseBall ) ;

        mDiagnosticVortonMaterial.SetColor( Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) );
        mDiagnosticVortonMaterial.SetDepthWrite( false ) ;
        mDiagnosticVortonMaterial.SetBlendMode( QdMaterial::BM_ALPHA ) ;
    }

    // Light & heavy simple vorton materials.
    // This material is used to render vortons for visual effects.  This
    // material lacks the obvious decorations, such as outlines and center dot.
    {
        static const unsigned width         = 256 ;
        static const unsigned height        = 256 ;
        static const unsigned numChannels   = 4 ;
        static const unsigned numPages      = 2 ;

        const Vec4 translucentYellowish( 0.9f , 0.9f , 0.1f , 0.125f ) ;
        const Vec4 translucentCyanish  ( 0.1f , 0.9f , 0.9f , 0.25f ) ;

        PeGaSys::Image noiseBall( width , height , numChannels , numPages ) ;
        PeGaSys::Image noiseBallLight( noiseBall , 0 ) ;   // Page 0 of noiseBall
        PeGaSys::Image noiseBallHeavy( noiseBall , 1 ) ;   // Page 1 of noiseBall
        ImgOpSeq_NoiseBall( noiseBallLight , 1.5f , translucentYellowish ) ;
        ImgOpSeq_NoiseBall( noiseBallHeavy , 0.5f , translucentCyanish   ) ;
        mSimpleVortonMaterial.AssignTexture( noiseBall ) ;

        mSimpleVortonMaterial.SetColor( Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) );
        mSimpleVortonMaterial.SetDepthWrite( false ) ;
        mSimpleVortonMaterial.SetBlendMode( QdMaterial::BM_ALPHA ) ;
    }

    // Light & heavy tracer dye materials
    {
        const Vec4 translucentRedish( 0.9f , 0.1f , 0.1f , 0.125f ) ; // redish
        const Vec4 translucentBluish( 0.1f , 0.1f , 0.9f , 0.125f ) ; // blueish

        PeGaSys::Image noiseBall( 128 , 128 , 4 , 2 ) ;
        PeGaSys::Image noiseBallLight( noiseBall , 0 ) ;   // Page 0 of noiseBall
        PeGaSys::Image noiseBallHeavy( noiseBall , 1 ) ;   // Page 1 of noiseBall
        ImgOpSeq_NoiseBall( noiseBallLight , 0.5f , translucentRedish ) ;
        ImgOpSeq_NoiseBall( noiseBallHeavy , 0.5f , translucentBluish   ) ;
        mDyeMaterial.AssignTexture( noiseBall ) ;

        mDyeMaterial.SetColor( Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) );
        mDyeMaterial.SetDepthWrite( false ) ;
        mDyeMaterial.SetBlendMode( QdMaterial::BM_ALPHA ) ;
        mDyeMaterial.SetScale( 2.0f ) ;
    #if ENABLE_FIRE
        mDyeMaterial.SetDensityVisibility( 255.0f * 1000.0f ) ;
    #else
        mDyeMaterial.SetDensityVisibility( FLT_MAX ) ;
    #endif
    }

    // Smoke material
    {
        const Vec4 translucentGray( 0.7f , 0.7f , 0.75f , 0.125f ) ; // light gray with a hint of blue.

        PeGaSys::Image noiseBall( 128 , 128 , 4 , 1 ) ;
        ImgOpSeq_NoiseBall( noiseBall , 0.5f , translucentGray ) ;
        mSmokeMaterial.AssignTexture( noiseBall ) ;

        mSmokeMaterial.SetColor( Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) );
        mSmokeMaterial.SetDepthWrite( false ) ;
        mSmokeMaterial.SetBlendMode( QdMaterial::BM_ALPHA ) ;
        mSmokeMaterial.SetScale( 4.0f ) ;
    #if ENABLE_FIRE
        mSmokeMaterial.SetDensityVisibility( 100.0f ) ;
    #else
        mSmokeMaterial.SetDensityVisibility( 0.0f ) ;
    #endif
    }

    // Flame material
    {
        const Vec4 orange( 1.0f , 0.5f , 0.1f , 1.0f ) ; // orange.

        PeGaSys::Image noiseBall( 128 , 128 , 4 , 1 ) ;
        ImgOpSeq_NoiseBall( noiseBall , 0.5f , orange ) ;
        mFlameMaterial.AssignTexture( noiseBall ) ;

        mFlameMaterial.SetColor( Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) );
        mFlameMaterial.SetDepthWrite( false ) ;
        mFlameMaterial.SetBlendMode( QdMaterial::BM_ADDITIVE ) ;
        mFlameMaterial.SetScale( 4.0f ) ;
    #if ENABLE_FIRE
        mFlameMaterial.SetDensityVisibility( 255.0f * 100.0f ) ;
    #else
        mFlameMaterial.SetDensityVisibility( 0.0f ) ;
    #endif
    }

    // Fuel material.
    {
        const Vec4 translucentBluish( 0.1f , 0.2f , 0.9f , 0.125f ) ; // blueish

        PeGaSys::Image noiseBall( 128 , 128 , 4 , 1 ) ;
        ImgOpSeq_NoiseBall( noiseBall , 0.5f , translucentBluish ) ;
        mFuelMaterial.AssignTexture( noiseBall ) ;

        mFuelMaterial.SetColor( Vec4( 1.0f , 1.0f , 1.0f , 1.0f ) );
        mFuelMaterial.SetDepthWrite( false ) ;
        mFuelMaterial.SetBlendMode( QdMaterial::BM_ALPHA ) ;
        mFuelMaterial.SetScale( 2.0f ) ;
    #if ENABLE_FIRE
        mFuelMaterial.SetDensityVisibility( 255.0f * 100.0f ) ;
    #else
        mFuelMaterial.SetDensityVisibility( FLT_MAX ) ;
    #endif
    }

}




/** Create rendering materials.
*/
void InteSiVis::CreateMaterials()
{

    CreateParticleMaterials() ;

    MakeSimpleNoiseTexture( mPhysObjMaterial , 256 , 256 , 1 , 0.5f ) ;
    mPhysObjMaterial.SetColor( Vec4( 0.8f , 0.4f , 0.8f , 1.0f ) ) ;

    MakeSimpleNoiseTexture( mContainerMaterial , 256 , 256 , 1 , 0.2f ) ;
    mContainerMaterial.SetColor( Vec4( 0.8f , 0.8f , 0.8f , 0.1f ) ) ;
    mContainerMaterial.SetCullFace( QdMaterial::CF_NONE ) ;
    mContainerMaterial.SetDepthWrite( false ) ;
    mContainerMaterial.SetBlendMode( QdMaterial::BM_ALPHA ) ;

    MakeSimpleNoiseTexture( mBoundingBoxMaterial , 8 , 8 , 1 , 0.0f ) ;
    mBoundingBoxMaterial.SetColor( Vec4( 0.7f , 0.7f , 0.7f , 0.1f ) ) ;
    mBoundingBoxMaterial.SetDepthWrite( false ) ;
    mBoundingBoxMaterial.SetBlendMode( QdMaterial::BM_ALPHA ) ;
    mBoundingBoxMaterial.SetCullFace( QdMaterial::CF_NONE ) ;

    MakeSimpleNoiseTexture( mPathlineMaterial , 8 , 8 , 1 , 0.0f ) ;
    mPathlineMaterial.SetColor( Vec4( 0.2f , 0.5f , 1.0f , 1.0f ) ) ;
    mPathlineMaterial.SetDepthWrite( false ) ;
    mPathlineMaterial.SetBlendMode( QdMaterial::BM_ALPHA ) ;

    MakeGradientNoiseTexture( mSkyMaterial , 512 , 512 , 1 , 0.2f ) ;
    mSkyMaterial.SetColor( Vec4( 0.3f , 0.3f , 0.3f , 1.0f ) ) ;

    mInitialized = true ;
}




/** Create simulation entities that couple physical objects to their render models.

    This should get called after both physical objects and render data exist.
    You cannot reliably call this, for example, from InitialConditions,
    since the render window might not have appeared yet, in which case
    various render resources will not yet exist.
*/
void InteSiVis::CreateEntities()
{

    const size_t numSpheres     = GetSpheres().size() ;
    const size_t numBoxes       = GetBoxes().size() ;
    const size_t numPhysObjs    = numSpheres + numBoxes ;

    mModels.clear() ;
    mEntities.clear() ;
    mModels.reserve( numPhysObjs ) ;
    mEntities.reserve( numPhysObjs ) ;

    size_t idxPhysObj = 0 ;

    for( size_t idxSphere = 0 ; idxSphere < numSpheres ; ++ idxSphere , ++ idxPhysObj )
    {   // For each sphere physical object...
        // Add a sphere model.
        const float radius = GetSpheres()[ idxSphere ].GetCollisionShape()->GetBoundingSphereRadius() ;
        GLuint sphereDisplayList = MakeSphere( radius ) ;
        mModels.push_back( QdModel( sphereDisplayList , & mPhysObjMaterial ) ) ;
        // Make an Entity to tie the render model to the physical object.
        mEntities.push_back( Entity( & GetSpheres()[ idxSphere ] , & mModels[ idxSphere ] ) ) ;
    }

    for( size_t idxBox = 0 ; idxBox < numBoxes ; ++ idxBox , ++ idxPhysObj )
    {   // For each box physical object...
        RbBox & box = GetBoxes()[ idxBox ] ;
        const float radius = box.GetCollisionShape()->GetBoundingSphereRadius() ;
        const Vec3 & dimensions = box.GetDimensions() ;
        QdMaterial * boxMaterial = & mPhysObjMaterial ;
        if( box.GetIsHole() )
        {   // This box is a container so render it translucent.
            boxMaterial = & mContainerMaterial ;
        }
        // Make a box.
        GLuint boxDisplayList = MakeBox( dimensions , radius ) ;
        // Add a cube model.
        mModels.push_back( QdModel( boxDisplayList , boxMaterial ) ) ;
        // Make an Entity to tie the render model to the physical object.
        mEntities.push_back( Entity( & box , & mModels[ idxPhysObj ] ) ) ;
    }
}




/** Perform initializations that cannot happen before render device is initialized.

    For example, the device cannot generate a texture until OpenGL has been initialized.
*/
void InteSiVis::InitializeRendering()
{

    CreateMaterials() ;

    mBallDisplayList = MakeSphere( 1.0f ) ;

    mInitialized = true ;
}




/** Render sky.
*/
void InteSiVis::RenderSky()
{
    QUERY_PERFORMANCE_ENTER ;
    mSkyMaterial.UseMaterial() ;
    QdLight::DisableLights() ;
    {
        glCullFace(GL_FRONT) ; // Render /inside/ of sphere
        glPushMatrix() ;
        static const float skyRadius = 500.0f ;
        const Vec3 & rEye    = mCamera.GetEye() ;
        glTranslatef( rEye.x , rEye.y , rEye.z ) ; // translate sky sphere to camera eye.
        glScalef( skyRadius , skyRadius , skyRadius ) ;
        glCallList( mBallDisplayList ) ;
        glPopMatrix() ;
    }
    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_Sky ) ;
}




/** Render rigid bodies.
*/
void InteSiVis::RenderRigidBodies()
{
    const size_t numEntities = mEntities.Size() ;
    for( size_t idxEntity = 0 ; idxEntity < numEntities ; ++ idxEntity )
    {
        Entity & entity = mEntities[ idxEntity ] ;
        entity.Update() ;
        entity.mRenderModel->Render( & mLights , mTimeNow ) ;
    }
}



void InteSiVis::SetDiagnosticPropertyValueRangeScale( float & rangeScale ) const
{
    VortonSim &                 rVortonSim  = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;

    float valMin = FLT_MAX , valMax = - FLT_MAX , valMean = 0.0f , valStdDev = 0.0f ;
    switch( mVortonProperty )
    {
    case VORTON_PROPERTY_POSITION:
        break ;

    case VORTON_PROPERTY_VELOCITY:
        {
            Vec3 centerOfVelocity ;
            Particles::ComputeVelocityStats( * reinterpret_cast< VECTOR< Particle > * >( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortons() ) , valMin , valMax , valMean , valStdDev , centerOfVelocity ) ;
        }
        break ;

    case VORTON_PROPERTY_VORTICITY:
        {
            Vec3 centerOfVorticity ;
            rVortonSim.GatherVorticityStats( valMin , valMax , valMean , valStdDev , centerOfVorticity ) ;
        }
        break ;

    case VORTON_PROPERTY_DENSITY:
        rVortonSim.GatherDensityStats( valMin , valMax , valMean , valStdDev ) ;
        break ;

    case VORTON_PROPERTY_DENSITY_SPH:
        rVortonSim.GatherDensitySphStats( valMin , valMax , valMean , valStdDev ) ;
        break ;

    case VORTON_PROPERTY_DENSITY_GRADIENT:
        rVortonSim.GatherDensityGradientStats( valMin , valMax , valMean , valStdDev ) ;
        break ;

    case VORTON_PROPERTY_PROXIMITY:
        rVortonSim.GatherProximityStats( valMin , valMax , valMean , valStdDev ) ;
        break ;
    }
    valMin *= ( 1.0f - FLT_EPSILON ) ;  // Account for diff between fsqrtf and sqrtf
    valMax *= ( 1.0f + FLT_EPSILON ) ;  // Account for diff between fsqrtf and sqrtf
    const float valMag          = Max2( fabsf( valMin ) , fabsf( valMax ) ) ;
    const float oneOverValMag   = 1.0f / Max2( valMag , FLT_EPSILON ) ;    // FLT_EPSILON to avoid divide-by-zero when range is empty.

    rangeScale = oneOverValMag ;
}




/** Render diagnostic text for vortex particles.
*/
void InteSiVis::RenderVortonDiagnosticText()
{
    QUERY_PERFORMANCE_ENTER ;


    // Enable opacity blending.
    glEnable( GL_BLEND ) ;
    glBlendFunc( GL_SRC_ALPHA , GL_ONE_MINUS_SRC_ALPHA ) ;  // Typical alpha blending

    VortonSim &                 rVortonSim  = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;
    VECTOR< Vorton > &          vortons     = * rVortonSim.GetVortons() ;
    const UniformGrid< Vec3 > & velGrid     = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid() ;

    float rangeScale ;
    SetDiagnosticPropertyValueRangeScale( rangeScale ) ;

    size_t numVortons = vortons.size() ;
    for( size_t iVort = 0 ; iVort < numVortons ; ++ iVort )
    {   // For each vorton...
        const Vorton & vort = vortons[ iVort ] ;

        if( velGrid.Encompasses( vort.mPosition ) )
        {   // Vorton is still inside grid.

            // Note that since this gets rendered after vorton advected, the
            // vorton is no longer necessarily near the gridpoint it was nearest
            // pre-advect. That means grid-derived diagnostic information is
            // suspect and has a good chance of being wrong fairly often.

            void * font = GLUT_BITMAP_HELVETICA_10 ;
            Vec4   color( 0.2f , 1.0f , 0.2f , 0.25f ) ;

            // Make opacity depend on distance to gridpoint.
            //color.w = 0.25f * Min2( vort.GetRadius() / distToNearestGridPoint , 1.0f ) ;

            unsigned    nearestGridPointIndices[ 4 ]    ; // indices of grid cell containing vorton.
            velGrid.IndicesOfNearestGridPoint( nearestGridPointIndices , vort.mPosition ) ; // Get indices of grid cell containing vorton.
            const Vec3  nearestGridPointPosition        = velGrid.PositionFromIndices( nearestGridPointIndices ) ;
            const Vec3  displacementToNearestGridPoint  = nearestGridPointPosition - vort.mPosition ;
            const float distToNearestGridPoint          = displacementToNearestGridPoint.Magnitude() ;

            if( ( distToNearestGridPoint < vort.GetRadius() ) && ( VORTON_PROPERTY_POSITION == mVortonProperty ) )
            {   // Gridpoint is inside vorton, and property being rendered is position.
                font  = GLUT_BITMAP_HELVETICA_12 ;
                color = Vec4( 1.0f , 0.2f , 0.2f , 0.5f ) ;
            }

            // Render diagnostic information about the vorton.
            char vortonPropertyText[ 128 ] = {} ;
            float val0to1 = 0.0f ;
            switch( mVortonProperty )
            {
            case VORTON_PROPERTY_POSITION:
                val0to1 = 1.0f ;
                sprintf( vortonPropertyText , "p=%g,%g,%g D=%g,%g,%s"
                    , vort.mPosition.x , vort.mPosition.y , vort.mPosition.z
                    , distToNearestGridPoint , distToNearestGridPoint / vort.GetRadius() , distToNearestGridPoint < vort.GetRadius() ? "IN" : "out" ) ;
                break ;

            case VORTON_PROPERTY_VELOCITY:
                val0to1 = vort.mVelocity.Magnitude() * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                sprintf( vortonPropertyText , "v=<%g,%g,%g> ||=%g"
                    , vort.mVelocity.x , vort.mVelocity.y , vort.mVelocity.z
                    , vort.mVelocity.Magnitude() ) ;
                break ;

            case VORTON_PROPERTY_VORTICITY:
                val0to1 = vort.GetVorticity().Magnitude() * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                sprintf( vortonPropertyText , "w=<%g,%g,%g> ||=%g"
                    , vort.GetVorticity().x , vort.GetVorticity().y , vort.GetVorticity().z
                    , vort.GetVorticity().Magnitude() ) ;
                break ;

            case VORTON_PROPERTY_DENSITY:
                val0to1 = vort.GetDensity() * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                sprintf( vortonPropertyText , "d=%g" , vort.GetDensity() ) ;
                break ;

            case VORTON_PROPERTY_DENSITY_SPH:
                {
                    const float & densitySph = rVortonSim.GetFluidDensitiesAtPcls()[ iVort ].mMassDensity ;
                    val0to1 = densitySph * rangeScale ;
                    color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                    sprintf( vortonPropertyText , "d=%g" , densitySph ) ;
                }
                break ;

            case VORTON_PROPERTY_DENSITY_GRADIENT:
            #if COMPUTE_DENSITY_AND_GRADIENT_AT_AND_WITH_VORTONS_USING_SPH || USE_SMOOTHED_PARTICLE_HYDRODYNAMICS
                if( ! rVortonSim.GetDensityGradientsAtPcls().Empty() )
                {
                    const Vec3 & densGrad = rVortonSim.GetDensityGradientsAtPcls()[ iVort ] ;
                    val0to1 = densGrad.Magnitude() * rangeScale ;
                    color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                    sprintf( vortonPropertyText , "dens grad=<%g,%g,%g> ||=%g"
                        , densGrad.x , densGrad.y , densGrad.z
                        , densGrad.Magnitude() ) ;
                }
            #endif
                break ;

            case VORTON_PROPERTY_PROXIMITY:
                {
                    const float & proximity = rVortonSim.GetProximities()[ iVort ] ;
                    const float val0to1     = proximity * rangeScale ;
                    color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                    sprintf( vortonPropertyText , "Prox=%g" , proximity ) ;
                }
                break ;

            }

            // Modulate opacity based on distance to camera target.
            color.w *= CameraFocusEmphasis( vort.mPosition ) ;

            if( color.w > 0.01f )
            {   // Diagnostic information is opaque enough to render.
                oglRenderStringWorld( vort.mPosition , font , color , "%i %s"
                    , iVort
                    , vortonPropertyText
                    ) ;
            }
        }
    }

    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderVortonDiagnosticText ) ;
}




/** Render diagnostic vectors for vortex particles.
*/
void InteSiVis::RenderVortonDiagnosticVectors()
{
    QUERY_PERFORMANCE_ENTER ;


    mPathlineMaterial.UseMaterial() ;

    VortonSim &                 rVortonSim  = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;
    VECTOR< Vorton > &          vortons     = * rVortonSim.GetVortons() ;
    const UniformGrid< Vec3 > & velGrid     = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid() ;

    float rangeScale ;
    SetDiagnosticPropertyValueRangeScale( rangeScale ) ;

    size_t numVortons = vortons.size() ;
    for( size_t iVort = 0 ; iVort < numVortons ; ++ iVort )
    {   // For each vorton...
        const Vorton & vort = vortons[ iVort ] ;

        Vec4   color( 1.0f , 1.0f , 1.0f , 0.5f ) ;

        // Make opacity depend on distance to gridpoint.
        //color.w = 0.25f * Min2( vort.GetRadius() / distToNearestGridPoint , 1.0f ) ;

        // Modulate opacity based on distance to camera target.
        float emphasis = CameraFocusEmphasis( vort.mPosition ) ;

        if( emphasis > 0.01f )
        {   // Diagnostic information is opaque enough to render.

            if( ( mGridDecorations == GRID_DECO_CELLS ) || ( mGridDecorations == GRID_DECO_CELLS_AND_POINTS ) )
            {   // Grid cell rendering is enabled.
                if( velGrid.Encompasses( vort.mPosition ) && ( VORTON_PROPERTY_POSITION == mVortonProperty ) )
                {   // Vorton is still inside grid.
                    // Note that since this gets rendered after vorton advected, the
                    // vorton is no longer necessarily near the gridpoint it was nearest
                    // pre-advect. That means some of this diagnostic information is
                    // suspect and has a good chance of being wrong fairly often.

                    unsigned    nearestGridPointIndices[ 4 ]    ; // indices of grid cell containing vorton.
                    velGrid.IndicesOfNearestGridPoint( nearestGridPointIndices , vort.mPosition ) ; // Get indices of grid cell containing vorton.
                    const Vec3  nearestGridPointPosition        = velGrid.PositionFromIndices( nearestGridPointIndices ) ;
                    // Draw a line from the vorton to the nearest gridpoint.
                    glBegin( GL_LINES ) ;
                    glColor4f( 1.0f , 0.0f , 1.0f , emphasis * color.w ) ; // Magenta
                    glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition           ) ) ;
                    glVertex3fv( reinterpret_cast< const GLfloat * >( & nearestGridPointPosition ) ) ;
                    glEnd() ;
                    glPointSize( 3.0f ) ;
                    glBegin( GL_POINTS ) ;
                    glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                    glEnd() ;
                }
            }

            if( mVortonProperty == VORTON_PROPERTY_VELOCITY )
            {
                const float val0to1 = vort.mVelocity.Magnitude() * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                glColor4f( color.x , color.y , color.z , emphasis * color.w ) ;
                // Draw a line indicating velocity direction.
                Vec3 velDirEnd = vort.mPosition + vort.mVelocity * vort.GetRadius() * rangeScale ;
                glBegin( GL_LINES ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & velDirEnd       ) ) ;
                glEnd() ;
                glPointSize( 3.0f ) ;
                glBegin( GL_POINTS ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glEnd() ;
            }

            if( VORTON_PROPERTY_VORTICITY == mVortonProperty )
            {
                const float val0to1 = vort.GetVorticity().Magnitude() * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                glColor4f( color.x , color.y , color.z , emphasis * color.w ) ;
                // Draw a line indicating vorticity.
                Vec3 vortDirEnd = vort.mPosition + vort.GetVorticity() * vort.GetRadius() * rangeScale ;
                glBegin( GL_LINES ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vortDirEnd      ) ) ;
                glEnd() ;
                glPointSize( 3.0f ) ;
                glBegin( GL_POINTS ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glEnd() ;
            }

            if( mVortonProperty == VORTON_PROPERTY_DENSITY )
            {
                const float val0to1 = vort.GetDensity() * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                glColor4f( color.x , color.y , color.z , emphasis * color.w ) ;

                // Draw a point indicating vorton position.
                glPointSize( 3.0f ) ;
                glBegin( GL_POINTS ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glEnd() ;
            }

            if( ( mVortonProperty == VORTON_PROPERTY_DENSITY_SPH ) && ! rVortonSim.GetFluidDensitiesAtPcls().Empty() )
            {
                const float & densitySph = rVortonSim.GetFluidDensitiesAtPcls()[ iVort ].mMassDensity ;
                const float val0to1 = densitySph * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                glColor4f( color.x , color.y , color.z , emphasis * color.w ) ;

                // Draw a point indicating vorton position.
                glPointSize( 3.0f ) ;
                glBegin( GL_POINTS ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glEnd() ;
            }

#if COMPUTE_DENSITY_AND_GRADIENT_AT_AND_WITH_VORTONS_USING_SPH || USE_SMOOTHED_PARTICLE_HYDRODYNAMICS
            if( ( VORTON_PROPERTY_DENSITY_GRADIENT == mVortonProperty ) && ! rVortonSim.GetDensityGradientsAtPcls().Empty()  )
            {
                const Vec3 & densGrad = rVortonSim.GetDensityGradientsAtPcls()[ iVort ] ;
                const float val0to1 = densGrad.Magnitude() * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                glColor4f( color.x , color.y , color.z , emphasis * color.w ) ;
                // Draw a line indicating density gradient.
                Vec3 densGradDirEnd = vort.mPosition + densGrad * vort.GetRadius() * rangeScale ;
                glBegin( GL_LINES ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & densGradDirEnd  ) ) ;
                glEnd() ;
                glPointSize( 3.0f ) ;
                glBegin( GL_POINTS ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glEnd() ;
            }
#endif

            if( ( mVortonProperty == VORTON_PROPERTY_PROXIMITY ) && ! rVortonSim.GetProximities().Empty() )
            {
                const float & proximity = rVortonSim.GetProximities()[ iVort ] ;
                const float val0to1     = proximity * rangeScale ;
                color = GetColorFromRamp( sColorRampRainbow , NUM_COLORS_IN_RAMP_RAINBOW , val0to1 ) ;
                glColor4f( color.x , color.y , color.z , emphasis * color.w ) ;

                // Draw a point indicating vorton position.
                glPointSize( 3.0f ) ;
                glBegin( GL_POINTS ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glEnd() ;
            }

#if POISON_DENSITY_GRADIENT_BASED_ON_VORTONS_HITTING_WALLS
            if( vort.mHitBoundary )
            {   // Vorton hit boundary. Draw a line indicating hit state.
                glPushAttrib( GL_ENABLE_BIT | GL_LINE_BIT ) ;
                glEnable( GL_LINE_SMOOTH ) ;
                //glLineWidth( 3.0f ) ;
                glBegin( GL_LINES ) ;
                glColor4f( 0.0f , 0.0f , 0.0f , emphasis * color.w ) ;
                Vec3 velHitEnd = vort.mPosition + Vec3( 1.0f , 0.0f , 0.0f ) * vort.GetRadius() * 0.5f;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & velHitEnd       ) ) ;
                velHitEnd = vort.mPosition + Vec3( 0.0f , 1.0f , 0.0f ) * vort.GetRadius() * 0.5f;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & velHitEnd       ) ) ;
                velHitEnd = vort.mPosition + Vec3( 0.0f , 0.0f , 1.0f ) * vort.GetRadius() * 0.5f;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & vort.mPosition  ) ) ;
                glVertex3fv( reinterpret_cast< const GLfloat * >( & velHitEnd       ) ) ;
                glEnd() ;
                glPopAttrib() ;
            }
#endif
        }
    }

    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderVortonDiagnosticVectors ) ;
}




/** Render vortex particles.
*/
void InteSiVis::RenderVortons()
{

    QUERY_PERFORMANCE_ENTER ;

    VortonSim &         rVortonSim  = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;
    VECTOR< Vorton > &  vortons     = * rVortonSim.GetVortons() ;

    if(     ( VORTON_RENDER_DIAGNOSTIC_PARTICLES  == mVortonRendering )
        ||  ( VORTON_RENDER_PARTICLES_AND_VECTORS == mVortonRendering )
        ||  ( VORTON_RENDER_PATHLINES             == mVortonRendering )
        ||  ( VORTON_RENDER_ALL                   == mVortonRendering ) )
    {   // Render vortons as diagnostic particles.
        //  // Modern STL supports vector.data() but not older versions like MSVS7.
        mDiagnosticVortonRenderer.SetParticleData( reinterpret_cast< char * >( & vortons.operator[]( 0 ) ) ) ; // Modern STL supports vector.data() but not older versions like MSVS7.
        mDiagnosticVortonRenderer.Render( mTimeNow , mTimeStep , vortons.Size() ) ;
        mSimpleVortonRenderer.SetParticleData( 0 ) ;
    }
    else if( VORTON_RENDER_SIMPLE_PARTICLES == mVortonRendering )
    {   // Render vortons as simple particles.
        mSimpleVortonRenderer.SetParticleData( reinterpret_cast< char * >( & vortons.operator[]( 0 ) ) ) ; // Modern STL supports vector.data() but not older versions like MSVS7.
        mSimpleVortonRenderer.Render( mTimeNow , mTimeStep , vortons.Size() ) ;
        mDiagnosticVortonRenderer.SetParticleData( 0 ) ;
    }
    else
    {   // Do not render vortons as particles.
        mDiagnosticVortonRenderer.SetParticleData( 0 ) ;
        mSimpleVortonRenderer.SetParticleData( 0 ) ;
    }

    if( mDiagnosticText == DIAG_TEXT_FULL )
    {
        RenderVortonDiagnosticText() ;
    }

    if(     ( mVortonRendering == VORTON_RENDER_VECTORS                 )
        ||  ( mVortonRendering == VORTON_RENDER_PARTICLES_AND_VECTORS   )
        ||  ( mVortonRendering == VORTON_RENDER_ALL                     ) )
    {
        RenderVortonDiagnosticVectors() ;
    }

    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderVortons ) ;
}




/** Render tracer particles.
*/
void InteSiVis::RenderTracers()
{
#if ENABLE_FIRE

    VECTOR< Particle > & tracers = mTracerPclGrpInfo.mParticleGroup->GetParticles() ;

    if( TRACER_RENDER_DYE == mTracerRendering )
    {
        // Render tracers as dye:
        QUERY_PERFORMANCE_ENTER ;
        mDyeRenderer.SetParticleData( (char*) & tracers[0] ) ;
        mDyeRenderer.Render( mTimeNow , mTimeStep , tracers.Size() ) ;
        QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderTracers ) ;
    }

    if( ( TRACER_RENDER_FIRE == mTracerRendering ) || ( TRACER_RENDER_FUEL == mTracerRendering ) )
    {
        // Render tracers as fuel:
        QUERY_PERFORMANCE_ENTER ;
        mFuelRenderer.SetParticleData( (char*) & tracers[0] ) ;
        mFuelRenderer.Render( mTimeNow , mTimeStep , tracers.Size() ) ;
        QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderTracers ) ;
    }

    if( ( TRACER_RENDER_FIRE == mTracerRendering ) || ( TRACER_RENDER_SMOKE == mTracerRendering ) )
    {
        // Render tracers as smoke:
        QUERY_PERFORMANCE_ENTER ;
        mSmokeRenderer.SetParticleData( (char*) & tracers[0] ) ;
        mSmokeRenderer.Render( mTimeNow , mTimeStep , tracers.Size() ) ;
        QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderTracers ) ;
    }

    if( ( TRACER_RENDER_FIRE == mTracerRendering ) || ( TRACER_RENDER_FLAME == mTracerRendering ) )
    {
        // Render tracers as flame:
        QUERY_PERFORMANCE_ENTER ;
        mFlameRenderer.SetParticleData( (char*) & tracers[0] ) ;
        mFlameRenderer.Render( mTimeNow , mTimeStep , tracers.Size() ) ;
        QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderTracers ) ;
    }

#else
    // Render tracers with texture that varies according to buoyancy:
    QUERY_PERFORMANCE_ENTER ;
    mDyeRenderer.SetParticleData( (char*) & tracer[0] ) ;
    mDyeRenderer.Render( mTimeNow , mTimeStep , tracer.Size() ) ;
    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderTracers ) ;
#endif
}




#if ENABLE_PARTICLE_POSITION_HISTORY
/** Render particle pathlines.
*/
void InteSiVis::RenderPathlines( VECTOR< Particle > & particles )
{
    QUERY_PERFORMANCE_ENTER ;

    mPathlineMaterial.UseMaterial() ;

    //  // Modern STL supports vector.data() but not older versions like MSVS7.

    size_t numParticles = particles.size() ;
    for( size_t iVort = 0 ; iVort < numParticles ; ++ iVort )
    {   // For each particle...
        const Particle & pcl = particles[ iVort ] ;

        Vec4   color( mPathlineMaterial.GetColor() ) ;

        // Modulate opacity based on distance to camera target.
        color.w *= CameraFocusEmphasis( pcl.mPosition ) ;

        // Render pathline for current particle.
        glBegin( GL_LINE_STRIP ) ;

        static const float oneOverNumHistoryPositions = 1.0f / float( Particle::NUM_HISTORICAL_POSITIONS ) ;
        size_t segmentIndex = 0 ;
        const size_t iHistoryEnd = pcl.HistoryEnd() ;
        for( size_t iHistory = pcl.HistoryBegin() ; iHistory != iHistoryEnd ; iHistory = Particle::NextHistoryIndex( iHistory ) )
        {   // For each element in the particle position history...
            // Fade each segment: more recent is more opaque.
            const float fade = float( segmentIndex ) * oneOverNumHistoryPositions ;
            glColor4f( color.x , color.y , color.z , color.w * fade ) ;
            if( pcl.mPositionHistory[ iHistory ].x != Particle::sNaN )
            {   // This history element is valid.
                // (Invalid element could come from particles too young to have full history.)
                glVertex3fv( reinterpret_cast< const GLfloat * >( & pcl.mPositionHistory[ iHistory ] ) ) ;
            }
            ++ segmentIndex ;
        }
        glVertex3fv( reinterpret_cast< const GLfloat * >( & pcl.mPositionHistory[ iHistoryEnd ] ) ) ;
        glEnd() ;
    }

    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderPathlines ) ;
}
#endif




/** Render NonFluid particles.
*/
void InteSiVis::RenderNonFluidParticles()
{
    QUERY_PERFORMANCE_ENTER ;

    for( ParticleSystemManager::Iterator ips = mPclSysMgr.Begin() ; ips != mPclSysMgr.End() ; ++ ips )
    {
        ParticleSystem * pclSys = * ips ;
        for( ParticleSystem::Iterator ipg = pclSys->Begin() ; ipg != pclSys->End() ; ++ ipg )
        {
            ParticleGroup * pclGrp = * ipg ;
            VECTOR< Particle > & pcls = pclGrp->GetParticles() ;

            mSmokeRenderer.SetParticleData( reinterpret_cast< char * >( & pcls.operator[]( 0 ) ) ) ; // Modern STL supports vector.data() but not older versions like MSVS7.
            mSmokeRenderer.Render( mTimeNow , mTimeStep , pcls.Size() ) ;
        }
    }

    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_RenderNonFluidParticles ) ;
}




/** Render particles.
*/
void InteSiVis::RenderParticles()
{
#if ! USE_ORIENTED_SORTED_PARTICLES
    QUERY_PERFORMANCE_ENTER ;
//    switch( mScenario )
//    {
//        case 0: case 1:
//            const Vec3      vCoM        = Particles::ComputeGeometricCenter( mTracers ) ;
//            const Vec3 &    vMin        = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid().GetMinCorner() ;
//            const Vec3 &    vExtent     = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid().GetExtent() ;
//            const Vec3      vMax        = vMin + vExtent ;
//            const Vec3      vFireball   = Vec3( 0.5f * ( vCoM.x + vMax.x ) , vCoM.y , vCoM.z ) ;
//            SetGlow( vFireball ) ;
//        break ;
//    }
    QdLight::DisableLights() ;
    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_DisableLightingForParticles ) ;
#endif

    if( mGridDecorations != GRID_DECO_NONE )
    {   // Grid decorations are enabled.
        QUERY_PERFORMANCE_ENTER ;
        const bool bRenderDiagnosticText = DIAG_TEXT_FULL  == mDiagnosticText  ;
        // Render bounding box -- intentionally before all opaque objects, with depth write disabled, so grid is behind everything else.
        if( GRID_FIELD_VELOCITY == mGridField )
        {
            DrawGrid( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid() , mBoundingBoxMaterial , mGridDecorations , bRenderDiagnosticText ) ;
        }
        else if( GRID_FIELD_DENSITY_GRADIENT == mGridField )
        {
            DrawGrid( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetDensityGradientGrid() , mBoundingBoxMaterial , mGridDecorations , bRenderDiagnosticText ) ;
        }
#if COMPUTE_PRESSURE_GRADIENT
        else if( GRID_FIELD_PRESSURE_GRADIENT == mGridField )
        {
            DrawGrid( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetPressureGradientGrid() , mBoundingBoxMaterial , mGridDecorations , bRenderDiagnosticText ) ;
        }
#endif
        QUERY_PERFORMANCE_EXIT( InteSiVis_Render_BoundingBox ) ;
    }

    // TODO: FIXME: Make particle rendering its own PclOp
    // HACK CODE START: Particle rendering should be another ParticleOperation but it is not.
    //                  Meanwhile, rendering non-fluid and fluid particles is essentially mutually exclusive.
    //                  One way to sort of repair this would be to make tracer and non-fluid rendering identical.
    //                  Even better, however, would be to make particle rendering a PclOp.
    if( ! mTracerPclGrpInfo.mParticleGroup->HasParticles() )
    {   // There is no fluid particle system.
        RenderNonFluidParticles() ;
    }
    else
    {   // There is a fluid particle system.
        if( mVortonRendering != VORTON_RENDER_NONE )
        {   // Vorton rendering is enabled.
            RenderVortons() ;
        }

        if(     ( mVortonRendering == VORTON_RENDER_PATHLINES )
            ||  ( mVortonRendering == VORTON_RENDER_ALL       )  )
        {   // Vorton Pathline rendering is enabled.
        #if ENABLE_PARTICLE_POSITION_HISTORY
            RenderPathlines( reinterpret_cast< VECTOR< Particle > & >( * mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortons() ) ) ;
        #endif
        }

        if( mTracerRendering != TRACER_RENDER_NONE )
        {   // Tracer rendering is enabled.
            RenderTracers() ;
        }
    }
}



static void RenderDiagnosticIntegralsText( float yPos , const char * name , const VortonSim::Integrals & integralsBefore , const VortonSim::Integrals & integralsAfter )
{
    const float delta = integralsBefore.ComputeMaxRelativeDifference( integralsAfter ) ;
    oglRenderString( Vec3( 10.0f , yPos , 0.0f ) , GLUT_BITMAP_8_BY_13 , "%s %6.1f I0={% 8.1f,% 8.1f,% 8.1f} I1={% 8.1f,% 8.1f,% 8.1f},{% 8.1f,% 8.1f,% 8.1f} I2={% 8.1f,% 8.1f,% 8.1f}"
        , name
        , delta
        , integralsAfter.mTotalCirculation.x            , integralsAfter.mTotalCirculation.y            , integralsAfter.mTotalCirculation.z
        , integralsAfter.mLinearImpulseFromVorticity.x  , integralsAfter.mLinearImpulseFromVorticity.y  , integralsAfter.mLinearImpulseFromVorticity.z
        , integralsAfter.mLinearImpulseFromVelocity.x   , integralsAfter.mLinearImpulseFromVelocity.y   , integralsAfter.mLinearImpulseFromVelocity.z
        , integralsAfter.mAngularImpulse.x              , integralsAfter.mAngularImpulse.y              , integralsAfter.mAngularImpulse.z              ) ;
}




/** Render summary diagnostic text.
*/
void InteSiVis::RenderSummaryDiagnosticText()
{
    QUERY_PERFORMANCE_ENTER ;

    VortonSim & vortonSim = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;

    if( mDiagnosticText != DIAG_TEXT_NONE )
    {   // Diagnostic text is enabled.
        // Render frame counter as text on screen.
        oglRenderString( Vec3( 10.0f , 10.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "frame=%5i time=%5.2g" , mFrame , mTimeNow ) ;
        // Render bounding box info.
        const Vec3 & bboxCenter = vortonSim.GetBoundingBoxCenter() ;
        const Vec3 & bboxSize   = vortonSim.GetBoundingBoxSize() ;
        oglRenderString( Vec3( 10.0f , 20.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "bbox@{ %g , %g , %g } spans { %g , %g , %g }" , bboxCenter.x , bboxCenter.y , bboxCenter.z , bboxSize.x , bboxSize.y , bboxSize.z ) ;
        // Render vorticity integrals.
        {
            const VortonSim::DiagnosticIntegrals & di = vortonSim.GetDiagnosticIntegrals() ;
            oglRenderString( Vec3( 10.0f , 30.0f , 0.0f ) , GLUT_BITMAP_8_BY_13 , "init        I0={% 8.1f,% 8.1f,% 8.1f} I1={% 8.1f,% 8.1f,% 8.1f},{% 8.1f,% 8.1f,% 8.1f} I2={% 8.1f,% 8.1f,% 8.1f}"
                , di.mInitial.mTotalCirculation.x                   , di.mInitial.mTotalCirculation.y                   , di.mInitial.mTotalCirculation.z
                , di.mInitial.mLinearImpulseFromVorticity.x         , di.mInitial.mLinearImpulseFromVorticity.y         , di.mInitial.mLinearImpulseFromVorticity.z
                , di.mInitial.mLinearImpulseFromVelocity.x          , di.mInitial.mLinearImpulseFromVelocity.y          , di.mInitial.mLinearImpulseFromVelocity.z
                , di.mInitial.mAngularImpulse.x                     , di.mInitial.mAngularImpulse.y                     , di.mInitial.mAngularImpulse.z              ) ;
            RenderDiagnosticIntegralsText(  40.0f , "befr" , di.mInitial         , di.mBefore           ) ;
            RenderDiagnosticIntegralsText(  50.0f , "advc" , di.mBefore          , di.mAfterAdvect      ) ;
            RenderDiagnosticIntegralsText(  60.0f , "rgrd" , di.mAfterAdvect     , di.mAfterRegrid      ) ;
            RenderDiagnosticIntegralsText(  70.0f , "velg" , di.mAfterRegrid     , di.mAfterVelGrid     ) ;
            RenderDiagnosticIntegralsText(  80.0f , "stre" , di.mAfterVelGrid    , di.mAfterStretch     ) ;
            RenderDiagnosticIntegralsText(  90.0f , "baro" , di.mAfterStretch    , di.mAfterBaroclinic  ) ;
            RenderDiagnosticIntegralsText( 100.0f , "difs" , di.mAfterBaroclinic , di.mAfterDiffuse     ) ;
            RenderDiagnosticIntegralsText( 110.0f , "heat" , di.mAfterDiffuse    , di.mAfterHeat        ) ;
        }
        // Render vorticity statistics.
        {
            float vorticityMin , vorticityMax , vorticityMean , vorticityStdDev ;
            Vec3 centerOfVorticity ;

            vortonSim.GatherVorticityStats( vorticityMin , vorticityMax , vorticityMean , vorticityStdDev , centerOfVorticity ) ;
            oglRenderString( Vec3( 10.0f , 120.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "omega:[%.0f,%.0f]<%g>+-%-.0f #=%i CoV=%g,%g,%g"
                , vorticityMin , vorticityMax , vorticityMean , vorticityStdDev , vortonSim.GetVortons()->size()
                , centerOfVorticity.x , centerOfVorticity.y , centerOfVorticity.z
                ) ;
            const VortonSim::VorticityTermsStatistics & vortTermStats = vortonSim.GetVorticityTermsStatistics() ;

            VortonSim::InvestigationTermE & vorticityInvestigationTerm = vortonSim.GetInvestigationTerm() ;
            oglRenderString( Vec3( 10.0f , 130.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "%c stretchTilt     :[%.7f,%.7f]<%g>+-%-.7f"
                , vorticityInvestigationTerm == VortonSim::INVESTIGATE_ALL || vorticityInvestigationTerm == VortonSim::INVESTIGATE_STRETCHING_TILTING ? '*' : ' '
                , vortTermStats.mStretchTilt.mMin , vortTermStats.mStretchTilt.mMax , vortTermStats.mStretchTilt.mMean , vortTermStats.mStretchTilt.mStdDev ) ;
            oglRenderString( Vec3( 10.0f , 140.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "%c baroclinic      :[%.7f,%.7f]<%g>+-%-.7f"
                , vorticityInvestigationTerm == VortonSim::INVESTIGATE_ALL || vorticityInvestigationTerm == VortonSim::INVESTIGATE_BAROCLINIC ? '*' : ' '
                , vortTermStats.mBaroclinic.mMin , vortTermStats.mBaroclinic.mMax , vortTermStats.mBaroclinic.mMean , vortTermStats.mBaroclinic.mStdDev ) ;
            oglRenderString( Vec3( 10.0f , 150.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "%c viscousDiffusion:[%.7f,%.7f]<%g>+-%-.7f"
                , vorticityInvestigationTerm == VortonSim::INVESTIGATE_ALL || vorticityInvestigationTerm == VortonSim::INVESTIGATE_VISCOUS_DIFFUSION ? '*' : ' '
                , vortTermStats.mViscousDiffusion.mMin , vortTermStats.mViscousDiffusion.mMax , vortTermStats.mViscousDiffusion.mMean , vortTermStats.mViscousDiffusion.mStdDev ) ;
        }
        // Render vorton velocity statistics.
        {
            float velocityMin , velocityMax , velocityMean , velocityStdDev ;
            Vec3 centerOfVelocity ;
            Particles::ComputeVelocityStats( * reinterpret_cast< VECTOR< Particle > * >( vortonSim.GetVortons() ) , velocityMin , velocityMax , velocityMean , velocityStdDev , centerOfVelocity ) ;

            oglRenderString( Vec3( 10.0f , 160.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "vel:[%.3f,%.3f]<%.3f>+-%-.3f"
                , velocityMin , velocityMax , velocityMean , velocityStdDev
                ) ;
        }
        // Render vorton fluid temperature statistics.
        {
            float temperatureMin , temperatureMax , temperatureMean , temperatureStdDev ;
            vortonSim.GatherTemperatureStats( temperatureMin , temperatureMax , temperatureMean , temperatureStdDev ) ;

            oglRenderString( Vec3( 10.0f , 170.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "Temp:[%.0f,%.0f]<%.0f>+-%-.0f"
                , temperatureMin , temperatureMax , temperatureMean , temperatureStdDev
                ) ;
        }
        // Render vorton fluid density statistics.
        {
            float densityMin , densityMax , densityMean , densityStdDev ;
            vortonSim.GatherDensityStats( densityMin , densityMax , densityMean , densityStdDev ) ;

            float densitySphMin , densitySphMax , densitySphMean , densitySphStdDev ;
            vortonSim.GatherDensitySphStats( densitySphMin , densitySphMax , densitySphMean , densitySphStdDev ) ;

            float numDensSphMin , numDensSphMax , numDensSphMean , numDensSphStdDev ;
            vortonSim.GatherNumberDensitySphStats( numDensSphMin , numDensSphMax , numDensSphMean , numDensSphStdDev ) ;

            float densGradSphMin , densGradSphMax , densGradSphMean , densGradSphStdDev ;
            vortonSim.GatherDensityGradientStats( densGradSphMin , densGradSphMax , densGradSphMean , densGradSphStdDev ) ;

            oglRenderString( Vec3( 10.0f , 180.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "dens:[%.3f,%.3f]<%.3f>+-%-.3f    densSph:[%.3f,%.3f]<%.3f>+-%-.3f    #densSph:[%.3f,%.3f]<%.3f>+-%-.3f    densGradSph:[%.3f,%.3f]<%.3f>+-%-.3f"
                , densityMin , densityMax , densityMean , densityStdDev
                , densitySphMin , densitySphMax , densitySphMean , densitySphStdDev
                , numDensSphMin , numDensSphMax , numDensSphMean , numDensSphStdDev
                , densGradSphMin , densGradSphMax , densGradSphMean , densGradSphStdDev
                ) ;
        }
        // Render vorton fluid proximity-to-walls statistics.
        {
            float proximityMin , proximityMax , proximityMean , proximityStdDev ;
            vortonSim.GatherProximityStats( proximityMin , proximityMax , proximityMean , proximityStdDev ) ;

            oglRenderString( Vec3( 10.0f , 190.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "proximity:[%.3f,%.3f]<%.3f>+-%-.3f"
                , proximityMin , proximityMax , proximityMean , proximityStdDev
                ) ;
        }
        {
            extern float gVortonSim_DisplacementMax ;
            extern int gVortonSim_NumRelaxationIters ;
            oglRenderString( Vec3( 10.0f , 200.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "diplacementMax=%g numIters=%i" , gVortonSim_DisplacementMax , gVortonSim_NumRelaxationIters ) ;
        }
        // Render tracer statistics.
        {
            oglRenderString( Vec3( 10.0f , 210.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "tracers: #=%i"
                , mTracerPclGrpInfo.mParticleGroup->GetParticles().Size()
                ) ;
        }
        // Render diagnostic state.
        oglRenderString( Vec3( 10.0f , 220.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "%s    %s    %s    %s    %s    %s"
            , GridDecorationString() , GridFieldString()
            , mEmphasizeCameraTarget ? "emphasis" : "NO_emp"
            , VortonRenderingString()
            , VortonPropertyString()
            , TracerRenderingString() ) ;

        // Render fluid simulation technique
        oglRenderString( Vec3( 10.0f , 230.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "%s"
            , FluidSimulationTechniqueString() ) ;

        // Render rigid body info.
        for( size_t iSphere = 0 ; iSphere < GetSpheres().size() ; ++ iSphere )
        {   // For each sphere...
            // Display its temperature, linear velocity and angular velocity.
            const float yPos = 240.0f + float( 10 * iSphere ) ;
            oglRenderString( Vec3( 10.0f , yPos , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "ball[%i].T=%g  v={%g,%g,%g} w={%g,%g,%g}"
                , iSphere
                , GetSpheres()[ iSphere ].GetThermalProperties().GetTemperature()
                , GetSpheres()[ iSphere ].GetBody()->GetVelocity().x
                , GetSpheres()[ iSphere ].GetBody()->GetVelocity().y
                , GetSpheres()[ iSphere ].GetBody()->GetVelocity().z
                , GetSpheres()[ iSphere ].GetBody()->GetAngularVelocity().x
                , GetSpheres()[ iSphere ].GetBody()->GetAngularVelocity().y
                , GetSpheres()[ iSphere ].GetBody()->GetAngularVelocity().z
                ) ;
        }
        for( size_t iBox = 0 ; iBox < GetBoxes().size() ; ++ iBox )
        {   // For each box...
            // Display its temperature, linear velocity and angular velocity.
            const float yPos = 250.0f + float( 10 * ( iBox + GetSpheres().size() ) ) ;
            oglRenderString( Vec3( 10.0f , yPos , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "box[%i].T=%g v={%g,%g,%g} w={%g,%g,%g}"
                , iBox
                , GetBoxes()[ iBox ].GetThermalProperties().GetTemperature()
                , GetBoxes()[ iBox ].GetBody()->GetVelocity().x
                , GetBoxes()[ iBox ].GetBody()->GetVelocity().y
                , GetBoxes()[ iBox ].GetBody()->GetVelocity().z
                , GetBoxes()[ iBox ].GetBody()->GetAngularVelocity().x
                , GetBoxes()[ iBox ].GetBody()->GetAngularVelocity().y
                , GetBoxes()[ iBox ].GetBody()->GetAngularVelocity().z
                ) ;
        }


    #if ENABLE_PARTICLE_JERK_RECORD
        // Render jerk statistics.
        {
            static FILE * jerkFile = 0 ;
            if( 0 == jerkFile )
            {   // Jerk output file is not yet open.
                char filename[ 256 ] = "TestData/jerk" ;
            #if ( VELOCITY_TECHNIQUE == VELOCITY_TECHNIQUE_DIRECT )
                strcat( filename , "-DIRECT" ) ;
            #elif ( VELOCITY_TECHNIQUE == VELOCITY_TECHNIQUE_TREE )
                strcat( filename , "-TREE" ) ;
            #elif ( VELOCITY_TECHNIQUE == VELOCITY_TECHNIQUE_POISSON_GAUSS_SEIDEL )
                strcat( filename , "-POISSON" ) ;
            #endif
            #if COMPUTE_VELOCITY_AT_VORTONS
                strcat( filename , "-VAV" ) ;
            #endif
            #if USE_PARTICLE_IN_CELL
                strcat( filename , "-PIC" ) ;
            #endif
            #if USE_ORIGINAL_VORTONS_IN_BASE_LAYER
                strcat( filename , "-OVIBL" ) ;
            #endif
            #if ENABLE_AUTO_MOLLIFICATION
                strcat( filename , "-AM" ) ;
            #endif
                strcat( filename , ".dat" ) ;

                jerkFile = fopen( filename , "w" ) ;

                fprintf( jerkFile , "# " ) ;
            #if ( VELOCITY_TECHNIQUE == VELOCITY_TECHNIQUE_DIRECT )
                fprintf( jerkFile , "DIRECT " ) ;
            #elif ( VELOCITY_TECHNIQUE == VELOCITY_TECHNIQUE_TREE )
                fprintf( jerkFile , "TREE " ) ;
            #elif ( VELOCITY_TECHNIQUE == VELOCITY_TECHNIQUE_POISSON_GAUSS_SEIDEL )
                fprintf( jerkFile , "POISSON " ) ;
            #endif
            #if COMPUTE_VELOCITY_AT_VORTONS
                fprintf( jerkFile , "COMPUTE_VELOCITY_AT_VORTONS " ) ;
            #endif
            #if USE_PARTICLE_IN_CELL
                fprintf( jerkFile , "USE_PARTICLE_IN_CELL " ) ;
            #endif
            #if USE_ORIGINAL_VORTONS_IN_BASE_LAYER
                fprintf( jerkFile , "USE_ORIGINAL_VORTONS_IN_BASE_LAYER " ) ;
            #endif
            #if ENABLE_AUTO_MOLLIFICATION
                fprintf( jerkFile , "ENABLE_AUTO_MOLLIFICATION " ) ;
            #endif
                fprintf( jerkFile , "\n" ) ;
                fflush( jerkFile ) ;
            }
            float jerk2Min , jerk2Max , jerk2Mean , jerk2StdDev ;
            VECTOR< Particle > & vortonsAsParticles = reinterpret_cast< VECTOR< Particle > & >( * vortonSim.GetVortons() ) ;
            Particles::ComputeJerkStatistics( vortonsAsParticles , jerk2Mean , jerk2StdDev , jerk2Min , jerk2Max ) ;
            if( IsNan( jerk2Mean ) || IsInf( jerk2Mean ) || IsNan( jerk2StdDev ) || IsInf( jerk2StdDev ) )
            {   // Not enough updates have occurred yet to compute jerk.  It takes 4 consecutive timesteps with TallyDiagnosticIntegrals enabled.
                oglRenderString( Vec3( 200.0f , 110.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "jerk: (insufficient data)" ) ;
            }
            else
            {   // Enough updates have occurred to compute jerk.
                oglRenderString( Vec3( 200.0f , 110.0f , 0.0f ) , GLUT_BITMAP_HELVETICA_10 , "jerk:[%.0f,%.0f]<%.0f>+-%-.0f" , sqrtf( jerk2Min ) , sqrtf( jerk2Max ) , sqrtf( jerk2Mean ) , sqrtf( jerk2StdDev ) ) ;
                fprintf( jerkFile , "%g %g %g %g %g\n" , mTimeNow , sqrtf( jerk2Mean ) , sqrtf( jerk2StdDev ) , sqrtf( jerk2Min ) , sqrtf( jerk2Max ) ) ;
                fflush( jerkFile ) ;
            }
        }
    #endif
    }
    else
    {
    }

    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_SummaryDiagnosticText ) ;
}




/** Gather and record performance profile data.
*/
void InteSiVis::GatherAndRecordProfileData()
{
#if defined( PROFILE )
    VortonSim & rVortonSim = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;

    if( mFrame == sFrameCountMax - 1 )
    {   // Reached last frame, so enable profile reporting
        gPrintProfileData = true ;
    }
    else if( mFrame >= sFrameCountMax )
    {   // Reached maximum number of frames, so exit.
        fprintf( stderr , "Ran %i frames\n" , mFrame ) ;
        fprintf( stderr , "region: { %g , %g , %g } , { %g , %g , %g } \n"
            , rVortonSim.GetMinCornerEternal().x
            , rVortonSim.GetMinCornerEternal().y
            , rVortonSim.GetMinCornerEternal().z
            , rVortonSim.GetMaxCornerEternal().x
            , rVortonSim.GetMaxCornerEternal().y
            , rVortonSim.GetMaxCornerEternal().z
            ) ;
        const float tracersPerFrame = (float) sNumTracersSum / (float) ( mFrame - 1 ) ;
        const float vortonsPerFrame = (float) sNumVortonsSum / (float) ( mFrame - 1 ) ;
        fprintf( stderr , "Particles, average per frame: %g tracers    %g vortons\n"
            , tracersPerFrame , vortonsPerFrame ) ;

        const float tracerHitsPerFrame = (float) FluidBodySim::GetNumTracerBodyHits() / (float) ( mFrame - 1 ) ;
        const float vortonHitsPerFrame = (float) FluidBodySim::GetNumVortonBodyHits() / (float) ( mFrame - 1 ) ;
        fprintf( stderr , "Particles-body hits, average per frame: %g tracers    %g vortons\n"
            , tracerHitsPerFrame , vortonHitsPerFrame ) ;
        fprintf( stderr , "Benchmark DONE: " __DATE__ " " __TIME__ "\n\n" ) ;

        exit( 0 ) ;
    }
#endif
}




/** Update particle systems.
*/
void InteSiVis::UpdateParticleSystems()
{
    QUERY_PERFORMANCE_ENTER ;
    if( mTimeStepping == PLAY || mTimeStepping == SINGLE_STEP )
    {
        #if USE_SMOOTHED_PARTICLE_HYDRODYNAMICS
        if(     ( VortonSim::FLUID_SIM_SMOOTHED_PARTICLE_HYDRODYNAMICS == mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetFluidSimulationTechnique() )
            ||  ( VortonSim::FLUID_SIM_VPM_SPH_HYBRID                  == mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetFluidSimulationTechnique() ) )
        {
            if( mVortonPclGrpInfo.mPclOpAssignVelocityFromField )
            {
                mVortonPclGrpInfo.mPclOpAssignVelocityFromField->mVelocityGrid   = 0 ; // obsolete
            }
            mVortonPclGrpInfo.mPclOpPopulateVelocityGrid->mVelocityGrid = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid() ;

            // In SPH mode, the density grid isn't updated, so has invalid bounds, which would make AssignScalarFromGrid fail.
            // Clear density grids to avoid that problem.
            // Note that in this mode, because density grid is not populated, tracers do not get their densities updated from vortons.
            // That has 2 effects: linear impulses are different, and rendering is different.
            const_cast< UniformGrid< float > & >( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetDensityGrid() ) .Clear() ;

        #if ENABLE_FIRE
            const_cast< UniformGrid< float > & >( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetFuelGrid() ).Clear() ;
            const_cast< UniformGrid< float > & >( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetFlameGrid() ).Clear() ;
            const_cast< UniformGrid< float > & >( mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetSmokeGrid() ).Clear() ;
        #endif
        }
        else
        #endif
        {
            if( mVortonPclGrpInfo.mPclOpAssignVelocityFromField )
            {
                // OBSOLETE
                mVortonPclGrpInfo.mPclOpAssignVelocityFromField->mVelocityGrid   = & mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVelocityGrid() ;
            }
            if( mVortonPclGrpInfo.mPclOpPopulateVelocityGrid )
            {
                mVortonPclGrpInfo.mPclOpPopulateVelocityGrid->mVelocityGrid = 0 ;
            }
        }

        mPclSysMgr.Update( mTimeStep , mFrame ) ;

        #if 0 && USE_SMOOTHED_PARTICLE_HYDRODYNAMICS // Silly VPM-SPH hybrid: Alternate between VPM and SPH.  This yields about the same results as running SPH after VPM each frame.
        if( VortonSim::FLUID_SIM_SMOOTHED_PARTICLE_HYDRODYNAMICS == mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetFluidSimulationTechnique() )
        {   // Previous fluid simulation update used SPH.
            // Switch to VPM for next Update.
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetFluidSimulationTechnique( VortonSim::FLUID_SIM_VORTEX_PARTICLE_METHOD ) ;
        }
        else
        {   // Previous fluid simulation update used VPM.
            // Switch to SPH for next Update.
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetFluidSimulationTechnique( VortonSim::FLUID_SIM_SMOOTHED_PARTICLE_HYDRODYNAMICS ) ;
        }
        #endif
    }
    QUERY_PERFORMANCE_EXIT( InteSiVis_ParticleSystem_Update ) ;

#if 0 // DO NOT SUBMIT -- unit testing
    {
        ParticleSystemManager * pclSysMgr = new ParticleSystemManager( mPclSysMgr ) ;
        pclSysMgr->Update( 0.0 , 0 ) ;
        mPclSysMgr.Update( mTimeStep , mFrame ) ;
        delete pclSysMgr ;
        mPclSysMgr.Update( mTimeStep , mFrame ) ;
    }
#endif
}




/** Update rigid bodies.
*/
void InteSiVis::UpdateRigidBodies()
{
    QUERY_PERFORMANCE_ENTER ;
    if( mTimeStepping == PLAY || mTimeStepping == SINGLE_STEP )
    {
        if( 28 == mScenario )
        {   // Box of fluid.
// TODO: eliminate this compile-time conditional.
#if USE_SMOOTHED_PARTICLE_HYDRODYNAMICS
            static Vec3 amplitude( 0.5f , 0.2f , 0.05f ) ;
            static Vec3 frequency( 3.0f , 5.0f , 7.0f ) ;
#else
            static Vec3 amplitude( 0.0f * 0.5f , 0.0f , 0.0f ) ;
            static Vec3 frequency( 0.2f , 0.0f , 0.0f ) ;
#endif
            if( ( amplitude.Mag2() != 0.0f ) && ( frequency.Mag2() != 0.0f ) )
            {
                RbBox & box = GetBoxes().Back() ;
                Impulsion::RigidBody * rigidBody = box.GetBody() ;
                Vec3 position = rigidBody->GetPosition() ;
                // Make box slosh back and forth.
                position.x = amplitude.x * sin( mTimeNow * frequency.x ) ;
                position.y = amplitude.y * cos( mTimeNow * frequency.y ) ;
                position.z = amplitude.z * sin( mTimeNow * frequency.z ) ;
                rigidBody->SetPosition( position ) ;
            }
        }
        PhysicalObject_UpdateSystem( GetPhysicalObjects() , mTimeStep , mFrame ) ;
    }
    QUERY_PERFORMANCE_EXIT( InteSiVis_RigidBody_UpdateSystem ) ;
}




const char * InteSiVis::FluidSimulationTechniqueString() const
{
    VortonSim & vortonSim = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;

    switch( vortonSim.GetFluidSimulationTechnique() )
    {
        CASE_STRING_FROM_TOKEN( VortonSim::FLUID_SIM_VORTEX_PARTICLE_METHOD             ) ;
        CASE_STRING_FROM_TOKEN( VortonSim::FLUID_SIM_SMOOTHED_PARTICLE_HYDRODYNAMICS    ) ;
        CASE_STRING_FROM_TOKEN( VortonSim::FLUID_SIM_VPM_SPH_HYBRID                     ) ;
        CASE_STRING_FROM_TOKEN( VortonSim::FLUID_SIM_NUM                                ) ;
    }
    return "(invalid FluidSimulationTechniqueE)" ;
}




const char * InteSiVis::GridDecorationString() const
{
    switch( mGridDecorations )
    {
        CASE_STRING_FROM_TOKEN( GRID_DECO_NONE              ) ;
        CASE_STRING_FROM_TOKEN( GRID_DECO_BOUNDING_BOX      ) ;
        CASE_STRING_FROM_TOKEN( GRID_DECO_CELLS             ) ;
        CASE_STRING_FROM_TOKEN( GRID_DECO_POINTS            ) ;
        CASE_STRING_FROM_TOKEN( GRID_DECO_CELLS_AND_POINTS  ) ;
    }
    return "(invalid GridDecorationE)" ;
}




const char * InteSiVis::GridFieldString() const
{
    switch( mGridField )
    {
        CASE_STRING_FROM_TOKEN( GRID_FIELD_VELOCITY          ) ;
        CASE_STRING_FROM_TOKEN( GRID_FIELD_DENSITY_GRADIENT  ) ;
        CASE_STRING_FROM_TOKEN( GRID_FIELD_PRESSURE_GRADIENT ) ;
    }
    return "(invalid GridFieldE)" ;
}




const char * InteSiVis::VortonRenderingString() const
{
    switch( mVortonRendering )
    {
        CASE_STRING_FROM_TOKEN( VORTON_RENDER_NONE                  ) ;
        CASE_STRING_FROM_TOKEN( VORTON_RENDER_DIAGNOSTIC_PARTICLES  ) ;
        CASE_STRING_FROM_TOKEN( VORTON_RENDER_SIMPLE_PARTICLES      ) ;
        CASE_STRING_FROM_TOKEN( VORTON_RENDER_VECTORS               ) ;
        CASE_STRING_FROM_TOKEN( VORTON_RENDER_PARTICLES_AND_VECTORS ) ;
        CASE_STRING_FROM_TOKEN( VORTON_RENDER_PATHLINES             ) ;
        CASE_STRING_FROM_TOKEN( VORTON_RENDER_ALL                   ) ;
    }
    return "(invalid VortonRenderingE)" ;
}




const char * InteSiVis::VortonPropertyString() const
{
    switch( mVortonProperty )
    {
        CASE_STRING_FROM_TOKEN( VORTON_PROPERTY_POSITION            ) ;
        CASE_STRING_FROM_TOKEN( VORTON_PROPERTY_VELOCITY            ) ;
        CASE_STRING_FROM_TOKEN( VORTON_PROPERTY_VORTICITY           ) ;
        CASE_STRING_FROM_TOKEN( VORTON_PROPERTY_DENSITY             ) ;
        CASE_STRING_FROM_TOKEN( VORTON_PROPERTY_DENSITY_SPH         ) ;
        CASE_STRING_FROM_TOKEN( VORTON_PROPERTY_DENSITY_GRADIENT    ) ;
        CASE_STRING_FROM_TOKEN( VORTON_PROPERTY_PROXIMITY           ) ;
    }
    return "(invalid VortonPropertyE)" ;
}




const char * InteSiVis::TracerRenderingString() const
{
    switch( mTracerRendering )
    {
        CASE_STRING_FROM_TOKEN( TRACER_RENDER_NONE  ) ;
        CASE_STRING_FROM_TOKEN( TRACER_RENDER_FUEL  ) ;
        CASE_STRING_FROM_TOKEN( TRACER_RENDER_FLAME ) ;
        CASE_STRING_FROM_TOKEN( TRACER_RENDER_SMOKE ) ;
        CASE_STRING_FROM_TOKEN( TRACER_RENDER_FIRE  ) ;
        CASE_STRING_FROM_TOKEN( TRACER_RENDER_DYE   ) ;
    }
    return "(invalid TracerRenderingE)" ;
}




/** Function that GLUT calls to display contents of window.
*/
/* static */ void InteSiVis::GlutDisplayCallback()
{
    QUERY_PERFORMANCE_ENTER ;

    if( ! sInstance->mInitialized )
    {   // This is the first time this app has displayed anything.
        sInstance->InitializeRendering() ;
        sInstance->CreateEntities() ;
    }

    sInstance->mCamera.SetCamera() ;
    sInstance->RenderSky() ;
    sInstance->RenderRigidBodies() ;
    sInstance->RenderParticles() ;
    sInstance->RenderSummaryDiagnosticText() ;

    QUERY_PERFORMANCE_ENTER ;
    glutSwapBuffers() ;
    QUERY_PERFORMANCE_EXIT( InteSiVis_Render_SwapBuffers ) ;

#if PROFILE > 1
    sInstance->GatherAndRecordProfileData() ;
#endif

    QUERY_PERFORMANCE_EXIT( InteSiVis_Render ) ;
}




/** Function that GLUT calls when nothing else is happening.
*/
/* static */ void InteSiVis::GlutIdleCallback()
{
    sInstance->Idle() ;
}




/** Function to handle mouse motion.
*/
/* static */ void InteSiVis::MouseMotionHandler( int x , int y , int modifierKeys )
{
    if( -999 == sMousePrevX )
    {
        sMousePrevX = x ;
        sMousePrevY = y ;
        return ;
    }

    static float fMouseMotionSensitivity = 0.005f ;    // Tune based on how fast you want mouse to move camera.

    const float dx = Clamp( fMouseMotionSensitivity * float( x - sMousePrevX ) , -100.0f , 100.0f ) ;
    const float dy = Clamp( fMouseMotionSensitivity * float( y - sMousePrevY ) , -100.0f , 100.0f ) ;

    if( mMouseButtons[0] )
    {   // User is left-click-dragging mouse
        if( GLUT_ACTIVE_SHIFT == modifierKeys )
        {   // Shift held.
            Vec3                        angles      ;
            Impulsion::PhysicalObject * physObj     = GetPhysicalObjectInFocus() ;
            if( physObj )
            {
                Impulsion::RigidBody *  rigidBody   = physObj->GetBody() ;
                const Mat33 &           orientOld   = rigidBody->GetOrientation() ;
                orientOld.GetAngles( angles ) ;
                angles.x += dy ;
                angles.z += dx ;
                Mat33 orientNew ;
                orientNew.SetRotationXYZ( angles ) ;
                rigidBody->SetOrientation( orientNew ) ;
            }
        }
        else
        {
            QdCamera & cam = mCamera ;
            float azimuth , elevation , radius ;
            // Obtain previous camera parameters.
            cam.GetOrbit( azimuth , elevation , radius ) ;
            // Avoid gimbal lock by limiting elevation angle to avoid the poles.
            static const float sAvoidPoles = 0.001f ;
            elevation = Clamp( elevation - dy , sAvoidPoles , PI * ( 1.0f - sAvoidPoles ) ) ;
            // Set new camera parameters based on how much mouse moved.
            cam.SetOrbit( azimuth - dx , elevation , radius ) ;
        }
    }
    else if( mMouseButtons[2] )
    {   // User is right-click-dragging mouse
        if( GLUT_ACTIVE_SHIFT == modifierKeys )
        {   // Shift held.
            Impulsion::PhysicalObject * physObj     = GetPhysicalObjectInFocus() ;
            if( physObj )
            {
                Mat44 viewMatrix ;
                glGetFloatv( GL_MODELVIEW_MATRIX , (GLfloat *) & viewMatrix ) ;
                Vec3 viewRight  ( viewMatrix.m[0][0] , viewMatrix.m[1][0] , viewMatrix.m[2][0] ) ;
                Vec3 viewUp     ( viewMatrix.m[0][1] , viewMatrix.m[1][1] , viewMatrix.m[2][1] ) ;
                Vec3 viewForward( viewMatrix.m[0][2] , viewMatrix.m[1][2] , viewMatrix.m[2][2] ) ;
                Vec3 delta      = 1.0f * viewForward * ( dx + dy ) ;

                Impulsion::RigidBody *  rigidBody   = physObj->GetBody() ;
                Vec3                    trans       = rigidBody->GetPosition() ;
                trans.x += delta.x ;
                trans.y += delta.y ;
                trans.z += delta.z ;
                rigidBody->SetPosition( trans ) ;            }
        }
        else
        {
            QdCamera & cam = mCamera ;
            float azimuth , elevation , radius ;
            // Obtain previous camera parameters.
            cam.GetOrbit( azimuth , elevation , radius ) ;
            // Set new camera parameters based on how much mouse moved.
            const float newRadius = Max2( radius - ( dx + dy ) , 1.0e-4f ) ;
            cam.SetOrbit( azimuth , elevation , newRadius ) ;
        }
    }
    else if( mMouseButtons[1] )
    {   // User is middle-click-dragging mouse
        // Extract world space direction vectors associated with view (used to compute camera-facing coordinates).
        // Note that these vectors are the unit vectors of the inverse of the view matrix.
        // They are the world-space unit vectors of the view transformation.
        Mat44 viewMatrix ;
        glGetFloatv( GL_MODELVIEW_MATRIX , (GLfloat *) & viewMatrix ) ;
        Vec3 viewRight  ( viewMatrix.m[0][0] , viewMatrix.m[1][0] , viewMatrix.m[2][0] ) ;
        Vec3 viewUp     ( viewMatrix.m[0][1] , viewMatrix.m[1][1] , viewMatrix.m[2][1] ) ;
        Vec3 viewForward( viewMatrix.m[0][2] , viewMatrix.m[1][2] , viewMatrix.m[2][2] ) ;
        Vec3 delta      = 1.0f * ( - viewRight * dx + viewUp * dy ) ;
        if( GLUT_ACTIVE_SHIFT == modifierKeys )
        {   // Shift held.
            Impulsion::PhysicalObject * physObj     = GetPhysicalObjectInFocus() ;
            if( physObj )
            {
                Impulsion::RigidBody *  rigidBody   = physObj->GetBody() ;
                Vec3                    trans       = rigidBody->GetPosition() ;
                trans.x -= delta.x ;
                trans.y -= delta.y ;
                trans.z -= delta.z ;
                rigidBody->SetPosition( trans ) ;
            }
        }
        else
        {
            QdCamera & cam = mCamera ;
            const Vec3 & rEye    = cam.GetEye() ;
            const Vec3 & rTarget = cam.GetTarget() ;
            cam.SetEye   ( Vec3( rEye.x    + delta.x , rEye.y    + delta.y , rEye.z    + delta.z ) ) ;
            cam.SetTarget( Vec3( rTarget.x + delta.x , rTarget.y + delta.y , rTarget.z + delta.z ) ) ;
        }
    }

    sMousePrevX = x ;
    sMousePrevY = y ;
}




void InteSiVis::Idle()
{
    QUERY_PERFORMANCE_ENTER ;
#if USE_TBB
    tbb::tick_count time0 = tbb::tick_count::now() ;
#endif

    UpdateParticleSystems() ;

#if USE_TBB
    tbb::tick_count timeFinal = tbb::tick_count::now() ;
    //fprintf( stderr , " tbb duration=%g second\n" , (timeFinal - time0).seconds() ) ;
#endif

    mCamera.Update() ;

    InteSiVis::GlutDisplayCallback() ;

    UpdateRigidBodies() ;

    QUERY_PERFORMANCE_EXIT( InteSiVis_UPDATE_Michael_J_Gourlay_2009 ) ;

    if( mTimeStepping == PLAY || mTimeStepping == SINGLE_STEP )
    {   // Step time by 1.
        if( mTimeStep > 0.0f )
        {   // Step time forward.
            ++ mFrame ;
        }
        else if( mTimeStep < 0.0f )
        {   // Step time backward.
            -- mFrame ;
        }
        mTimeNow += mTimeStep ;
        if( mTimeStepping == SINGLE_STEP )
        {   // Now that simulation single-stepped, return to pause mode.
            mTimeStepping = PAUSE ;
        }
    }

#if PROFILE
    sNumTracersSum += mTracerPclGrpInfo.mParticleGroup->GetNumParticles() ;
    sNumVortonsSum += mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.GetVortons()->Size() ;
#endif
}



/** Function to handle a special key.
*/
/* static */ void InteSiVis::SpecialKeyHandler( int key , const int modifierKeys , int /* windowRelativeMouseX */ , int /* windowRelativeMouseY */ )
{
    QdCamera &      cam     = mCamera ;
    Vec3            vTarget = cam.GetTarget() ;

    if( ! modifierKeys )
    {   // No modifier keys (Shift, Ctrl or Alt) were pressed.
        switch( key )
        {
            case GLUT_KEY_F1 : InitialConditions(  1 ) ; break;
            case GLUT_KEY_F2 : InitialConditions(  2 ) ; break;
            case GLUT_KEY_F3 : InitialConditions(  3 ) ; break;
            case GLUT_KEY_F4 : InitialConditions(  4 ) ; break;
            case GLUT_KEY_F5 : InitialConditions(  5 ) ; break;
            case GLUT_KEY_F6 : InitialConditions(  6 ) ; break;
            case GLUT_KEY_F7 : InitialConditions(  7 ) ; break;
            case GLUT_KEY_F8 : InitialConditions(  8 ) ; break;
            case GLUT_KEY_F9 : InitialConditions(  9 ) ; break;
            case GLUT_KEY_F11: InitialConditions( 10 ) ; break;

            case GLUT_KEY_UP   : mTimeStepping   = PLAY  ; break ;
            case GLUT_KEY_DOWN : mTimeStepping   = PAUSE ; break ;
            case GLUT_KEY_RIGHT: mTimeStep       =  sTimeStep ; if( mTimeStepping == PAUSE ) mTimeStepping = SINGLE_STEP ; break ;
            case GLUT_KEY_LEFT : mTimeStep       = -sTimeStep ; if( mTimeStepping == PAUSE ) mTimeStepping = SINGLE_STEP ; break ;

            default:
            return;
            break;
        }
    }
    else if( GLUT_ACTIVE_SHIFT == modifierKeys )
    {   // Shift held while special key pressed.
        switch( key )
        {
            case GLUT_KEY_F1 : InitialConditions( 11 ) ; break;
            case GLUT_KEY_F2 : InitialConditions( 12 ) ; break;
            case GLUT_KEY_F3 : InitialConditions( 13 ) ; break;
            case GLUT_KEY_F4 : InitialConditions( 14 ) ; break;
            case GLUT_KEY_F5 : InitialConditions( 15 ) ; break;
            case GLUT_KEY_F6 : InitialConditions( 16 ) ; break;
            case GLUT_KEY_F7 : InitialConditions( 17 ) ; break;
            case GLUT_KEY_F8 : InitialConditions( 18 ) ; break;
            case GLUT_KEY_F9 : InitialConditions( 19 ) ; break;
            case GLUT_KEY_F11: InitialConditions( 20 ) ; break;

            case GLUT_KEY_UP   : mPhysObjFocus  ++ ; break ;
            case GLUT_KEY_DOWN : mPhysObjFocus  -- ; break ;
            case GLUT_KEY_RIGHT: mPhysObjFocus  ++ ; break ;
            case GLUT_KEY_LEFT : mPhysObjFocus  -- ; break ;

            default:
            return;
            break;
        }
    }
    else if( GLUT_ACTIVE_CTRL == modifierKeys )
    {   // Ctrl held while special key pressed.
        switch( key )
        {
            case GLUT_KEY_F1 : InitialConditions( 21 ) ; break;
            case GLUT_KEY_F2 : InitialConditions( 22 ) ; break;
            case GLUT_KEY_F3 : InitialConditions( 23 ) ; break;
            case GLUT_KEY_F4 : InitialConditions( 24 ) ; break;
            case GLUT_KEY_F5 : InitialConditions( 25 ) ; break;
            case GLUT_KEY_F6 : InitialConditions( 26 ) ; break;
            case GLUT_KEY_F7 : InitialConditions( 27 ) ; break;
            case GLUT_KEY_F8 : InitialConditions( 28 ) ; break;
            case GLUT_KEY_F9 : InitialConditions( 29 ) ; break;
            case GLUT_KEY_F11: InitialConditions( 30 ) ; break;

            default:
            return;
            break;
        }
    }
    else if( GLUT_ACTIVE_ALT == modifierKeys )
    {   // Alt held while special key pressed.
        switch( key )
        {
            case GLUT_KEY_F1 : InitialConditions( 31 ) ; break;
            case GLUT_KEY_F2 : InitialConditions( 32 ) ; break;
            case GLUT_KEY_F3 : InitialConditions( 33 ) ; break;
            case GLUT_KEY_F4 : InitialConditions( 34 ) ; break;
            case GLUT_KEY_F5 : InitialConditions( 35 ) ; break;
            case GLUT_KEY_F6 : InitialConditions( 36 ) ; break;
            case GLUT_KEY_F7 : InitialConditions( 37 ) ; break;
            case GLUT_KEY_F8 : InitialConditions( 38 ) ; break;
            case GLUT_KEY_F9 : InitialConditions( 39 ) ; break;
            case GLUT_KEY_F11: InitialConditions( 40 ) ; break;

            default:
            return;
            break;
        }
    }

    // Make sure mPhysObjFocus has a legitimate value for the current scenario.
    if( ! mPhysicalObjects.Empty() )
    {
        mPhysObjFocus %= mPhysicalObjects.size() ;
    }
    else
    {   // Avoid divide-by-zero.
        mPhysObjFocus = 0 ;  // Actual value does not matter; code elsewhere cannot legally reference mPhysObjFocus if mPhysicalObjects is empty.
    }
}





/** Function that GLUT calls to handle a special key.
*/
/* static */ void InteSiVis::GlutSpecialKeyHandler( int key , int windowRelativeMouseX , int windowRelativeMouseY )
{
    sInstance->mModifierKeys = glutGetModifiers() ;

    sInstance->SpecialKeyHandler( key , sInstance->mModifierKeys , windowRelativeMouseX , windowRelativeMouseY ) ;

    glutPostRedisplay() ;
}




void InteSiVis::SetTallyDiagnosticIntegrals()
{
    switch( mDiagnosticText )
    {
        case DIAG_TEXT_SUMMARY   :
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetTallyDiagnosticIntegrals( true ) ;
            break ;
        case DIAG_TEXT_FULL:
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetTallyDiagnosticIntegrals( true ) ;
            break ;
        case DIAG_TEXT_NONE   :
            mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetTallyDiagnosticIntegrals( false ) ;
            break ;
    }
}




/** Function to handle a regular key.
*/
void InteSiVis::KeyboardHandler( unsigned char key , int /* mouseX */ , int /* mouseY */ )
{
    mModifierKeys = glutGetModifiers() ;

    QdCamera & cam = mCamera ;
    float azimuth , elevation , radius ;
    static bool animateCamera = false ;
    cam.GetOrbit( azimuth , elevation , radius ) ;
    switch( key )
    {
        case '?': gPrintProfileData = ! gPrintProfileData ; break ;

        case '.': radius -= 0.1f ; break ;  // Move camera in.
        case ',': radius += 0.1f ; break ;  // Move camera out.

        case 'c':   // Toggle camera rotation.
        case 'C':
        {
            animateCamera = ! animateCamera ;
            if( animateCamera )
            {   // Camera animation is now on.
                if( 'c' == key )
                {   // Rotate azimuth faster.
                    cam.SetOrbitalTrajectory( Vec3( -0.005f , -0.00f , 0.0f ) ) ;
                }
                else
                {   // Rotate elevation faster.
                    cam.SetOrbitalTrajectory( Vec3( -0.002f , -0.005f , 0.0f ) ) ;
                }
            }
            else
            {
                cam.SetOrbitalTrajectory( Vec3( 0.0f , 0.0f , 0.0f ) ) ;
            }
        }
        break ;

        case 'd':   // Cycle through diagnostic text options.
        case 'D':
        {
            switch( mDiagnosticText )
            {
                case DIAG_TEXT_NONE   :
                    mDiagnosticText = DIAG_TEXT_SUMMARY ;
                    mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetTallyDiagnosticIntegrals( true ) ;
                    break ;
                case DIAG_TEXT_SUMMARY:
                    mDiagnosticText = DIAG_TEXT_FULL    ;
                    mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetTallyDiagnosticIntegrals( true ) ;
                    break ;
                case DIAG_TEXT_FULL   :
                    mDiagnosticText = DIAG_TEXT_NONE    ;
                    mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim.SetTallyDiagnosticIntegrals( false ) ;
                    break ;
            }
        }
        break ;

        case 'e':   // Toggle emphasis on objects near camera look-at location.
        case 'E':
        {
            mEmphasizeCameraTarget = ! mEmphasizeCameraTarget ;
        }
        break ;

        case 'f':   // Cycle through which field to render as grid.
        case 'F':
        {
            switch( mGridField )
            {
                case GRID_FIELD_VELOCITY         : mGridField = GRID_FIELD_DENSITY_GRADIENT ; break ;
                case GRID_FIELD_DENSITY_GRADIENT :
                #if COMPUTE_PRESSURE_GRADIENT
                                                   mGridField = GRID_FIELD_PRESSURE_GRADIENT;
                #else
                                                   mGridField = GRID_FIELD_VELOCITY;
                #endif
                    break ;
                case GRID_FIELD_PRESSURE_GRADIENT: mGridField = GRID_FIELD_VELOCITY         ; break ;
            }
        }
        break ;

        case 'g':   // Cycle through grid decoration options.
        case 'G':
        {
            switch( mGridDecorations )
            {
            case GRID_DECO_NONE             : mGridDecorations = GRID_DECO_BOUNDING_BOX     ; break ;
            case GRID_DECO_BOUNDING_BOX     : mGridDecorations = GRID_DECO_CELLS            ; break ;
            case GRID_DECO_CELLS            : mGridDecorations = GRID_DECO_POINTS           ; break ;
            case GRID_DECO_POINTS           : mGridDecorations = GRID_DECO_CELLS_AND_POINTS ; break ;
            case GRID_DECO_CELLS_AND_POINTS : mGridDecorations = GRID_DECO_NONE             ; break ;
            }
        }
        break ;

        case 's':   // Cycle through fluid simulation techniques.
        case 'S':
        {
            VortonSim & vortonSim = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;
            switch( vortonSim.GetFluidSimulationTechnique() )
            {
            case VortonSim::FLUID_SIM_VORTEX_PARTICLE_METHOD            : vortonSim.SetFluidSimulationTechnique( VortonSim::FLUID_SIM_SMOOTHED_PARTICLE_HYDRODYNAMICS   ) ; break ;
            case VortonSim::FLUID_SIM_SMOOTHED_PARTICLE_HYDRODYNAMICS   : vortonSim.SetFluidSimulationTechnique( VortonSim::FLUID_SIM_VPM_SPH_HYBRID                    ) ; break ;
            case VortonSim::FLUID_SIM_VPM_SPH_HYBRID                    : vortonSim.SetFluidSimulationTechnique( VortonSim::FLUID_SIM_VORTEX_PARTICLE_METHOD            ) ; break ;
            }
        }
        break ;

        case 't':   // Cycle through tracer rendering options.
        case 'T':
        {
            switch( mTracerRendering )
            {
            case TRACER_RENDER_NONE : mTracerRendering = TRACER_RENDER_FUEL  ; break ;
            case TRACER_RENDER_FUEL : mTracerRendering = TRACER_RENDER_FLAME ; break ;
            case TRACER_RENDER_FLAME: mTracerRendering = TRACER_RENDER_SMOKE ; break ;
            case TRACER_RENDER_SMOKE: mTracerRendering = TRACER_RENDER_FIRE  ; break ;
            case TRACER_RENDER_FIRE : mTracerRendering = TRACER_RENDER_DYE   ; break ;
            case TRACER_RENDER_DYE  : mTracerRendering = TRACER_RENDER_NONE  ; break ;
            }
        }
        break ;

        case 'p':   // Cycle through vorton properties.
        case 'P':
        {
            switch( mVortonProperty )
            {
            case VORTON_PROPERTY_POSITION           : mVortonProperty = VORTON_PROPERTY_VELOCITY        ; break ;
            case VORTON_PROPERTY_VELOCITY           : mVortonProperty = VORTON_PROPERTY_VORTICITY       ; break ;
            case VORTON_PROPERTY_VORTICITY          : mVortonProperty = VORTON_PROPERTY_DENSITY         ; break ;
            case VORTON_PROPERTY_DENSITY            : mVortonProperty = VORTON_PROPERTY_DENSITY_SPH     ; break ;
            case VORTON_PROPERTY_DENSITY_SPH        : mVortonProperty = VORTON_PROPERTY_DENSITY_GRADIENT; break ;
            case VORTON_PROPERTY_DENSITY_GRADIENT   : mVortonProperty = VORTON_PROPERTY_PROXIMITY       ; break ;
            case VORTON_PROPERTY_PROXIMITY          : mVortonProperty = VORTON_PROPERTY_POSITION        ; break ;
            }
        }
        break ;

        case 'v':   // Cycle through vorton rendering options.
        case 'V':
        {
            switch( mVortonRendering )
            {
            case VORTON_RENDER_NONE                 : mVortonRendering = VORTON_RENDER_DIAGNOSTIC_PARTICLES ; break ;
            case VORTON_RENDER_DIAGNOSTIC_PARTICLES : mVortonRendering = VORTON_RENDER_SIMPLE_PARTICLES     ; break ;
            case VORTON_RENDER_SIMPLE_PARTICLES     : mVortonRendering = VORTON_RENDER_VECTORS              ; break ;
            case VORTON_RENDER_VECTORS              : mVortonRendering = VORTON_RENDER_PARTICLES_AND_VECTORS; break ;
            case VORTON_RENDER_PARTICLES_AND_VECTORS: mVortonRendering = VORTON_RENDER_PATHLINES            ; break ;
            case VORTON_RENDER_PATHLINES            : mVortonRendering = VORTON_RENDER_ALL                  ; break ;
            case VORTON_RENDER_ALL                  : mVortonRendering = VORTON_RENDER_NONE                 ; break ;
            }
        }
        break ;

        case 'w':   // Cycle through vorticity equation terms
        case 'W':
        {
            VortonSim & vortonSim   = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;
            VortonSim::InvestigationTermE & vorticityInvestigationTerm = vortonSim.GetInvestigationTerm() ;
            switch( vorticityInvestigationTerm )
            {
            case VortonSim::INVESTIGATE_ALL               : vorticityInvestigationTerm = VortonSim::INVESTIGATE_STRETCHING_TILTING  ; break ;
            case VortonSim::INVESTIGATE_STRETCHING_TILTING: vorticityInvestigationTerm = VortonSim::INVESTIGATE_BAROCLINIC          ; break ;
            case VortonSim::INVESTIGATE_BAROCLINIC        : vorticityInvestigationTerm = VortonSim::INVESTIGATE_VISCOUS_DIFFUSION   ; break ;
            case VortonSim::INVESTIGATE_VISCOUS_DIFFUSION : vorticityInvestigationTerm = VortonSim::INVESTIGATE_NONE                ; break ;
            case VortonSim::INVESTIGATE_NONE              : vorticityInvestigationTerm = VortonSim::INVESTIGATE_ALL                 ; break ;
            }
        }
        break ;

        case 'z':   // Zero velocity everywhere
        {
            VortonSim &         vortonSim   = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;
            VECTOR< Vorton > &  vortons     = * vortonSim.GetVortons() ;
            const size_t numVortons = vortons.size() ;
            for( size_t vortIdx = 0 ; vortIdx < numVortons ; ++ vortIdx )
            {
                Vorton & vorton = vortons[ vortIdx ] ;
                vorton.mVelocity = Vec3( 0.0f , 0.0f , 0.0f ) ;
            }
        }
        break ;

        case 'Z':   // Zero vorticity everywhere
        {
            VortonSim &         vortonSim   = mVortonPclGrpInfo.mPclOpVortonSim->mVortonSim ;
            VECTOR< Vorton > &  vortons     = * vortonSim.GetVortons() ;
            const size_t numVortons = vortons.size() ;
            for( size_t vortIdx = 0 ; vortIdx < numVortons ; ++ vortIdx )
            {
                Vorton & vorton = vortons[ vortIdx ] ;
                vorton.mAngularVelocity = Vec3( 0.0f , 0.0f , 0.0f ) ;
            }
        }
        break ;

        case 127 /* delete */ : ; break;

        case 27 /* escape */ :
            exit( 0 ) ;
        break;

        default:
            return;
            /*NOTREACHED*/
        break;
    }
    cam.SetOrbit( azimuth , elevation , radius ) ;
}




/* static */ void InteSiVis::GlutKeyboardHandler( unsigned char key , int mouseX , int mouseY )
{
    sInstance->KeyboardHandler( key , mouseX , mouseY ) ;
}




/** Function that GLUT calls to handle mouse button events.
*/
/* static */ void InteSiVis::GlutMouseHandler( int button , int state , int x , int y )
{
    sInstance->mModifierKeys = glutGetModifiers() ;

    if( GLUT_DOWN == state )
    {   // User is clicking a mouse button
        sInstance->mMouseButtons[ button ] = 1 ;
    }
    else
    {   // User released a mouse button
        sInstance->mMouseButtons[ button ] = 0 ;
    }
    sMousePrevX = x ;
    sMousePrevY = y ;
}




/** Function that GLUT calls to handle mouse motion.
*/
/* static */ void InteSiVis::GlutMouseMotionHandler( int x , int y )
{
    sInstance->MouseMotionHandler( x , y , sInstance->mModifierKeys ) ;
}




/** Function that GLUT calls when the window obtains focus.
*/
/* static */ void InteSiVis::GlutEntryHandler( int state )
{
    if( GLUT_LEFT == state )
    {   // Focus left this window so act like mouse buttons got released.
        sInstance->mMouseButtons[0] =
        sInstance->mMouseButtons[1] =
        sInstance->mMouseButtons[2] = 0 ;
    }
}




/// Return PhysicalObject in focus, if any.
Impulsion::PhysicalObject * InteSiVis::GetPhysicalObjectInFocus()
{
    if( ! GetPhysicalObjects().Empty() )
    {   // Simulation has at at least one PhysicalObject.
        mPhysObjFocus %= GetPhysicalObjects().Size() ;  // Make sure mPhysObjFocus has a legitimate value for current scenario.
        return GetPhysicalObjects()[ mPhysObjFocus ] ;
    }
    else
    {   // Simulation has no PhysicalObjects so return nothing.
        mPhysObjFocus = 0 ;  // Actual value does not matter; code elsewhere cannot legally reference mPhysObjFocus if mPhysicalObjects is empty.
        return 0 ;
    }
}




/// Typedef for pointer to SwapInterval function.
typedef BOOL (APIENTRY *PFNWGLSWAPINTERVALFARPROC)( int );

/// Pointer to SwapInterval function.
PFNWGLSWAPINTERVALFARPROC wglSwapIntervalEXT = 0;

/** Set the vertical synchronization interval.

    \param interval Number of frames before swapping buffers.
*/
void setVSync( int interval = 1 )
{
    const char * extensions = (char*) glGetString( GL_EXTENSIONS );

    if( strstr( extensions, "WGL_EXT_swap_control" ) == 0 )
    {   // Your computer does NOT support WGL_EXT_swap_control extension.
        return ;
    }
    else
    {
        wglSwapIntervalEXT = (PFNWGLSWAPINTERVALFARPROC)wglGetProcAddress( "wglSwapIntervalEXT" );
        if( wglSwapIntervalEXT )
        {   // Set the vsync mode
            wglSwapIntervalEXT(interval);
        }
    }
}




/// Callback meant for use with glDebugMessageCallbackARB, which nVidia apparently does not yet support.
static void CALLBACK DebugCallback(unsigned int /* source */ , unsigned int /* type */ , unsigned int /* id */ , unsigned int /* severity */ ,  int /* length */ , const char * /* message*/ , void * /* userParam */ )
{
}




/** Initialize display device.
*/
void InteSiVis::InitDevice( int * pArgc , char ** argv )
{
#if 0
    for(int iz=-1 ; iz < 2 ; ++ iz )
    for(int iy=-1 ; iy < 2 ; ++ iy )
    for(int ix=-1 ; ix < 2 ; ++ ix )
    {
        int offset = ix + iy * 3 + iz * 9 ;
        if( offset >= 0 )
        {
            printf("[ %2i , %2i , %2i ]=%i\n", ix, iy, iz , offset ) ;
        }
    }
#endif

    glutInit( pArgc , argv ) ;

    unsigned int mode =     GLUT_RGBA
                        |   GLUT_DEPTH
                        |   GLUT_DOUBLE
                        //|   GLUT_ACCUM
                        //|   GLUT_STENCIL
                        //|   GLUT_MULTISAMPLE
                        ;

    // Initialize the GL utility toolkit.
    glutInitDisplayMode( mode ) ;

    // Create render window
    glutInitWindowPosition( 0 , 0 ) ;
    glutInitWindowSize( sInstance->mCamera.GetWidth() , sInstance->mCamera.GetHeight() ) ;
    mRenderWindow = glutCreateWindow( "(C) 2009-2013 Michael J. Gourlay: VorteGrid Render Window" ) ;

    // Register GLUT callbacks for main window
    glutIdleFunc( GlutIdleCallback) ;
    glutDisplayFunc( GlutDisplayCallback) ;
    glutKeyboardFunc( GlutKeyboardHandler ) ;
    glutSpecialFunc( GlutSpecialKeyHandler ) ;
    glutReshapeFunc( GlutReshapeGlutCallback ) ;
    glutMotionFunc( GlutMouseMotionHandler ) ;
    glutMouseFunc( GlutMouseHandler ) ;
    glutPassiveMotionFunc( GlutMouseMotionHandler ) ;
    glutEntryFunc( GlutEntryHandler ) ;

    //glDebugMessageCallbackARB( & DebugCallback , NULL ) ;

    // Set the current window to the main window
    glutSetWindow( mRenderWindow ) ;

    setVSync( 0 ) ;

    glutMainLoop() ; // Relinquish control to GLUT.  This never returns.
}




/// Application entry point.
int main( int argc , char ** argv )
{
    fprintf( stderr , "Benchmark: " __DATE__ " " __TIME__ "\n" ) ;

#if 0 // DO NOT SUBMIT
    extern int other_sph_main (int argc, char** argv) ;
    other_sph_main( argc , argv ) ;
#endif

    XMM_Set_FlushToZero_DenormalAreZero() ;
    Setx87Precision( PRECISION_SINGLE ) ;
    //Unmaskx87FpExceptions() ;

#if PROFILE
    {   // Redirect stderr to a file to collect profile data.
        freopen( "profile.log" , "a" , stderr ) ;
        fprintf( stderr , "Benchmark: " __DATE__ " " __TIME__ "\n" ) ;
    }
#endif

    InteSiVis inteSiVis ;
    inteSiVis.InitDevice( & argc , argv ) ;
    return 0 ;
}
